﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Smith.NED
{
    public partial class ConsistencyChecker : UserControl
    {
        public ConsistencyChecker()
        {
            InitializeComponent();
        }

        JKLevel level
        {
            get
            {
                return GlobalVars.Level;
            }
        }

        public enum ErrorType
        {
            None,
            Sector,
            Surface,
            Thing,
            Cog
        }

        public class ConsistencyError
        {
            public string text = string.Empty;
            public ErrorType etype = ErrorType.None;
            public object itm = null;

            public override string ToString()
            {
                return text;
            }
        }

        public void ClearErrors()
        {
            Errors.Items.Clear();
        }

        public void AddError(string text, object itm)
        {
            ConsistencyError ce = new ConsistencyError();
            if (itm == null)
                ce.etype = ErrorType.None;
            else if (itm is JKSector)
                ce.etype = ErrorType.Sector;
            else if (itm is JKSurface)
                ce.etype = ErrorType.Surface;
            else if (itm is JKThing)
                ce.etype = ErrorType.Thing;
            else if (itm is JKCOG)
                ce.etype = ErrorType.Cog;
            else
                return;

            ce.itm = itm;

            int ni;
            switch (ce.etype)
            {
                case ErrorType.Sector:
                    ni = (itm as JKSector).num;
                    ce.text = text + "; sector " + ni.ToString();
                    break;

                case ErrorType.Surface:
                    JKSurface surf = itm as JKSurface;
                    ce.text = text + "; surface " + surf.sector.num.ToString() + "," + surf.num.ToString();
                    break;

                case ErrorType.Thing:
                    ni = (itm as JKThing).num;
                    ce.text = text + "; thing " + ni.ToString();
                    break;

                case ErrorType.Cog:
                    ni = level.cogs.IndexOf(itm as JKCOG);
                    ce.text = text + "; Cog " + ni.ToString();
                    break;

                case ErrorType.None:
                    ce.text = text;
                    break;
            }

            Errors.Items.Add(ce);
        }

        class FileList : Dictionary<string, bool>
        {
            public FileList()
                : base(StringComparer.InvariantCultureIgnoreCase)
            {

            }

            public bool NoFile(string name)
            {
                if (!ContainsKey(name))
                {
                    try
                    {
                        using (BaseFile bf = FileOperations.OpenGameFile(name))
                            Add(name, true);
                    }
                    catch
                    {
                        Add(name, false);
                    }
                }

                return !this[name];
            }
        }

        public void CheckResources()
        {
            Show();
            ClearErrors();

            FileList cmps = new FileList();
            FileList mats = new FileList();
            FileList wavs = new FileList();
            FileList cogs = new FileList();

            if (!string.IsNullOrEmpty(level.masterCMP) && cmps.NoFile(level.masterCMP))
                AddError("Level's Master CMP is not found", null);

            for (int i = 0; i < level.sectors.Count; i++)
            {
                JKSector sec = level.sectors[i];

                if (cmps.NoFile(sec.colormap))
                    AddError("Colormap not found", sec);

                if (!string.IsNullOrEmpty(sec.sound) && wavs.NoFile(sec.sound))
                    AddError("Sound not found", sec);

                for (int j = 0; j < sec.surfaces.Count; j++)
                {
                    JKSurface surf = sec.surfaces[j];
                    if (!string.IsNullOrEmpty(surf.material) && mats.NoFile(surf.material))
                        AddError("Material not found", surf);
                }
            }

            for (int i = 0; i < level.things.Count; i++)
            {
                JKThing th = level.things[i];

                if (Templates.IndexOfName(th.name) == -1)
                    AddError("Template not found", th);
            }

            for (int i = 0; i < level.cogs.Count; i++)
            {
                JKCOG cg = level.cogs[i];
                if (cogs.NoFile(cg.name))
                    AddError("Cog not found", cg);
            }

            if (GlobalVars.IsMOTS)
            {
                if (cmps.Count > 1)
                    AddError("Warning! More than 1 CMP used in the level", null);
            }
            else
            {
                if (cmps.Count > 3)
                    AddError("Warning! More than 3 CMP used in the level", null);
            }

            if (Errors.SelectedIndex <= 0 && Errors.Items.Count > 0)
                Errors.SelectedIndex = 0;
            else
                Errors_SelectedIndexChanged(Errors, new EventArgs());
        }

        public new void Show()
        {
            if (IsHandleCreated)//Visible)
            {
                //BringToFront();
                if (GlobalVars.NedMain != null)
                    GlobalVars.NedMain.tabControl3.SelectedTab = GlobalVars.NedMain.tabConsistencyCheck;
                return;
            }

            base.Show();
        }

        public void Check()
        {
            Show();
            ClearErrors();

            for (int i = 0; i < level.sectors.Count; i++)
            {
                JKSector sec = level.sectors[i];
                if (string.IsNullOrEmpty(sec.colormap.Trim()))
                    AddError("No colormap set for sector!", sec);

                if (sec.surfaces.Count < 4)
                {
                    AddError("Sector has less than 4 surfaces!", sec);
                    continue;
                }

                if (sec.vertices.Count < 4)
                {
                    AddError("Sector has less than 4 vertices!", sec);
                    continue;
                }

                if ((sec.flags & Sector.Flag.Preview3DO) == 0)
                {
                    if (!Utils.IsSectorConvex(sec))
                        AddError("Sector is not convex", sec);
                }

                for (int j = 0; j < sec.surfaces.Count; j++)
                {
                    JKSurface surf = sec.surfaces[j];

                    if (surf.adjoin == null && (surf.faceFlags & Surface.FaceFlag.Transparent) != 0)
                        AddError("Incorrect flags on solid surface", surf);

                    if (surf.vertices.Count > 24)
                        AddError("More than 24 vertices in a surface", surf);
                    if (surf.vertices.Count < 3)
                        AddError("Less than 3 vertices in a surface!", surf);

                    if (MathLib.CompareScalar(surf.normal.Magnitude, 1.0) != 0)
                        AddError("Incorrect normal - surface must be invalid!", surf);

                    if (surf.adjoin != null)
                    {
                        if (surf.adjoin.adjoin != surf)
                            AddError("Invalid reverse adjoin", surf);
                        if (!Utils.Do_Surf_Overlap(surf, surf.adjoin))
                            AddError("Adjoined surfaces don't overlap", surf);


                        if (surf.vertices.Count != surf.adjoin.vertices.Count)
                            AddError("Adjoined surfaces don't have same vertex count", surf);
                        // doesnt look like this check matters
                        /*else
                        {
                            int mismatches = 0;
                            for (int p = 0; p < surf.vertices.Count; p++)
                                if (surf.vertices[p].v != surf.adjoin.vertices[p].v)
                                    mismatches++;

                            if (mismatches != 0)
                                AddError($"Adjoined surfaces' vertex position(s) differ;  {mismatches}/{surf.vertices.Count} mismatches", surf);
                        }*/
                    }

                    for (int v = 0; v < surf.vertices.Count; v++)
                    {
                        Vertex vx = surf.vertices[v];

                        double d = surf.CalcD();
                        if (Math.Abs(surf.normal * vx.v - d) > 0.001)
                        {
                            AddError("The face is not planar", surf);
                            break;
                        }
                    }

                    if (!Utils.IsSurfConvex(surf))
                        AddError("The face is not convex", surf);
                }
            }

            if (GlobalVars.CheckOverlaps)
            {
                Box3[] boxes = new Box3[level.sectors.Count];
                for (int i = 0; i < level.sectors.Count; i++)
                    boxes[i] = Utils.FindBBox(level.sectors[i]);

                for (int i = 0; i < level.sectors.Count; i++)
                {
                    JKSector sec = level.sectors[i];

                    for (int j = i - 1; j >= 0; j--)
                    {
                        JKSector sec1 = level.sectors[j];
                        bool b = false;
                        for(int k=0; k<sec.surfaces.Count; k++)
                            if (sec.surfaces[k].adjoin != null && sec.surfaces[k].adjoin.sector == sec1)
                            {
                                b = true;
                                break;
                            }

                        if (b)
                            continue;

                        if (!Utils.DoBoxesIntersect(boxes[i], boxes[j]))
                            continue;

                        if (Utils.DoSectorsOverlap(sec, sec1))
                        {
                            AddError("Sectors " + i.ToString() + " and " + j.ToString() + " overlap", sec);
                            AddError("Sectors " + i.ToString() + " and " + j.ToString() + " overlap", level.sectors[j]);
                        }
                    }
                }
            }

            for (int i = 0; i < level.things.Count; i++)
            {
                JKThing th = level.things[i];
                if (th.sec == null)
                {
                    AddError("Thing not in sector", th);
                    continue;
                }
                else if (!Utils.IsInSector(th.sec, th.pos))
                    AddError("Thing is not in specified sector", th);
            }

            for (int i = 0; i < level.cogs.Count; i++)
            {
                JKCOG cog = level.cogs[i];
                foreach (COGValue cv in cog.vals)
                    switch (cv.cogType)
                    {
                        case COGType.AI:
                        case COGType.KEY:
                        case COGType.MAT:
                        case COGType.Wav:
                        case COGType.Model:
                            string ext = Path.GetExtension(cv.AsString).ToLower();
                            switch (cv.cogType)
                            {
                                case COGType.AI: if (ext == ".ai") continue; break;
                                case COGType.KEY: if (ext == ".key") continue; break;
                                case COGType.MAT: if (ext == ".mat") continue; break;
                                case COGType.Model: if (ext == ".3do") continue; break;

                                case COGType.Wav:
                                    if (Audio.IsAcceptableExtension(ext))
                                    {
                                        if (ext != ".wav")
                                        {
                                            // HELL .. We could probably load the wavs and even test their sample rate and stuff
                                            //  wouldnt that be nice =)


                                            AddError("Warning, audio file format used is incompatible with JK " + cv.name, cog);
                                        }
                                        continue;
                                    }
                                    break;
                            }
                            AddError("Incorrect file name for parameter " + cv.name, cog);
                            break;
                    }
            }

            if (Errors.SelectedIndex <= 0)
                Errors.SelectedIndex = Errors.Items.Count == 0 ? -1 : 0;
            else
                Errors_SelectedIndexChanged(Errors, new EventArgs());

            if (Errors.Items.Count == 0)
                AddError("No errors", null);
        }

        private void Errors_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (Errors.SelectedIndex == -1)
            {
                BNGoto.Enabled = false;
                return;
            }

            ConsistencyError ce = Errors.SelectedItem as ConsistencyError;
            BNGoto.Enabled = (ce != null);
        }

        private void Errors_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            Errors.SelectedIndex = Errors.IndexFromPoint(e.Location);
            BNGoto_Click(BNGoto, new EventArgs());
        }

        private void BNGoto_Click(object sender, EventArgs e)
        {
            int index = Errors.SelectedIndex;
            if (index < 0)
                return;

            ConsistencyError ce = Errors.SelectedItem as ConsistencyError;
            if (ce == null)
                return;

            switch (ce.etype)
            {
                case ErrorType.Sector:
                    {
                        int sc = level.sectors.IndexOf(ce.itm as JKSector);
                        if (sc == -1)
                        {
                            MessageBox.Show("Sector not found", "Error", MessageBoxButtons.OK);
                            return;
                        }
                        GlobalVars.NedMain.GotoSC(sc);
                    }
                    break;

                case ErrorType.Surface:
                    {
                        int sc = -1;
                        int sf = -1;
                        for (int i = 0; i < level.sectors.Count; i++)
                        {
                            JKSector sec = level.sectors[i];
                            sc = i;
                            sf = sec.surfaces.IndexOf(ce.itm as JKSurface);
                            if (sf != -1)
                                break;
                        }

                        if (sf == -1)
                        {
                            MessageBox.Show("Surface not found", "Error", MessageBoxButtons.OK);
                            return;
                        }

                        GlobalVars.NedMain.GotoSF(sc, sf);
                    }
                    break;

                case ErrorType.Thing:
                    {
                        int th = level.things.IndexOf(ce.itm as JKThing);
                        if (th == -1)
                        {
                            MessageBox.Show("Thing not found", "Error", MessageBoxButtons.OK);
                            return;
                        }

                        GlobalVars.NedMain.GotoTH(th);
                    }
                    break;

                case ErrorType.Cog:
                    {
                        int cg = level.cogs.IndexOf(ce.itm as JKCOG);
                        if (cg == -1)
                        {
                            MessageBox.Show("Cog not found", "Error", MessageBoxButtons.OK);
                            return;
                        }

                        GlobalVars.CogForm.GotoCOG(cg);
                    }
                    break;
            }
        }

        private void BNClose_Click(object sender, EventArgs e)
        {

        }

        private void BNCheck_Click(object sender, EventArgs e)
        {
            Check();
        }

        private void BNCheckRes_Click(object sender, EventArgs e)
        {
            CheckResources();
        }
    }
}
