﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;

namespace Smith.NED
{
    [StructLayout(LayoutKind.Sequential, Pack=1)]
    public class GOBHeader
    {
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 4)]
        public string magic;
        public int index_ofs;
    }

    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public class GOBEntry
    {
        public int offs;
        public int size;
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 13)]
        public string name;
    }

    public class GOBCreator : ContainerCreator
    {
        public List<GOBEntry> pes = new List<GOBEntry>();
        public int centry = 0;

        public GOBCreator(string name)
            : base(name)
        {

        }

        public override void PrepareHeader(List<string> newfiles)
        {
            pes.Clear();
            cf.Position = Marshal.SizeOf(typeof(GOBHeader));
            foreach (string s in newfiles)
            {
                GOBEntry pe = new GOBEntry();
                pe.name = FileOperations.ExtractName(ValidateName(s));
                pes.Add(pe);
            }
        }

        public override void AddFile(BaseFile f)
        {
            GOBEntry pe = pes[centry++];
            pe.offs = cf.Position;
            pe.size = f.Length;
            cf.Write(f.Read(f.Length));
        }

        public override string ValidateName(string name)
        {
            string ext = FileOperations.ExtractExt(name);
            string n = FileOperations.ExtractNameWithoutExtension(name);

            if (ext.Length > 4)
                ext = ext.Substring(0, 4);

            if (n.Length > 8)
                n = n.Substring(0, 8);

            return n + ext;
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (pes.Count > 0)
                {
                    GOBHeader gh = new GOBHeader();
                    gh.magic = "GOB\n";
                    gh.index_ofs = cf.Position;
                    cf.Position = 0;
                    cf.Write(FileOperations.StructToBytes<GOBHeader>(gh));
                    cf.Position = gh.index_ofs;
                    cf.Write(FileOperations.StructToBytes<int>(pes.Count));
                    foreach (GOBEntry pe in pes)
                        cf.Write(FileOperations.StructToBytes<GOBEntry>(pe));
                }
            }

            base.Dispose(disposing);
        }
    }

    public class GOBDirectory : ContainerFile
    {
        public GOBHeader gh;

        public GOBDirectory(string _name)
            : base(_name)
        {

        }

        protected override void Refresh()
        {
            ClearIndex();

            using (BaseFile f = FileOperations.OpenFileRead(Name))
            {
                gh = FileOperations.BytesToStruct<GOBHeader>(f.Read(Marshal.SizeOf(typeof(GOBHeader))));
                if (gh.magic != "GOB\n")
                    throw new Exception("bad gob");

                f.Position = gh.index_ofs;
                int num = FileOperations.BytesToStruct<int>(f.Read(4));
                for (int x = 0; x < num; x++)
                {
                    GOBEntry ge = FileOperations.BytesToStruct<GOBEntry>(f.Read(Marshal.SizeOf(typeof(GOBEntry))));
                    FInfo fi = new FInfo();
                    fi.offs = ge.offs;
                    fi.size = ge.size;
                    Files.Add(ge.name, fi);
                }
            }
        }

        public override ContainerCreator GetContainerCreator(string name)
        {
            return new GOBCreator(name);
        }
    }

    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public class GOB2Header
    {
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
        public byte[] magic;
        public int unknown0;
        public int unknown1;
        public int numentries;
    }

    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public class GOB2Entry
    {
        public int offs;
        public int size;
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 128)]
        public string name;
    }

    public class GOB2Creator : ContainerCreator
    {
        public List<GOB2Entry> pes = new List<GOB2Entry>();
        public int centry = 0;

        public GOB2Creator(string name)
            : base(name)
        {

        }

        public override void PrepareHeader(List<string> newfiles)
        {
            pes.Clear();

            cf.Position = Marshal.SizeOf(typeof(GOB2Header)) + newfiles.Count * Marshal.SizeOf(typeof(GOB2Entry));
            foreach (string file in newfiles)
            {
                GOB2Entry pe = new GOB2Entry();
                pe.name = ValidateName(file);
                pes.Add(pe);
            }

            centry = 0;
        }

        public override void AddFile(BaseFile f)
        {
            GOB2Entry pe = pes[centry++];
            pe.offs = cf.Position;

            if (f == null)
            {
                pe.size = 0;
            }
            else
            {
                pe.size = f.Length;
                cf.Write(f.Read(f.Length));
            }
        }

        public override string ValidateName(string name)
        {
            return name.ToLower();
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                GOB2Header ge = new GOB2Header();
                ge.magic = new byte[4] { (byte)'G', (byte)'O', (byte)'B', (byte)' '};
                ge.unknown0 = 0x14;
                ge.unknown1 = 0x0C;
                ge.numentries = pes.Count;
                cf.Position = 0;
                cf.Write(FileOperations.StructToBytes<GOB2Header>(ge));
                foreach (GOB2Entry pe in pes)
                    cf.Write(FileOperations.StructToBytes<GOB2Entry>(pe));
            }

            base.Dispose(disposing);
        }
    }

    public class GOB2Directory : ContainerFile
    {
        public GOB2Header gh;

        public GOB2Directory(string _name)
            : base(_name)
        {

        }

        protected override void Refresh()
        {
            ClearIndex();

            using (BaseFile f = FileOperations.OpenFileRead(Name))
            {
                gh = FileOperations.BytesToStruct<GOB2Header>(f.Read(Marshal.SizeOf(typeof(GOB2Header))));
                if(gh.magic[0] != (byte)'G' || gh.magic[1] != (byte)'O' || gh.magic[2] != (byte)'B' || gh.magic[3] != (byte)' ')
                    throw new Exception("Not a gob 2 file");

                for(int x=0; x<gh.numentries; x++)
                {
                    GOB2Entry ge = FileOperations.BytesToStruct<GOB2Entry>(f.Read(Marshal.SizeOf(typeof(GOB2Entry))));

                    if (Files.ContainsKey(ge.name))
                    {
                        Utils.PanMessage(Utils.MessageType.Warning, "container specifies same file twice.. using first instance");
                        continue;
                    }

                    FInfo fi = new FInfo();
                    fi.offs = ge.offs;
                    fi.size = ge.size;
                    Files.Add(ge.name, fi);

                    string dir = System.IO.Path.GetDirectoryName(ge.name).TrimEnd(new char[] { '/', '\\' });
                    if (!string.IsNullOrEmpty(dir))
                    {
                        if (!Dirs.Contains(dir))
                            Dirs.Add(dir);
                    }
                }
            }
        }

        public override int GetColWidth(int n)
        {
            if (n == 0)
                return 31;

            return base.GetColWidth(n);
        }

        public override ContainerCreator GetContainerCreator(string name)
        {
            throw new Exception("gob 2 creator not implemented");
        }
    }
}
