﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using System.Text.RegularExpressions;

namespace Smith.NED
{
    public static class FileOperations
    {
        public class WildCardMask
        {
            private List<string> masks = new List<string>();

            public string mask
            {
                get
                {
                    string m = string.Empty;
                    for (int x = 0; x < masks.Count; x++)
                    {
                        m += masks[x];
                        if (x < (masks.Count - 1))
                            m += ";";
                    }
                    return m;
                }

                set
                {
                    masks.Clear();
                    foreach (string m in value.Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries))
                        masks.Add(m.Trim());
                }
            }

            public bool Match(string s)
            {
                foreach (string m in masks)
                    if (WildCardMatch(m, s))
                        return true;
                return false;
            }

            private string[] SplitAndKeepDelimiters(string s, char[] delims)
            {
                List<string> parts = new List<string>();
                for (int pos = 0; pos < s.Length; )
                {
                    int delimpos = s.IndexOfAny(delims, pos);
                    if (delimpos == -1)
                    {
                        // done. add remainder of string, if exist
                        if (pos < s.Length)
                            parts.Add(s.Substring(pos));
                        break;
                    }

                    // add everything we found up to the delim
                    if (delimpos > pos)
                        parts.Add(s.Substring(pos, delimpos - pos));

                    // add the delimiter
                    parts.Add(s.Substring(delimpos, 1));

                    // set position past current delimiter
                    pos = delimpos + 1;
                }
                return parts.ToArray();
            }

            private bool WildCardMatch(string mask, string s)
            {
                string[] quotes = SplitAndKeepDelimiters(mask, new char[] { '?', '*' });
                int ps = 0;
                for (int iquote = 0; iquote < quotes.Length; iquote++)
                {
                    string quote = quotes[iquote];
                    switch (quote)
                    {
                        case "?":
                            if (ps < s.Length - 1)
                                ps++;
                            break;

                        case "*":
                            if (iquote == quotes.Length - 1)
                            {
                                ps = s.Length;
                                continue;
                            }

                            if (quotes[iquote + 1].IndexOfAny(new char[] { '?', '*' }) != -1)
                                continue;

                            quote = quotes[++iquote];

                            {
                                int p = s.IndexOf(quote, ps, StringComparison.InvariantCultureIgnoreCase);
                                if (p == -1)
                                    return false;

                                ps = p + quote.Length;
                            }
                            break;

                        default:
                            if (!s.Substring(ps, Math.Min(s.Length - ps, quote.Length)).Equals(quote, StringComparison.InvariantCultureIgnoreCase))
                                return false;

                            ps += quote.Length;
                            break;
                    }
                }

                return (ps == s.Length);
            }

            public WildCardMask()
            {

            }

            public WildCardMask(string _mask)
            {
                mask = _mask;
            }
        }

        public static string FindProjDirFile(string name)
        {
            string ext = ExtractExt(name).ToLowerInvariant();

            List<string> checkpaths = new List<string>();
            checkpaths.Add(string.Empty);
            if(ext == ".uni")
            {
                checkpaths.Add(@"misc\");
                checkpaths.Add(@"ui\");
            }

            foreach (string subdir in checkpaths)
            {
                string dir = Path.Combine(GlobalVars.ProjectDir, subdir);

                string f = Path.Combine(dir, name);
                if (File.Exists(f))
                    return f;
            }

            return string.Empty;
        }

        public static string ExtractExt(string path)
        {
            return Path.GetExtension(path.Replace('>', '\\'));
        }

        public static string ExtractPath(string path)
        {
            if (string.IsNullOrEmpty(path))
                return path;

            return Path.GetDirectoryName(path.Replace('>', '\\'));
        }

        public static string ExtractName(string path)
        {
            return Path.GetFileName(path.Replace('>', '\\'));
        }

        public static string ExtractNameWithoutExtension(string path)
        {
            return Path.GetFileNameWithoutExtension(path.Replace('>', '\\'));
        }

        public static string ChangeFileExt(string path, string ext)
        {
            int i = path.LastIndexOf('.');
            if (i == -1)
                return path;

            return path.Substring(0, i) + ext;
        }

        public static byte[] StructToBytes<T>(T obj)
        {
            int sz = Marshal.SizeOf(typeof(T));
            IntPtr p = Marshal.AllocHGlobal(sz);
            Marshal.StructureToPtr(obj, p, false);
            byte[] buf = new byte[sz];
            Marshal.Copy(p, buf, 0, sz);
            Marshal.FreeHGlobal(p);
            return buf;
        }

        public static T BytesToStruct<T>(byte[] buf)
        {
            int sz = Marshal.SizeOf(typeof(T));
            IntPtr p = Marshal.AllocHGlobal(sz);
            Marshal.Copy(buf, 0, p, sz);
            T t = (T)Marshal.PtrToStructure(p, typeof(T));
            Marshal.FreeHGlobal(p);
            return t;
        }

        public static string BackUpFile(string file)
        {
            if (!File.Exists(file))
                return string.Empty;

            string cext = ExtractExt(file);
            cext = cext.Insert(1, "~");
            if (cext.Length > 4)
                cext = cext.Substring(0, 4);
            string backfile = ChangeFileExt(file, cext);
            if (File.Exists(backfile))
                File.Delete(backfile);
            File.Copy(file, backfile);
            return backfile;
        }

        public static List<string> ListDirMask(string path, string mask)
        {
            List<string> sl = new List<string>();

            if (Directory.Exists(path))
            {
                string[] masks = mask.Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries);

                foreach(string curmask in masks)
                    foreach (string s in Directory.GetFiles(path, curmask))
                        sl.Add(Path.GetFileName(s));
            }

            return sl;

            // not entirely sure if this is doing anything particularly weird
            // that we have to accommodate for

            /*Procedure ListDirMask(const path,mask:String;sl:TStringList);
var sr:TSearchRec;
    Res:Integer;
    CurMask:array[0..128] of char;
    P,PM:Pchar;
begin
 sl.Clear;
 PM:=PChar(Mask);
Repeat
 p:=StrScan(PM,';');
 if p=nil then p:=StrEnd(PM);
 StrLCopy(CurMask,PM,P-PM);

 Res:=FindFirst(path+curMask,faAnyFile,sr);
 While Res=0 do
 begin
  if (Sr.Attr and (faVolumeID+faDirectory))=0 then sl.Add(sr.Name);
  Res:=FindNext(sr);
 end;
 FindClose(sr);
  if P^=#0 then break else PM:=P+1;
Until False;
end;   */
        }

        public static BaseFile TryProjectFile(string subdir, string name)
        {
            string projectFilePath = Path.Combine(Path.Combine(GlobalVars.ProjectDir, subdir), name);
            if (File.Exists(projectFilePath))
                return OpenFileRead(projectFilePath);

            return null;
        }

        public static BaseFile OpenFileRead(string path)
        {
            int ps = path.IndexOf('>');
            if (ps == -1)
                return DiskFile.OpenRead(path);

            string contName = path.Substring(0, ps);
            string fname = path.Substring(ps + 1).TrimStart(new char[] { '/', '\\' });

            ContainerFile cf;
            if (Containers.ContainsKey(contName))
                cf = Containers[contName];
            else
            {
                if (Containers.Count > 10)
                {
                    foreach(string k in Containers.Keys)
                        if (!Containers[k].Permanent)
                        {
                            Containers[k].Dispose();
                            Containers.Remove(k);
                            break;
                        }
                }

                cf = OpenContainer(contName);
                Containers.Add(contName, cf);
            }

            return cf.OpenFile(fname, 0);
        }

        public static BaseFile OpenFileWrite(string path)
        {
            return DiskFile.OpenWrite(path);
        }

        public static bool IsInGOB(string gob, string subdir, string file)
        {
            return IsInGOB(gob, Path.Combine(subdir, file));
        }

        public static Dictionary<string, ContainerFile> Containers = new Dictionary<string, ContainerFile>(StringComparer.InvariantCultureIgnoreCase);

        public enum ContainerType
        {
            Unknown,
            GOB,
            GOB2,
            GOO,
            WAD,
            LAB,
            LFD,
            NotFound
        }

        public static ContainerType WhatContainer(string file)
        {
            if (!File.Exists(file))
                return ContainerType.NotFound;

            switch (Path.GetExtension(file).ToUpperInvariant())
            {
                case ".WAD": return ContainerType.WAD;
                case ".LAB": return ContainerType.LAB;
                case ".LFD": return ContainerType.LFD;
                case ".GOO": return ContainerType.GOO;
                case ".GOB":
                    using (BaseFile tmp = OpenFileRead(file))
                    {
                        byte[] buf = tmp.Read(4);
                        if (buf[0] == (byte)'G' && buf[1] == (byte)'O' && buf[2] == (byte)'B' && buf[3] == (byte)' ')
                            return ContainerType.GOB2;

                        return ContainerType.GOB;
                    }
                default: return ContainerType.Unknown;
            }
        }

        public static bool IsContainer(string file)
        {
            ContainerType ct = WhatContainer(file);
            return (ct != ContainerType.Unknown && ct != ContainerType.NotFound);
        }

        public static bool IsInContainer(string file)
        {
            return file.IndexOf('>') != -1;
        }

        public static ContainerFile OpenContainer(string gob)
        {
            switch (WhatContainer(gob))
            {
                case ContainerType.GOB:
                    return new GOBDirectory(gob);

                case ContainerType.GOO: goto case ContainerType.GOB2;
                case ContainerType.GOB2:
                    return new GOB2Directory(gob);

                case ContainerType.NotFound:
                    throw new Exception(gob + " not found");

                default:
                    throw new Exception(gob + " is not a container");
            }
        }

        public static ContainerFile OpenGameContainer(string gob)
        {
            ContainerFile cf = null;

            if (Containers.ContainsKey(gob))
                cf = Containers[gob];

            if(cf == null)
            {
                cf = OpenContainer(gob);
                if (cf == null)
                    return null;

                Containers[gob] = cf;
            }

            cf.Permanent = true;
            return cf;
        }

        public class MaskedDirectoryControl
        {
            public ListBox LBDir;
            //public string Mask;
            public ContainerFile Dir = null;

            public MaskedDirectoryControl(ListBox l)
            {
                LBDir = l;
                //Mask = "*.*";
            }

            public void SetDir(ContainerFile d)
            {
                Dir = d;
                if (d == null)
                    return;

                LBDir.ColumnWidth = (int)((double)LBDir.Width / ((double)Dir.GetColWidth(0) * (double)LBDir.Font.Size * 1.0));
            }

            public class ListEntry
            {
                public string Name;
                public FInfo FInfo;

                public ListEntry(string _Name, FInfo _FInfo)
                {
                    Name = _Name;
                    FInfo = _FInfo;
                }

                public override string ToString()
                {
                    if (FInfo == null)
                        return "[" + Name + "]";

                    return Name;
                }
            }

            public void AddFile(string s, FInfo fi)
            {
                LBDir.Items.Add(new ListEntry(s, fi));
            }

            public void SetMask(string mask)
            {
                WildCardMask matcher = new WildCardMask();
                matcher.mask = mask;

                LBDir.BeginUpdate();
                LBDir.Items.Clear();
                foreach (string ts in Dir.ListDirs)
                    AddFile(ts, null);
                Dictionary<string, FInfo> files = Dir.ListFiles;
                foreach (string ts in files.Keys)
                    if (matcher.Match(ts))
                        AddFile(ts, files[ts]);
                LBDir.EndUpdate();
            }
        }

        public static bool IsInGOB(string gob, string file)
        {
            if (string.IsNullOrEmpty(gob))
                return false;

            ContainerFile cf = OpenGameContainer(gob);
            if (cf == null || !cf.FileExists(file))
                return false;

            return true;
        }

        public static BaseFile TryGOBFile(string gob, string subdir, string name)
        {
            if (!IsInGOB(gob, subdir, name))
                return null;

            return OpenFileRead(gob + ">" + subdir + @"\" + name);
        }

        public static BaseFile OpenGameFile(string name)
        {
            BaseFile s = _OpenGameFile(name);
            if (s != null)
                return s;

            // if not found with JED's method, try resource system
            Stream cs = Resource.GetFile(name);
            if(cs != null)
            {
                return new StreamFile(cs);
            }

            string msg = "Can't find file anywhere: " + name;
            Utils.PanMessage(Utils.MessageType.Warning, msg);
            return null;
        }

        private static BaseFile _OpenGameFile(string name)
        {
            string projectFilePath = Path.Combine(GlobalVars.ProjectDir, name);
            if (File.Exists(projectFilePath))
                return OpenFileRead(projectFilePath);

            string projGob = GlobalVars.ProjectGob;

            string fileExt = Path.GetExtension(name).ToUpperInvariant();
            switch (fileExt)
            {
                case ".JKL":
                    return TryGOBFile(projGob, "jkl", name) ??
                        TryGOBFile(projGob, "mission", name);

                case ".MAT":
                    return TryProjectFile("mat", name) ?? TryProjectFile(@"3do\mat", name) ??
                        TryGOBFile(projGob, "mat", name) ?? TryGOBFile(projGob, @"3do\mat", name) ??
                           TryGOBFile(GlobalVars.res2_gob, "mat", name) ?? TryGOBFile(GlobalVars.res2_gob, @"3do\mat", name);

                case ".DAT":
                    return TryProjectFile("misc", name) ??
                        TryGOBFile(projGob, "misc", name) ??
                           TryGOBFile(GlobalVars.res2_gob, "misc", name);

                case ".3DO":
                    return TryProjectFile("3do", name) ??
                        TryGOBFile(projGob, "3do", name) ??
                        TryGOBFile(GlobalVars.res2_gob, "3do", name);

                case ".KEY":
                    return TryProjectFile(@"3do\key", name) ??
                        TryGOBFile(projGob, @"edo\key", name) ??
                        TryGOBFile(GlobalVars.res2_gob, @"3do\key", name);

                case ".AI0": goto case ".AI";
                case ".AI2": goto case ".AI";
                case ".AI":
                    return TryProjectFile(@"misc\ai", name) ??
                        TryGOBFile(projGob, @"misc\ai", name) ??
                        TryGOBFile(GlobalVars.res2_gob, @"misc\ai", name);

                case ".CMP":
                    return TryProjectFile(@"misc\cmp", name) ??
                        TryGOBFile(projGob, @"misc\cmp", name) ??
                        TryGOBFile(GlobalVars.res2_gob, @"misc\cmp", name);

                case ".PAR":
                    return TryProjectFile(@"misc\par", name) ??
                        TryGOBFile(projGob, @"misc\par", name) ??
                        TryGOBFile(GlobalVars.res2_gob, @"misc\par", name);

                case ".PER":
                    return TryProjectFile(@"misc\per", name) ??
                        TryGOBFile(projGob, @"misc\per", name) ??
                        TryGOBFile(GlobalVars.res2_gob, @"misc\per", name);

                case ".PUP":
                    return TryProjectFile(@"misc\pup", name) ??
                        TryGOBFile(projGob, @"misc\pup", name) ??
                        TryGOBFile(GlobalVars.res2_gob, @"misc\pup", name);

                case ".SND":
                    return TryProjectFile(@"misc\snd", name) ??
                        TryGOBFile(projGob, @"misc\snd", name) ??
                        TryGOBFile(GlobalVars.res2_gob, @"misc\snd", name);

                case ".SPR":
                    return TryProjectFile(@"misc\spr", name) ??
                        TryGOBFile(projGob, @"misc\spr", name) ??
                        TryGOBFile(GlobalVars.res2_gob, @"misc\spr", name);

                case ".COG":
                    return TryProjectFile("cog", name) ??
                        TryGOBFile(projGob, "cog", name) ??
                        TryGOBFile(GlobalVars.res2_gob, "cog", name) ??
                        TryGOBFile(GlobalVars.sp_gob, "cog", name) ??
                        TryGOBFile(GlobalVars.mp1_gob, "cog", name) ??
                        TryGOBFile(GlobalVars.mp2_gob, "cog", name) ??
                        TryGOBFile(GlobalVars.mp3_gob, "cog", name);

                default:
                    //case ".WAV":
                    foreach (string checkext in Audio.AcceptableAudioExtensions)
                    {
                        if (!fileExt.Equals(checkext, StringComparison.InvariantCultureIgnoreCase))
                            continue;

                        return TryProjectFile("sound", name) ?? TryProjectFile("voice", name) ?? TryProjectFile("voiceuu", name) ??
                            TryGOBFile(projGob, "sound", name) ??
                               TryGOBFile(GlobalVars.res1_gob, "sound", name) ?? TryGOBFile(GlobalVars.res1_gob, "voice", name) ??
                               (GlobalVars.IsMOTS ? TryGOBFile(GlobalVars.res1_gob, "voiceuu", name) : null);
                    }


                    Utils.PanMessage(Utils.MessageType.Warning, "Unrecognized file extension for file " + name);
                    return null;
            }
        }
    }
}
