﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Smith.NED
{
    public class FileChangeControl
    {
        private long size;
        private DateTime modifiedTime;
        private string name;

        public FileChangeControl(string _name)
        {
            name = _name;

            FileInfo fi = FileInfo;
            size = fi.Length;
            modifiedTime = fi.LastWriteTime;
        }

        private FileInfo FileInfo
        {
            get
            {
                return new FileInfo(name);
            }
        }

        public bool Changed
        {
            get
            {
                FileInfo fi = FileInfo;
                return (fi.Length != size || !fi.LastWriteTime.Equals(modifiedTime));
            }
        }
    }

    public class FInfo
    {
        public int offs;
        public int size;
    }

    public abstract class BaseFile : IDisposable
    {
        public abstract string FullName { get; }
        public abstract string ShortName { get; }
        public abstract byte[] Read(int count);
        public abstract void Write(byte[] buf);
        public abstract int Length { get; }
        public abstract int Position { get; set; }

        public Stream GetStream()
        {
            return new MemoryStream(Read(Length));
        }

        ~BaseFile()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private bool disposed = false;
        protected virtual void Dispose(bool disposing)
        {
            if (disposed)
                return;

            if (disposing)
            {

            }

            disposed = true;
        }
    }

    public class DiskFile : BaseFile
    {
        public FileStream f;

        public DiskFile(FileStream _f)
        {
            f = _f;
        }

        public static DiskFile OpenRead(string path)
        {
            return new DiskFile(File.OpenRead(path));
        }

        public static DiskFile OpenWrite(string path)
        {
            return new DiskFile(File.Create(path));
        }

        public override string FullName
        {
            get { return f.Name; }
        }

        public override string ShortName
        {
            get { return Path.GetFileName(FullName); }
        }

        public override int Length
        {
            get { return (int)f.Length; }
        }

        public override int Position
        {
            get
            {
                return (int)f.Position;
            }
            set
            {
                f.Position = (long)value;
            }
        }

        public override byte[] Read(int count)
        {
            byte[] buf = new byte[count];
            f.Read(buf, 0, count);
            return buf;
        }

        public override void Write(byte[] buf)
        {
            f.Write(buf, 0, buf.Length);
        }

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);

            if (disposing)
            {
                f.Dispose();
            }
        }
    }

    public class StreamFile : BaseFile
    {
        public Stream f;

        public StreamFile(Stream _f)
        {
            f = _f;
        }

        public static DiskFile OpenRead(string path)
        {
            return new DiskFile(File.OpenRead(path));
        }

        public static DiskFile OpenWrite(string path)
        {
            return new DiskFile(File.Create(path));
        }

        public override string FullName
        {
            get { return "stream"/*f.Name*/; }
        }

        public override string ShortName
        {
            get { return "stream"/*Path.GetFileName(FullName)*/; }
        }

        public override int Length
        {
            get { return (int)f.Length; }
        }

        public override int Position
        {
            get
            {
                return (int)f.Position;
            }
            set
            {
                f.Position = (long)value;
            }
        }

        public override byte[] Read(int count)
        {
            byte[] buf = new byte[count];
            f.Read(buf, 0, count);
            return buf;
        }

        public override void Write(byte[] buf)
        {
            f.Write(buf, 0, buf.Length);
        }

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);

            if (disposing)
            {
                f.Dispose();
            }
        }
    }

    public abstract class ContainerCreator : IDisposable
    {
        public BaseFile cf;
        public abstract void PrepareHeader(List<string> newfiles);
        public abstract void AddFile(BaseFile f);
        public abstract string ValidateName(string name);

        public ContainerCreator(string name)
        {
            cf = FileOperations.OpenFileWrite(name);
        }

        ~ContainerCreator()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private bool disposed = false;
        protected virtual void Dispose(bool disposing)
        {
            if (disposed)
                return;

            if (disposing)
            {
                cf.Dispose();
            }

            disposed = true;
        }
    }

    public class SubFile : BaseFile
    {
        public int off;
        public int size;
        public BaseFile f;
        public string Name;

        public SubFile(string fn, string name, int apos, int asize)
        {
            Name = fn + ">" + name;
            off = apos;
            size = asize;
            f = FileOperations.OpenFileRead(fn);
            f.Position = off;
        }

        public override string FullName
        {
            get { return Name; }
        }

        public override string ShortName
        {
            get { return FileOperations.ExtractName(Name); }
        }

        public override byte[] Read(int count)
        {
            return f.Read(count);
        }

        public override void Write(byte[] buf)
        {
            throw new Exception("subfile doesnt support writing");
        }

        public override int Length
        {
            get { return size; }
        }

        public override int Position
        {
            get
            {
                return f.Position - off;
            }
            set
            {
                f.Position = value + off;
            }
        }

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);

            if (disposing)
            {
                f.Dispose();
            }
        }
    }

    public abstract class ContainerFile : IDisposable
    {
        public bool Permanent = false;
        public string Name;
        public Dictionary<string, FInfo> Files = new Dictionary<string, FInfo>(StringComparer.InvariantCultureIgnoreCase);
        public List<string> Dirs = new List<string>();
        public Dictionary<string, FInfo> tmpflist = new Dictionary<string, FInfo>(StringComparer.InvariantCultureIgnoreCase);
        public List<string> tmpdlist = new List<string>();
        public BaseFile f;
        public FileChangeControl fc;
        public bool HasDirs = false;
        public string CurDir = string.Empty;

        public ContainerFile(string _name)
        {
            Name = _name;
            fc = new FileChangeControl(Name);
            Refresh();
        }

        public virtual string GetColName(int n)
        {
            switch (n)
            {
                case 0:
                    return "Name";

                case 1:
                    return "Size";

                default:
                    return string.Empty;
            }
        }

        public virtual int GetColWidth(int n)
        {
            switch (n)
            {
                case 0:
                    return 13;
                default:
                    return 10;
            }
        }

        public virtual int ColumnCount
        {
            get
            {
                return 2;
            }
        }

        public virtual string FullName
        {
            get
            {
                return Name;
            }
        }

        public virtual string ShortName
        {
            get
            {
                return Path.GetFileName(Name);
            }
        }

        public virtual BaseFile OpenFile(string fname, int mode)
        {
            if (fc.Changed)
                Refresh();

            if (mode != 0)
                throw new Exception("Cannot write file inside container");

            if (!Files.ContainsKey(fname))
                throw new Exception("File not found " + fname);

            FInfo fi = Files[fname];

            return new SubFile(Name, fname, fi.offs, fi.size);
        }

        /*public virtual BaseFile OpenFileByFI(FInfo fi)
        {

        }*/

        public virtual bool ChDir(string d)
        {
            if (!string.IsNullOrEmpty(d) && !Dirs.Contains(d))
                return false;

            CurDir = d;
            return true;
        }

        public virtual void RenameFile(string fname, string newname)
        {

        }

        public bool FileExists(string fname)
        {
            return Files.ContainsKey(fname);
        }

        public virtual Dictionary<string, FInfo> ListFiles
        {
            get
            {
                if (fc.Changed)
                    Refresh();

                tmpflist.Clear();
                string dir = CurDir;
                if(!string.IsNullOrEmpty(dir))
                    dir += @"\";
                foreach(string file in Files.Keys)
                {
                    if(!file.StartsWith(dir, StringComparison.InvariantCultureIgnoreCase))
                        continue;

                    // wtf is this
                    if(!IsLastInPath(file, dir.Length))
                        continue;

                    tmpflist.Add(GetDirAt(file, dir.Length), Files[file]);
                }

                return tmpflist;
            }
        }

        public bool IsLastInPath(string path, int p)
        {
            return (path.IndexOf('\\', p) == -1);
        }

        public string GetDirAt(string path, int p)
        {
            int sp = path.IndexOf('\\', p);
            if(sp == -1)
                return path.Substring(p);
            else
                return path.Substring(p, sp-p);
        }

        public virtual List<string> ListDirs
        {
            get
            {
                if (fc.Changed)
                    Refresh();

                tmpdlist.Clear();
                string dir = CurDir;
                if (!string.IsNullOrEmpty(dir))
                    dir += "\\";
                foreach (string dname in Dirs)
                {
                    if (!dname.StartsWith(dir, StringComparison.InvariantCultureIgnoreCase))
                        continue;

                    if (!IsLastInPath(dname, dir.Length))
                        continue;

                    tmpdlist.Add(GetDirAt(dname, dir.Length));
                }

                return tmpdlist;
            }
        }

        public virtual FInfo GetFileInfo(string fname)
        {
            if (!Files.ContainsKey(fname))
                return null;

            return Files[fname];
        }

        public virtual bool AddFiles(List<string> addFiles)
        {
            string newName = FileOperations.ChangeFileExt(Name, ".tmp");
            using (ContainerCreator ccreator = GetContainerCreator(newName))
            {
                Dictionary<string, FInfo> newFiles = new Dictionary<string, FInfo>(Files, StringComparer.InvariantCultureIgnoreCase);
                foreach (string addFile in addFiles)
                {
                    string fname = FileOperations.ExtractName(addFile);
                    if (Files.ContainsKey(fname))
                        newFiles.Remove(fname);

                    newFiles.Add(addFile, null);
                }

                ccreator.PrepareHeader(new List<string>(newFiles.Keys));

                foreach (string file in newFiles.Keys)
                {
                    BaseFile bf;

                    if (newFiles[file] == null)
                        bf = OpenFile(file, 0); // from container.. this doesnt make sense;  shouldnt we be getting this externally if FInfo is bad?
                    else
                        bf = FileOperations.OpenFileRead(file);

                    ccreator.AddFile(bf);
                }
            }

            FileOperations.BackUpFile(Name);
            File.Delete(Name);
            File.Move(newName, Name);
            Refresh();
            return true;
        }

        public virtual bool DeleteFiles(List<string> files)
        {
            return false;
        }

        public virtual bool AddFile(string fname)
        {
            List<string> f = new List<string>();
            f.Add(fname);
            return AddFiles(f);
        }

        public bool DeleteFile(string fname)
        {
            List<string> f = new List<string>();
            f.Add(fname);
            return DeleteFiles(f);
        }

        public abstract ContainerCreator GetContainerCreator(string name);

        protected virtual void ClearIndex()
        {
            Files.Clear();
        }

        protected abstract void Refresh();

        ~ContainerFile()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private bool disposed = false;
        protected virtual void Dispose(bool disposing)
        {
            if (disposed)
                return;

            if (disposing)
            {

            }

            disposed = true;
        }
    }
}
