﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Smith.NED
{
    public static class ListRes
    {
        public static void LoadJKLLists(string name, ref List<string> checklist)
        {
            Stream stream = null;
            using (BaseFile bf = FileOperations.OpenFileRead(name))
                stream = new MemoryStream(bf.Read(bf.Length));

            Tokenizer t = new Tokenizer(stream);

            foreach (string sec in new string[] { "sounds", "materials", "models", "aiclass", "sprites", "keyframes", "animclass", "soundclass", "cogscripts" })
                if (t.SeekToLine("section: " + sec))
                {
                    t.EatLine();
                    for (; ; )
                    {
                        string ln = t.ReadLine();
                        if (string.IsNullOrEmpty(ln) || ln.Equals("end", StringComparison.InvariantCultureIgnoreCase))
                            break;

                        // sounds are not prefixed with index; everything else is. some sections have more data after filename but ignore
                        string[] parts = ln.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                        string file;
                        if (parts.Length == 1)
                            file = parts[0];
                        else
                            file = parts[1];

                        if (!checklist.Contains(file, StringComparer.InvariantCultureIgnoreCase))
                            checklist.Add(file);
                    }
                }
        }
    }

    public class ModelsDatFile
    {
        public List<string> snds = new List<string>();
        public List<string> a3dos = new List<string>();

        public void Load(string fname)
        {
            MemoryStream stream;
            try
            {
                using (BaseFile bf = FileOperations.OpenGameFile(fname))
                    stream = new MemoryStream(bf.Read(bf.Length));
            }
            catch
            {
                return;
            }

            Tokenizer t = new Tokenizer(stream);

            t.EatToken();
            int num = t.ReadInt();
            for (int x = 0; x < num; x++)
            {
                t.EatToken();
                a3dos.Add(t.ReadToken());
                snds.Add(t.ReadToken());
            }
        }
    }

    public class ItemsDatFile
    {
        public List<string> COGs = new List<string>();

        public void Load(string fname)
        {
            COGs.Clear();

            MemoryStream stream;
            try
            {
                using (BaseFile bf = FileOperations.OpenGameFile(fname))
                    stream = new MemoryStream(bf.Read(bf.Length));
            }
            catch
            {
                return;
            }

            Tokenizer t = new Tokenizer(stream);

            while (!t.EndOfStream)
            {
                string ln = t.ReadLine();
                if (string.IsNullOrEmpty(ln))
                    break;

                int p = ln.IndexOf("cog=", StringComparison.InvariantCultureIgnoreCase);
                if (p != -1)
                    COGs.Add(ln.Substring(p+4));
            }
        }
    }

    public class PUPFile
    {
        public List<string> keys = new List<string>();

        public void Load(string fname)
        {
            keys.Clear();

            MemoryStream stream;
            try
            {
                using (BaseFile bf = FileOperations.OpenGameFile(fname))
                    stream = new MemoryStream(bf.Read(bf.Length));
            }
            catch
            {
                return;
            }

            Tokenizer t = new Tokenizer(stream);

            while (!t.EndOfStream)
            {
                string ln = t.ReadLine();
                if (string.IsNullOrEmpty(ln))
                    break;

                int i = ln.IndexOfAny(new char[] { ' ', '\t' });
                if (i == -1)
                    continue;

                string w = ln.Substring(0, i);

                if (FileOperations.ExtractExt(w).Equals(".key", StringComparison.InvariantCultureIgnoreCase))
                    keys.Add(w);
            }
        }
    }

    public class SNDFile
    {
        public List<string> wavs = new List<string>();

        public void Load(string fname)
        {
            wavs.Clear();

            MemoryStream stream;
            try
            {
                using (BaseFile bf = FileOperations.OpenGameFile(fname))
                    stream = new MemoryStream(bf.Read(bf.Length));
            }
            catch
            {
                return;
            }

            Tokenizer t = new Tokenizer(stream);

            while (!t.EndOfStream)
            {
                string ln = t.ReadLine();
                if (string.IsNullOrEmpty(ln))
                    break;

                string[] parts = ln.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                if (parts.Length >= 2)
                    wavs.Add(parts[1]);
            }
        }
    }

    public class T3DOFile
    {
        public List<string> mats = new List<string>();

        public void Load(string fname)
        {
            mats.Clear();

            MemoryStream stream;
            try
            {
                using (BaseFile bf = FileOperations.OpenGameFile(fname))
                    stream = new MemoryStream(bf.Read(bf.Length));
            }
            catch
            {
                return;
            }

            Tokenizer t = new Tokenizer(stream);

            if (t.SeekToLine("section: modelresource"))
            {
                t.EatToken();
                int numMats = t.ReadInt();
                for (int x = 0; x < numMats; x++)
                {
                    t.EatToken();
                    string mat = t.ReadToken();
                    mats.Add(mat);
                }
            }
        }
    }

    public class CogFile
    {
        public COGValues vals = new COGValues();

        public void Load(string fname)
        {
            LoadFile(fname, true);
        }

        public void LoadNoLocals(string fname)
        {
            LoadFile(fname, false);
        }

        public int Count
        {
            get
            {
                return vals.Count;
            }
        }

        public COGValue this[int i]
        {
            get
            {
                return vals[i];
            }
        }

        private void Clear()
        {
            vals.Clear();
        }

        private string GetLine(Tokenizer t)
        {
            while (!t.EndOfStream)
            {
                string ln = t.ReadLine();
                if (string.IsNullOrEmpty(ln))
                    return null;

                int i = ln.IndexOf("//");
                if (i != -1)
                    ln = ln.Substring(0, i);

                ln = ln.Trim(new char[] { ' ', '\t' });

                if (string.IsNullOrEmpty(ln))
                    continue;

                return ln;
            }

            return null;
        }

        private void LoadFile(string fname, bool locals)
        {
            vals.Clear();

            MemoryStream stream = null;
            using (BaseFile bf = FileOperations.OpenGameFile(fname))
                if(bf != null)
                    stream = new MemoryStream(bf.Read(bf.Length));

            if (stream == null)
                return;

            Tokenizer t = new Tokenizer(stream);

            for (; ; )
            {
                string ln = GetLine(t);
                if (string.IsNullOrEmpty(ln))
                    break;

                if (ln.Equals("symbols", StringComparison.InvariantCultureIgnoreCase))
                {
                    for(;;)
                    {
                        ln = GetLine(t);
                        if(string.IsNullOrEmpty(ln))
                            break;

                        if (ln.Equals("end", StringComparison.InvariantCultureIgnoreCase))
                            break;

                        COGValue cv = COGValue.FromString(ln);
                        if (cv == null)
                        {
                            Log.Error($"Couldnt parse symbol string into cogvalue: {ln}");
                            continue;
                        }

                        if (!locals && (cv.local || cv.cogType == COGType.Msg))
                            continue;

                        vals.Add(cv);
                    }

                    return;
                }
            }
        }
    }
}
