﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.IO;

namespace Smith.NED
{
    public partial class MaterialAttributeEditor : GUI.BaseForm
    {
        public class MatEntry
        {
            public string Name;
            public string Colormap = "dflt.cmp";
            public string Class = "default";

            public void ToXML(XmlElement root)
            {
                root.SetAttribute("FNM", Name);
                root.SetAttribute("CMP", Colormap);
                root.SetAttribute("CLS", Class);
            }

            public static MatEntry FromXML(XmlNode root)
            {
                XmlAttribute atr;

                MatEntry me = new MatEntry();

                atr = root.Attributes["FNM"];
                if (atr != null)
                    me.Name = Path.GetFileName(atr.InnerText).ToLowerInvariant();

                atr = root.Attributes["CMP"];
                if (atr != null)
                    me.Colormap = atr.InnerText;

                atr = root.Attributes["CLS"];
                if (atr != null)
                    me.Class = atr.InnerText;

                return me;
            }
        }

        public class MatAttributes
        {
            public string Name;
            public double DirectOcclusion = 0.5;
            public double ReverbOcclusion = 0.5;

            public void ToXML(XmlElement root)
            {
                root.SetAttribute("CLS", Name);
                root.SetAttribute("DO", DirectOcclusion.ToString());
                root.SetAttribute("RO", ReverbOcclusion.ToString());
            }

            public static MatAttributes FromXML(XmlNode root)
            {
                XmlAttribute atr;

                MatAttributes ma = new MatAttributes();

                atr = root.Attributes["CLS"];
                if (atr != null)
                    ma.Name = atr.InnerText;

                atr = root.Attributes["DO"];
                if (atr != null)
                    ma.DirectOcclusion = double.Parse(atr.InnerText);

                atr = root.Attributes["RO"];
                if (atr != null)
                    ma.ReverbOcclusion = double.Parse(atr.InnerText);

                return ma;
            }
        }

        public class MaterialDB
        {
            public List<MatAttributes> Classes = new List<MatAttributes>();
            public List<MatEntry> Mats = new List<MatEntry>();

            public MatAttributes GetClass(string name)
            {
                foreach (MatAttributes ma in Classes)
                    if (ma.Name.Equals(name, StringComparison.InvariantCultureIgnoreCase))
                        return ma;

                return null;
            }

            public MatAttributes GetOrCreateClass(string name)
            {
                MatAttributes ma = GetClass(name);
                if (ma == null)
                {
                    ma = new MatAttributes();

                    ma.Name = name;

                    Classes.Add(ma);
                }

                return ma;
            }

            public MatEntry GetMat(string name)
            {
                if (string.IsNullOrEmpty(name))
                    return null;

                name = Path.GetFileName(name).ToLowerInvariant();

                foreach (MatEntry me in Mats)
                    if (me.Name.Equals(name, StringComparison.InvariantCultureIgnoreCase))
                        return me;

                return null;
            }

            public MatEntry GetOrCreateMat(string name)
            {
                name = Path.GetFileName(name).ToLowerInvariant();

                MatEntry me = GetMat(name);
                if (me == null)
                {
                    me = new MatEntry();

                    me.Name = name;

                    Mats.Add(me);
                }

                return me;
            }

            public void Clear()
            {
                Classes.Clear();
                Mats.Clear();

                MatAttributes ma = new MatAttributes();
                ma.Name = "default";
                Classes.Add(ma);
            }

            public void Save(string filename)
            {
                XmlDocument doc = new XmlDocument();
                XmlElement root = doc.CreateElement("MaterialDB");
                doc.AppendChild(root);

                ToXML(root);

                doc.Save(filename);
            }

            public void Load(string filename)
            {
                string dummy;
                Load(filename, out dummy);
            }

            public void Load(string filename, out string localFilePath)
            {
                Load(Jeddata.GetFile(filename, out localFilePath));
            }

            public void Load(Stream stream)
            {
                XmlDocument doc = new XmlDocument();
                doc.Load(stream);

                FromXML(doc.SelectSingleNode("MaterialDB"));
            }

            public void ToXML(XmlElement root)
            {
                XmlDocument doc = root.OwnerDocument;
                XmlElement elem;

                /*elem = doc.CreateElement("Duration");
                elem.InnerText = Duration.ToString();
                root.AppendChild(elem);

                elem = doc.CreateElement("TimeSnap");
                elem.InnerText = TimeSnap.ToString();
                root.AppendChild(elem);

                elem = doc.CreateElement("SyncSound");
                elem.InnerText = SyncSound;
                root.AppendChild(elem);*/

                foreach (MatAttributes ma in Classes)
                {
                    elem = doc.CreateElement("ATR");
                    ma.ToXML(elem);
                    root.AppendChild(elem);
                }

                foreach (MatEntry me in Mats)
                {
                    elem = doc.CreateElement("MAT");
                    me.ToXML(elem);
                    root.AppendChild(elem);
                }
            }

            public void FromXML(XmlNode root)
            {
                //XmlNode node;

                Clear();

                /*UpdatingControls = true;

                node = root.SelectSingleNode("Duration");
                if (node != null)
                    Duration = double.Parse(node.InnerText);

                node = root.SelectSingleNode("TimeSnap");
                if (node != null)
                    TimeSnap = double.Parse(node.InnerText);

                node = root.SelectSingleNode("SyncSound");
                if (node != null)
                    SyncSound = node.InnerText;*/

                XmlNodeList atrs = root.SelectNodes("ATR");
                foreach (XmlNode atr in atrs)
                {
                    MatAttributes ma = MatAttributes.FromXML(atr);
                    if (ma.Name.Equals("default", StringComparison.InvariantCultureIgnoreCase))
                        Classes[0] = ma;
                    else
                        Classes.Add(ma);
                }

                XmlNodeList mats = root.SelectNodes("MAT");
                foreach (XmlNode mat in mats)
                    Mats.Add(MatEntry.FromXML(mat));

                //UpdatingControls = false;
            }
        }

        public MaterialAttributeEditor()
        {
            InitializeComponent();
        }

        private void matlist_DrawItem(object sender, DrawItemEventArgs e)
        {
            e.DrawBackground();

            if (e.Index < 0 || e.Index >= matlist.Items.Count)
                return;

            MatEntry me = matlist.Items[e.Index] as MatEntry;
            if(me == null)
                return;

            Box2 rc = new Box2(e.Bounds);

            Box2 iconrc = rc;
            iconrc.Left = iconrc.Right - iconrc.Height;

            Box2 txtrc = rc;
            txtrc.Right = iconrc.Left;

            StringFormat sf = new StringFormat();
            sf.Alignment = StringAlignment.Near;
            sf.LineAlignment = StringAlignment.Center;
            e.Graphics.DrawString(me.Name + ", " + me.Colormap + ", " + me.Class, e.Font, new SolidBrush(e.ForeColor), txtrc, sf);

            Material mat = Resource.GetMaterial(me.Name);
            if (mat == null)
                mat = Resource.GetMaterial("dflt.mat");

            if (mat != null)
            {
                Bitmap bmp, embmp;
                mat.GenerateBitmap(out bmp, out embmp, Resource.GetColormap(me.Colormap));

                e.Graphics.FillRectangle(new SolidBrush(Color.Black), (Rectangle)iconrc);
                System.Drawing.Drawing2D.InterpolationMode oldinterp = e.Graphics.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
                double wrat, hrat;
                if (bmp.Width > bmp.Height)
                {
                    wrat = 1.0;
                    hrat = (double)bmp.Height / (double)bmp.Width;
                }
                else
                {
                    wrat = (double)bmp.Width / (double)bmp.Height;
                    hrat = 1.0;
                }
                Box2 imgrc = Box2.Around(iconrc.Center, iconrc.Size.MultComps(wrat, hrat));
                e.Graphics.FillRectangle(new SolidBrush(Color.FromArgb(255, 0, 255)), (Rectangle)imgrc);
                e.Graphics.DrawImage(bmp, (Rectangle)imgrc, new Rectangle(0, 0, bmp.Width, bmp.Height), GraphicsUnit.Pixel);
                e.Graphics.InterpolationMode = oldinterp;

                if (bmp != null)
                    bmp.Dispose();

                if (embmp != null)
                    embmp.Dispose();
            }
            
            e.DrawFocusRectangle();
        }

        MaterialDB db = new MaterialDB();

        string dbFilePath;

        private void MaterialAttributeEditor_Load(object sender, EventArgs e)
        {
            // try to load from file first
            db.Clear();
            db.Load("materials.db", out dbFilePath);

            if (string.IsNullOrEmpty(dbFilePath))
            {
                if (!string.IsNullOrEmpty(GlobalVars.ProjectDir))
                    dbFilePath = Path.Combine(GlobalVars.ProjectDir, dbFilePath);
                else
                    if (!string.IsNullOrEmpty(GlobalVars.BaseDir))
                        dbFilePath = Path.Combine(Path.Combine(GlobalVars.BaseDir, "jeddata"), "materials.db");
            }

            label5.Text = dbFilePath;

            List<string> cmpnames = new List<string>();

            // create entries for any mat we dont have yet
            foreach (GOB gob in Resource.GOBs)
                foreach (string file in gob)
                {
                    if (file.StartsWith(@"mat\", StringComparison.InvariantCultureIgnoreCase))  // not 3do mats?
                    {
                        MatEntry me = db.GetOrCreateMat(file);
                    }

                    if (file.EndsWith(".cmp", StringComparison.InvariantCultureIgnoreCase))
                    {
                        string cmpname = Path.GetFileName(file).ToLower();
                        if (!cmpnames.Contains(cmpname))
                            cmpnames.Add(cmpname);
                    }
                }


            foreach (MatEntry me in db.Mats)
            {
                string cmpname = Path.GetFileName(me.Colormap).ToLower();
                if (!cmpnames.Contains(cmpname))
                    cmpnames.Add(cmpname);
            }

            updatingmat = true;
            updatingclass = true;

            cbColormap.BeginUpdate();
            cbColormap.Items.Clear();
            cbColormap.Items.AddRange(cmpnames.ToArray());
            cbColormap.EndUpdate();

            cbClass.BeginUpdate();
            cbClass.Items.Clear();
            foreach (MatAttributes ma in db.Classes)
                cbClass.Items.Add(ma.Name);
            cbClass.EndUpdate();
            

            matlist.BeginUpdate();
            matlist.Items.Clear();
            foreach (MatEntry me in db.Mats)
                matlist.Items.Add(me);
            matlist.EndUpdate();

            updatingmat = false;
            updatingclass = false;
        }

        private void MaterialAttributeEditor_FormClosing(object sender, FormClosingEventArgs e)
        {
            if(string.IsNullOrEmpty(dbFilePath))
                MessageBox.Show("Couldn't figure out a good path to save materials.db;  save the level or set your JK directory");
            else
            {
                if (MessageBox.Show("Do you want to save\n" + dbFilePath, "Create/Update File", MessageBoxButtons.OKCancel) == System.Windows.Forms.DialogResult.OK)
                {
                    string dir = Path.GetDirectoryName(dbFilePath);
                    if (!Directory.Exists(dir))
                        Directory.CreateDirectory(dir);

                    db.Save(dbFilePath);
                }
            }
        }

        private void matlist_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (matlist.SelectedIndices.Count == 0)
            {
                matgroup.Enabled = false;
                return;
            }

            matgroup.Enabled = true;

            string cmp = null;
            string cls = null;
            foreach(MatEntry me in matlist.SelectedItems)
            {
                if (cmp == null)
                {
                    // grab from first
                    cmp = me.Colormap;
                    cls = me.Class;
                }
                else
                {
                    if (!cmp.Equals(me.Colormap, StringComparison.InvariantCultureIgnoreCase))
                        cmp = string.Empty;

                    if (!cls.Equals(me.Class, StringComparison.InvariantCultureIgnoreCase))
                        cls = string.Empty;
                }
            }

            updatingmat = true;
            cbColormap.Text = cmp.ToLower();
            cbClass.Text = cls.ToLower();
            updatingmat = false;
        }

        bool updatingmat = false;
        bool updatingclass = false;

        private void cbColormap_TextChanged(object sender, EventArgs e)
        {
            if (updatingmat)
                return;

            foreach (MatEntry me in matlist.SelectedItems)
                me.Colormap = cbColormap.Text;

            matlist.Refresh();
        }

        private void cbClass_SelectedIndexChanged(object sender, EventArgs e)
        {
            MatAttributes ma = db.GetClass(cbClass.Text);
            if (ma == null)
            {
                updatingclass = true;
                directo.Text = string.Empty;
                reverbo.Text = string.Empty;
                updatingclass = false;
                return;
            }

            updatingclass = true;
            directo.Text = ma.DirectOcclusion.ToString();
            reverbo.Text = ma.ReverbOcclusion.ToString();
            updatingclass = false;
            

            if (updatingmat)
                return;

            foreach (MatEntry me in matlist.SelectedItems)
                me.Class = cbClass.Text;

            matlist.Refresh();
        }

        private void bnClassDelete_Click(object sender, EventArgs e)
        {
            if (cbClass.SelectedIndex == 0)
            {
                MessageBox.Show("Can't delete default");
                return;
            }

            MatAttributes ma = db.GetClass(cbClass.Text);
            if (ma == null)
                return;

            // select all entries with this class
            matlist.SelectedItems.Clear();
            foreach (MatEntry me in matlist.Items)
                if (me.Class.Equals(ma.Name, StringComparison.InvariantCultureIgnoreCase))
                    matlist.SelectedItems.Add(me);

            if (matlist.SelectedItems.Count > 0)
            {
                if (MessageBox.Show("Any materials assigned to this class will be reset to default.", "Warning", MessageBoxButtons.OKCancel) != System.Windows.Forms.DialogResult.OK)
                    return;
            }

            object oldobj = cbClass.SelectedItem;
            cbClass.SelectedIndex = 0;  // will actually update all materials
            cbClass.Items.Remove(oldobj);

            db.Classes.Remove(ma);
        }

        private void bnClassAdd_Click(object sender, EventArgs e)
        {
            MatAttributes prevma = db.GetClass(cbClass.Text);

            string s = string.Empty;
            Utils.InputBox("New Class", "Please enter new class name", ref s);
            if (string.IsNullOrEmpty(s))
            {
                MessageBox.Show("Classname cannot be empty");
                return;
            }

            if (db.GetClass(s) != null)
            {
                MessageBox.Show("Classname already exists");
                return;
            }

            MatAttributes ma = new MatAttributes();
            ma.Name = s;
            if (prevma != null)
            {
                ma.DirectOcclusion = prevma.DirectOcclusion;
                ma.ReverbOcclusion = prevma.ReverbOcclusion;
            }
            db.Classes.Add(ma);
            cbClass.SelectedIndex = cbClass.Items.Add(ma.Name);
        }

        private void directo_TextChanged(object sender, EventArgs e)
        {
            if (updatingclass)
                return;

            MatAttributes ma = db.GetClass(cbClass.Text);
            if (ma == null)
                return;

            double d;
            if (!double.TryParse(directo.Text, out d))
                return;

            ma.DirectOcclusion = d;
        }

        private void reverbo_TextChanged(object sender, EventArgs e)
        {
            if (updatingclass)
                return;

            MatAttributes ma = db.GetClass(cbClass.Text);
            if (ma == null)
                return;

            double d;
            if (!double.TryParse(reverbo.Text, out d))
                return;

            ma.ReverbOcclusion = d;
        }
    }
}
