﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;

namespace Smith.NED
{
    public unsafe class PreviewRenderer : RendererBase
    {
        public const double FrontPlane = 0.05;
        public const double BackPlane = 5000.0;

        public enum PickType
        {
            Nothing,
            Surface,
            Thing
        }

        public class T3DPTexture
        {
            public string cmpname;
            public int width;
            public int height;
        }

        public class TOGLTexture : T3DPTexture
        {
            public uint ObjIdx;
        }

        public class TVXdets
        {
            public Vec3 v;
            public Vec2 tv;
            public Color4 clr;

            public static TVXdets Quick(Vec3 _v)
            {
                TVXdets tv = new TVXdets();
                tv.v = _v;
                tv.tv = Vec2.Zero;
                tv.clr = Color4.white();
                return tv;
            }
        }

        public class T3DPSurf
        {
            public Vec3 normal;
            public List<TVXdets> vxds = new List<TVXdets>();
            public TOGLTexture tx = null;
        }

        public class T3DPMesh
        {
            public List<T3DPSurf> surfs = new List<T3DPSurf>();
            public Vec3 center;
            public double radius;

            public void CalculateSphere()
            {
                center = Vec3.Zero;
                radius = 0.0;

                Box3 bbox = Box3.Expandable;
                foreach (T3DPSurf surf in surfs)
                    foreach (TVXdets vxd in surf.vxds)
                        bbox.Expand(vxd.v);

                if (!bbox.HasArea)
                {
                    // if bad box but valid surfaces, try to get from first surface and call it a day
                    if (surfs.Count >= 1 && surfs[0].vxds.Count >= 2)
                    {
                        center = Vec3.Zero;
                        foreach (TVXdets vxd in surfs[0].vxds)
                            center += vxd.v;
                        center /= (double)surfs[0].vxds.Count;

                        radius = 0.0;
                        foreach (TVXdets vxd in surfs[0].vxds)
                            radius = Math.Max(radius, (vxd.v - center).Magnitude);
                    }

                    return;
                }

                center = bbox.Center;

                foreach (T3DPSurf surf in surfs)
                    foreach (TVXdets vxd in surf.vxds)
                        radius = Math.Max(radius, (vxd.v - center).Magnitude);
            }
        }

        public class TMeshes : List<T3DPMesh>
        {

        }

        public Vec3 CamPos = Vec3.Zero;
        public Orient CamOrient = Orient.Zero;
        public double gamma = 1.0;
        public string thing_cmp = "dflt.cmp";

        public Dictionary<string, TOGLTexture> TXList = new Dictionary<string, TOGLTexture>(StringComparer.InvariantCultureIgnoreCase);
        public Dictionary<string, Colormap> CmpList = new Dictionary<string, Colormap>(StringComparer.InvariantCultureIgnoreCase);
        public JKSectors sectors = new JKSectors();
        public JKThings things = new JKThings();

        public TMeshes sclist = new TMeshes();
        public TMeshes thlist = new TMeshes();

        public double vwidth = 0.0;
        public double vheight = 0.0;

        public override void SetupViewport(System.Drawing.Rectangle windowClient, Box2 displayRCMM)
        {
            base.SetupViewport(windowClient, displayRCMM);

            vwidth = (double)windowClient.Width;
            vheight = (double)windowClient.Height;
        }

        public const double Near = 0.05;

        public void SetMatrix()
        {
            using (Context)
            {
                gl.MatrixMode(gl.PROJECTION);
                gl.LoadIdentity();

                Vec2 scrnsz = new Vec2(vwidth, vheight);//320.0, 240.0);

                double a = 2.0 * MathLib.ToDeg(Math.Atan(scrnsz.y / scrnsz.x));
                gl.uPerspective(a, scrnsz.x/ scrnsz.y, Near, 5000.0);
                gl.Rotated(-90.0, 1.0, 0.0, 0.0);

                // this is all terrible-  shouldnt we be setting camera in modelview?
                gl.Rotated(CamOrient.pitch, 1.0, 0.0, 0.0);
                gl.Rotated(CamOrient.yaw, 0.0, 0.0, 1.0);
                gl.Translate(-CamPos);

                gl.MatrixMode(gl.MODELVIEW);
                gl.LoadIdentity();
                gl.MultMatrix(WorldOrient.Matrix.Inverse);  
            }
        }

        public void SetViewToThing(JKThing th)
        {
            Box3 bbox = th.model.GetBBox();

            Vec3 tcen = bbox.Center;
            double trad = (tcen - bbox.LLB).Magnitude;

            CamPos = tcen - new Vec3(0.0, Near + trad*1.0, 0.0);
        }

        public Orient WorldOrient = Orient.Zero;

        public void SetPCHYAW(double pch, double yaw)
        {
            if (things.Count == 0)
                return;

            //JKModel.Mesh mesh = thMeshes[0];
            JKThing th = things[0];

            if (th.model == null)
                return;

            Box3 box = th.model.GetBBox();

            //throw new Exception("SetPCHYAW not implemented");
            WorldOrient.pitch = pch;
            WorldOrient.yaw = yaw;
        }

        public void GetWorldLine(Vec2 pt, out Vec3 near, out Vec3 far)
        {
            using (Context)
            {
                SetMatrix();

                double* pModelViewMatrix = stackalloc double[16];
                double* pProjectionMatrix = stackalloc double[16];
                int* pViewport = stackalloc int[4];
                double ox, oy, oz;

                gl.GetDoublev(gl.MODELVIEW_MATRIX, (IntPtr)pModelViewMatrix);
                gl.GetDoublev(gl.PROJECTION_MATRIX, (IntPtr)pProjectionMatrix);
                gl.GetIntegerv(gl.VIEWPORT, (IntPtr)pViewport);

                gl.uUnProject(pt.x, (double)pViewport[3] - pt.y, 0.0, (IntPtr)pModelViewMatrix, (IntPtr)pProjectionMatrix, (IntPtr)pViewport, (IntPtr)(&ox), (IntPtr)(&oy), (IntPtr)(&oz));
                near = new Vec3(ox, oy, oz);

                gl.uUnProject(pt.x, (double)pViewport[3] - pt.y, 1.0, (IntPtr)pModelViewMatrix, (IntPtr)pProjectionMatrix, (IntPtr)pViewport, (IntPtr)(&ox), (IntPtr)(&oy), (IntPtr)(&oz));
                far = new Vec3(ox, oy, oz);
            }
        }

        public static int IsInFront(Vec3 p, Vec3 n, Vec3 c, double rad)
        {
            double d = (c - p) * n;
            if (d > rad)
                return 1;
            if (d > -rad)
                return 2;
            return 0;
        }


        public void DrawNEDShit()
        { 
            switch (GlobalVars.NedMain.map_mode)
            {
                case NedMain.MapMode.Surface:
                    {
                        JKSurface surf = GlobalVars.Level.sectors[GlobalVars.NedMain.Cur_SC].surfaces[GlobalVars.NedMain.Cur_SF];

                        DebugDrawPolygon(surf.Poly3, Color4.yellow());
                    }
                    break;
            }


#if true
            // HIGHLIGHT SURFACES THAT ARE OPAQUE BUT PASSABLE
            if (GlobalVars.NedMain.ShowWeirdAdjoins)
            {

                foreach (JKSector sec in GlobalVars.Level.sectors)
                {
                    foreach (JKSurface surf in sec.surfaces)
                    {
                        if (surf.adjoin == null)
                            continue;

                        bool renderedOpaque = (surf.geo != Surface.GeoModeType.NotRendered) && ((surf.faceFlags & Surface.FaceFlag.Transparent) == 0);
                        bool passable = ((surf.adjoinFlags & Adjoin.Flag.Passable) != 0);

                        // normal adjoin
                        if (!renderedOpaque && passable)
                            continue;

                        // "normal" wall = adjoin, but is rendered and impassable;  prolly not typical in official levels, we'll see
                        if (renderedOpaque && !passable)
                        {
                            // if connected to a cog then its prolly openable and is therefore OK;  some other color
                            bool foundCOGLink = false;
                            foreach (JKCOG cog in GlobalVars.Level.cogs)
                                foreach (COGValue cv in cog.vals)
                                    if (cv.cogType == COGType.Surface)
                                    {
                                        JKSurface checkSurf = cv.obj as JKSurface;

                                        if (object.ReferenceEquals(checkSurf, surf))
                                        {
                                            foundCOGLink = true;
                                            break;
                                        }
                                    }


                            DebugDrawPolygon(surf.Poly3, foundCOGLink ? Color4.cyan() : Color4.pink());

                            continue;
                        }

                        // secret wall;  adjoin is rendered but also passable
                        if (renderedOpaque && passable)
                        {
                            // if either sector is water then its prolly just opaque water surface
                            if (
                                ((sec.flags & Sector.Flag.Underwater) != 0) ||
                                ((surf.adjoin.sector.flags & Sector.Flag.Underwater) != 0)
                                )
                                continue;


                            DebugDrawPolygon(surf.Poly3, Color4.green());

                            continue;
                        }

                        // fake wall;  adjoin is invisible but not passable
                        if (!renderedOpaque && !passable)
                        {
                            DebugDrawPolygon(surf.Poly3, Color4.red());

                            continue;
                        }

                        // dunno wut condition this would be
                    }
                }
            }
#endif
        }

        public JKSector SelSC = null;
        public JKThing SelTH = null;
        public int SelSF = -1;

        private int GetIntersectingSurf(T3DPMesh m, ref double cdist, Vec3 a1, Vec3 a2, Vec3 lvec)
        {
            if (m == null)
                return -1;

            Vec3 cv = m.center - a1;
            double c2 = Utils.Sqr(cv.x) + Utils.Sqr(cv.y) + Utils.Sqr(cv.z);
            double a = lvec * cv;

            if (a < (-m.radius))
                return -1;

            double b = Math.Sqrt(c2 - Utils.Sqr(a));
            if (b > m.radius)
                return -1;

            int result = -1;

            for (int i = 0; i < m.surfs.Count; i++)
            {
                T3DPSurf s = m.surfs[i];

                if ((s.normal * lvec) > 0.0)
                    continue;

                if (s.vxds.Count == 0)
                    continue;

                TVXdets vxd = s.vxds[0];
                Vec3 r;
                if (!Utils.PlaneLineXnNew(s.normal, vxd.v, a1, a2, out r))
                    continue;

                double dist = (r - a1) * lvec;
                if (dist < 0.0)
                    continue;

                if (dist > cdist)
                    continue;

                if (!IsPointOn3DPSurf(s, r))
                    continue;

                cdist = dist;
                result = i;
            }

            return result;
        }

        private bool IsPointOn3DPSurf(T3DPSurf surf, Vec3 p)
        {
            int vc = surf.vxds.Count;

            for (int i = 0; i < vc; i++)
            {
                Vec3 v1 = surf.vxds[i].v;
                Vec3 v2 = surf.vxds[Utils.NextIdx(i, vc)].v;

                Vec3 vct = surf.normal ^ (v2 - v1);
                double dist = vct * (p - v1);
                if (MathLib.CompareScalar(dist, 0.0) < 0)
                    return false;
            }

            return true;
        }

        public PickType PickAt(Vec2 pt)
        {
            double cdist = (double)1e100;
            PickType result = PickType.Nothing;

            Vec3 a1, a2;
            GetWorldLine(Vec2.Zero, out a1, out a2);// wtf? why are we doing this

            GetWorldLine(pt, out a1, out a2);
            Vec3 lvec = (a2 - a1).Unit;

            for (int i = 0; i < sclist.Count; i++)
            {
                T3DPMesh m = sclist[i];
                int n = GetIntersectingSurf(m, ref cdist, a1, a2, lvec);
                if (n == -1)
                    continue;

                result = PickType.Surface;
                SelSC = sectors[i];
                SelSF = n;
                for (int j = 0; j < SelSC.surfaces.Count; j++)
                {
                    if (SelSC.surfaces[j].d3did == n)
                    {
                        SelSF = j;
                        break;
                    }
                }
            }

            if (!GlobalVars.NedMain.AllSectors3DOPreview)
            {
                for (int i = 0; i < thlist.Count; i++)
                {
                    T3DPMesh m = thlist[i];
                    int n = GetIntersectingSurf(m, ref cdist, a1, a2, lvec);
                    if (n == -1)
                        continue;

                    result = PickType.Thing;
                    SelTH = things[i];
                }
            }

            return result;
        }

        public void Redraw()
        {
            using (Context)
            {
                SetMatrix();

                Vec3 a1, a2;
                Vec3 b1, b2;

                // front plane
                GetWorldLine(new Vec2(vwidth / 2.0, vheight / 2.0), out a1, out a2);
                Vec3 fn = (a2 - a1).Unit;
                Vec3 fp = CamPos + FrontPlane * fn;

                // left plane
                GetWorldLine(new Vec2(0.0, 0.0), out a1, out a2);
                GetWorldLine(new Vec2(0.0, vheight), out b1, out b2);
                Vec3 ln = ((b1 - a1) ^ (a2 - a1)).Unit;
                Vec3 lp = a1;

                // right plane
                GetWorldLine(new Vec2(vwidth, 0.0), out a1, out a2);
                GetWorldLine(new Vec2(vwidth, vheight), out b1, out b2);
                Vec3 rn = ((a2 - a1) ^ (b1 - a1)).Unit;
                Vec3 rp = a1;

                // top plane
                GetWorldLine(new Vec2(0.0, 0.0), out a1, out a2);
                GetWorldLine(new Vec2(vwidth, 0.0), out b1, out b2);
                Vec3 tn = ((a2 - a1) ^ (b1 - a1)).Unit;
                Vec3 tp = a1;

                // bottom plane
                GetWorldLine(new Vec2(0.0, vheight), out a1, out a2);
                GetWorldLine(new Vec2(vwidth, vheight), out b1, out b2);
                Vec3 bn = ((b1 - a1) ^ (a2 - a1)).Unit;
                Vec3 bp = a1;


                foreach (T3DPMesh m in sclist)
                {
                    if (m == null)
                        continue;

                    if (!SkipViewFrustumTest &&
                        (IsInFront(fp, fn, m.center, m.radius) == 0 ||
                        IsInFront(lp, ln, m.center, m.radius) == 0 ||
                        IsInFront(rp, rn, m.center, m.radius) == 0 ||
                        IsInFront(tp, tn, m.center, m.radius) == 0 ||
                        IsInFront(bp, bn, m.center, m.radius) == 0))
                        continue;

                    DrawMesh(m);
                }

                if (!GlobalVars.NedMain.AllSectors3DOPreview)
                {
                    foreach (T3DPMesh m in thlist)
                    {
                        if (m == null)
                            continue;

                        if (!SkipViewFrustumTest &&
                            (IsInFront(fp, fn, m.center, m.radius) == 0 ||
                            IsInFront(lp, ln, m.center, m.radius) == 0 ||
                            IsInFront(rp, rn, m.center, m.radius) == 0 ||
                            IsInFront(tp, tn, m.center, m.radius) == 0 ||
                            IsInFront(bp, bn, m.center, m.radius) == 0))
                            continue;

                        DrawMesh(m);
                    }
                }
            }
        }


        // not really sure why the view frustum test is not working in respicker 3do preview.  perhaps because
        // the IsInFront logic assumes no transformation of world geometry;  our renderer was copied mainly from
        // the base class or GL code of jed, whereas the 3do preview seems to be direct3d specifically;  and its
        // pitch/yaw model rotation is done directly via some direct3d frame class. long story short, this is
        // an easy hack to make sure the meshes are rendered in 3do preview.
        public bool SkipViewFrustumTest = true;// false;    // u know what it doesnt work in 3d preview either.  screw it.

        void DrawMesh(T3DPMesh m)
        {
            if (m == null)
                return;

            foreach (T3DPSurf surf in m.surfs)
            {
                if(surf.tx != null)
                    BindTexture(surf.tx.ObjIdx);

                gl.Begin(gl.POLYGON);
                foreach (TVXdets vxd in surf.vxds)
                {
                    gl.Color(vxd.clr);
                    if(surf.tx != null)
                        gl.TexCoord(vxd.tv);
                    gl.Vertex(vxd.v);
                }
                gl.End();
            }
        }

        public void AddThing(JKThing th)
        {
            T3DPMesh m = ThingToMesh(th);
            if (m == null)
                m = BadThingMesh(th);

            thlist.Add(m);
            things.Add(th);
        }

        public void DeleteThing(JKThing th)
        {
            int i = things.IndexOf(th);
            if (i == -1)
                return;

            things.RemoveAt(i);
            thlist.RemoveAt(i);
        }

        public void UpdateThing(JKThing th)
        {
            int i = things.IndexOf(th);
            if (i == -1)
                return;

            T3DPMesh m = ThingToMesh(th);
            if (m == null)
                m = BadThingMesh(th);

            thlist[i] = m;
        }

        public void ClearThings()
        {
            things.Clear();
            thlist.Clear();
        }

        public void AddSector(JKSector s)
        {
            T3DPMesh m = SecToMesh(s);
            sclist.Add(m);
            sectors.Add(s);
        }

        public void DeleteSector(JKSector s)
        {
            int i = sectors.IndexOf(s);
            if (i == -1)
                return;

            sectors.RemoveAt(i);
            sclist.RemoveAt(i);
        }

        public void UpdateSector(JKSector s)
        {
            int i = sectors.IndexOf(s);
            if (i == -1)
                return;

            sclist[i] = SecToMesh(s);
        }

        public void SetGamma(double g)
        {
            gamma = g;
        }

        public void ClearSectors()
        {
            //MasterPalSet = false;
            ClearTXList();
            sectors.Clear();
            sclist.Clear();
        }

        public T3DPMesh BadThingMesh(JKThing th)
        {
            T3DPMesh m = new T3DPMesh();
            T3DPSurf surf;

            surf = new T3DPSurf();
            m.surfs.Add(surf);
            surf.normal = Vec3.PosY;
            surf.tx = null;
            surf.vxds.AddRange(new TVXdets[] {
                TVXdets.Quick(th.pos + new Vec3(0.05, 0.0, 0.05)),
                TVXdets.Quick(th.pos + new Vec3(0.05, 0.0, -0.05)),
                TVXdets.Quick(th.pos + new Vec3(-0.05, 0.0, -0.05)),
                TVXdets.Quick(th.pos + new Vec3(-0.05, 0.0, 0.05)),
            });

            surf = new T3DPSurf();
            m.surfs.Add(surf);
            surf.normal = Vec3.NegY;
            surf.tx = null;
            surf.vxds.AddRange(new TVXdets[] {
                TVXdets.Quick(th.pos + new Vec3(-0.05, 0.0, 0.05)),
                TVXdets.Quick(th.pos + new Vec3(-0.05, 0.0, -0.05)),
                TVXdets.Quick(th.pos + new Vec3(0.05, 0.0, -0.05)),
                TVXdets.Quick(th.pos + new Vec3(0.05, 0.0, 0.05)),
            });

            m.CalculateSphere();

            return m;
        }

        public T3DPMesh SecToMesh(JKSector sec)
        {
            T3DPMesh m = new T3DPMesh();

            foreach (JKSurface surf in sec.surfaces)
            {
                if ((sec.flags & Sector.Flag.Preview3DO) == 0)
                {
                    if (surf.geo == Surface.GeoModeType.NotRendered || string.IsNullOrEmpty(surf.material))
                    {
                        surf.d3did = -1;
                        continue;
                    }
                }

                T3DPSurf msurf = new T3DPSurf();
                m.surfs.Add(msurf);

                msurf.normal = surf.normal;
                if ((sec.flags & Sector.Flag.Preview3DO) != 0)
                    msurf.normal *= -1.0;

                int nvx = Math.Min(24, surf.vertices.Count);

                surf.d3did = m.surfs.Count - 1;

                TOGLTexture ttx = null;
                try
                {
                    ttx = GetTexture(surf.material, sec.colormap);
                    msurf.tx = ttx;
                }
                catch (Exception e)
                {
                    Utils.PanMessage(Utils.MessageType.Warning, "Cannot load " + surf.material + " for sector " + sec.num.ToString() + " surface " + sec.surfaces.IndexOf(surf).ToString() + ": " + e.Message);
                }


                for (int j = 0; j < nvx; j++)
                {
                    int nv;
                    if ((sec.flags & Sector.Flag.Preview3DO) == 0)
                        nv = j;
                    else
                        nv = nvx - j - 1;

                    TexVertex tv = surf.texvertices[nv];

                    TVXdets vd = new TVXdets();
                    msurf.vxds.Add(vd);

                    if (GlobalVars.P3DFullLit || (sec.flags & Sector.Flag.Preview3DO) != 0 || (surf.surfaceFlags & (Surface.Flag.Ceiling | Surface.Flag.Horizon)) != 0)
                        vd.clr = Color4.white();
                    else
                    {
                        if (sec.level.mots && GlobalVars.P3DColoredLights)
                        {
                            double elt = sec.extra + surf.extraLight;
                            vd.clr = new Color4(tv.r + elt, tv.g + elt, tv.b + elt).clamped();
                        }
                        else
                        {
                            double elt = tv.intensity + sec.extra + surf.extraLight;
                            vd.clr = Color4.white().ScaledRGB(elt).clamped();
                        }
                    }

                    if (ttx != null)
                        vd.tv = tv.v.DivComps((double)Math.Min(ttx.width, 256), (double)Math.Min(ttx.height, 256));

                    vd.v = surf.vertices[nv].v;
                }
            }

            m.CalculateSphere();

            return m;
        }

        public T3DPMesh ThingToMesh(JKThing th)
        {
            if (th.model == null)
                return null;

            T3DPMesh m = new T3DPMesh();

            // dunno?
            if (th.sec != null)
                thing_cmp = th.sec.colormap;

            Mat4 mx = Utils.CreateRotMatrix(th.orient.pitch, th.orient.yaw, th.orient.roll);

            foreach (JKModel.Mesh mesh in th.model.meshes)
            {
                foreach (JKModel.Face face in mesh.faces)
                {
                    string mat = th.model.GetMat(face.imat);
                    if (face.geo == 0 || string.IsNullOrEmpty(mat))
                        continue;

                    T3DPSurf surf = new T3DPSurf();
                    m.surfs.Add(surf);

                    face.CalcNormal();
                    surf.normal = face.normal;

                    TOGLTexture ttx = null;

                    try
                    {
                        ttx = GetTexture(mat, thing_cmp);
                        surf.tx = ttx;
                    }
                    catch (Exception)
                    {
                        Utils.PanMessage(Utils.MessageType.Warning, "Cannot load " + mat + " for thing " + th.num.ToString());
                    }

                    int nvx = Math.Min(24, face.vertices.Count);

                    for (int j = 0; j < nvx; j++)
                    {
                        TVXdets vd = new TVXdets();
                        surf.vxds.Add(vd);

                        vd.v = th.pos + face.vertices[j].v.Transformed(mx);
                        vd.clr = Color4.white();

                        if (ttx != null)
                            vd.tv = face.texvertices[j].v.DivComps((double)ttx.width, (double)ttx.height);
                        else
                            vd.tv = Vec2.Zero;
                    }
                }
            }

            m.CalculateSphere();

            return m;
        }

        public static Colormap LoadColormap(string cmp)
        {
            MemoryStream cmpstream = null;
            using (BaseFile bf = FileOperations.OpenGameFile(cmp))
                if (bf != null)
                    cmpstream = new MemoryStream(bf.Read(bf.Length));

            if (cmpstream == null)
            {
                Utils.PanMessage(Utils.MessageType.Warning, "Cannot load " + cmp);
                return null;
            }

            return new Colormap(cmp, cmpstream);
        }

        public static Material LoadMaterial(string name)
        {
            MemoryStream matstream = null;
            using (BaseFile bf = FileOperations.OpenGameFile(name))
                if (bf != null)
                    matstream = new MemoryStream(bf.Read(bf.Length));

            if (matstream == null)
            {
                Utils.PanMessage(Utils.MessageType.Warning, "Cannot load " + name);
                return null;
            }

            return new Material(name, matstream, true/*doesnt matter for NED*/);
        }

        public static Bitmap LoadMatBitmap(string name, string cmp, int index=0)
        {
            return LoadMatBitmap(name, LoadColormap(cmp), index);
        }

        public static Bitmap LoadMatBitmap(string name, Colormap cmp, int index=0)
        {
            if (cmp == null)
                return null;

            Material mat = LoadMaterial(name);
            if (mat == null)
                return null;

            Bitmap bmp, ebmp;
            mat.GenerateBitmap(out bmp, out ebmp, cmp, index);

            return bmp;
        }

        TOGLTexture GetTexture(string name, string cmp)
        {
            string combname = name + cmp;

            if (TXList.ContainsKey(combname))
                return TXList[combname];

            Colormap pcmp = null;
            if (CmpList.ContainsKey(cmp))
                pcmp = CmpList[cmp];
            else
            {
                pcmp = LoadColormap(cmp);

                CmpList.Add(cmp, pcmp);
            }

            TOGLTexture ttx = null;

            try
            {
                Bitmap bmp = LoadMatBitmap(name, pcmp);
                if (bmp == null)
                    return null;

                using (Context)
                {
                    ttx = new TOGLTexture();
                    ttx.cmpname = cmp;
                    ttx.width = bmp.Width;
                    ttx.height = bmp.Height;
                    ttx.ObjIdx = GenerateTexture(bmp);
                }
            }
            finally
            {
                TXList.Add(combname, ttx);
            }

            return ttx;
        }

        public unsafe uint GenerateTexture(Bitmap bmp)
        {
            uint id;
            gl.GenTextures(1, (IntPtr)(&id));
            BindTexture(id);

            // assume 24 RGB or 32 ARGB
            int colorbits = (bmp.PixelFormat == System.Drawing.Imaging.PixelFormat.Format24bppRgb ? 24 : 32);

            BitmapData bmdat = bmp.LockBits(new Rectangle(0, 0, bmp.Width, bmp.Height), ImageLockMode.ReadOnly, bmp.PixelFormat);

            gl.TexImage2D(gl.TEXTURE_2D, 0, colorbits / 8, bmp.Width, bmp.Height, 0, (colorbits == 24) ? gl.BGR_EXT : gl.BGRA_EXT, gl.UNSIGNED_BYTE, bmdat.Scan0);

            bmp.UnlockBits(bmdat);

            // set up texture parameters
            gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, (int)gl.NEAREST);//(int)(Program.MainForm.gpuMipmaps.Checked ? (Program.MainForm.bilinearFiltering.Checked ? gl.LINEAR_MIPMAP_LINEAR : gl.NEAREST_MIPMAP_NEAREST) : (Program.MainForm.bilinearFiltering.Checked ? gl.LINEAR : gl.NEAREST)));
            gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, (int)gl.NEAREST);//(int)(Program.MainForm.bilinearFiltering.Checked ? gl.LINEAR : gl.NEAREST));
            gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, (int)gl.REPEAT);
            gl.TexParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, (int)gl.REPEAT);

#if false
            gl.TexParameterf(gl.TEXTURE_2D, gl.TEXTURE_MAX_ANISOTROPY_EXT, Program.MainForm.anisotropicTextures.Checked ? AnisotropicFilter : 1.0f/*"off"*/);

            if (Program.MainForm.gpuMipmaps.Checked)
            {
                gl.Hint(gl.GENERATE_MIPMAP_HINT, gl.NICEST);
                gl.GenerateMipmap(gl.TEXTURE_2D);
            }
#endif

            return id;
        }

        //bool MasterPalSet = false;

        public void ClearTXList()
        {
            using (Context)
            {
                foreach (string key in TXList.Keys)
                {
                    TOGLTexture tex = TXList[key];
                    if (tex == null)
                        continue;

                    uint i = tex.ObjIdx;
                    gl.DeleteTextures(1, (IntPtr)(&i));
                }

                TXList.Clear();
            }

            CmpList.Clear();
        }

        public PreviewRenderer(Control ctrl)
            : base(ctrl.Handle, 32, 16, 8, 0, true)
        {
            using (Context)
            {
                /*gl.Disable(gl.LINE_SMOOTH);
                gl.Disable(gl.POLYGON_SMOOTH);
                gl.Disable(gl.POINT_SMOOTH);*/

                gl.Disable(gl.LIGHTING);

                gl.Clear(gl.COLOR_BUFFER_BIT);
                gl.SwapBuffers(hDC);
            }
        }

        protected override uint LoadTexture(string filename)
        {
            return 0;
        }
    }
}
