﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace Smith.NED
{
    public unsafe class Renderer : RendererBase
    {
        public Vec3 CamPos = Vec3.Zero;
        public Vec3 XVec = Vec3.PosX;
        public Vec3 YVec = Vec3.PosY;
        public Vec3 ZVec = Vec3.PosZ;
        public Vec3 GridPos = Vec3.Zero;
        public Vec3 GridXVec = Vec3.PosX;
        public Vec3 GridYVec = Vec3.PosY;
        public Vec3 GridNormal = Vec3.PosZ;
        public double scale = 1.0;
        public int vpx, vpy, vpw, vph;

        public override void SetupViewport(System.Drawing.Rectangle windowClient, Box2 displayRCMM)
        {
            base.SetupViewport(windowClient, displayRCMM);

            vpx = windowClient.Left;
            vpy = windowClient.Top;
            vpw = windowClient.Width;
            vph = windowClient.Height;
        }

        public Renderer(Control ctrl)
            : base(ctrl.Handle, 32, 16, 8, 0, true)
        {
            using (Context)
            {
                /*gl.Disable(gl.LINE_SMOOTH);
                gl.Disable(gl.POLYGON_SMOOTH);
                gl.Disable(gl.POINT_SMOOTH);*/

                gl.Disable(gl.LIGHTING);

                gl.FrontFace(gl.CCW);

                gl.Clear(gl.COLOR_BUFFER_BIT);
                gl.SwapBuffers(hDC);
            }
        }

        public const int XRange = 20000;
        public const int YRange = 20000;
        public const int ZRange = 20000;
        public const int ppunit = 16;

        public List<int> Selected = new List<int>();

        public void SetRenderMatrix()
        {
            gl.MatrixMode(gl.PROJECTION);
            gl.LoadIdentity();

            double dpx = (double)vpw / (double)ppunit * scale;
            double dpy = (double)vph / (double)vpw * dpx;

            gl.Ortho(-dpx / 2.0, dpx / 2.0, -dpy / 2.0, dpy / 2.0, -(double)ZRange / 2.0, (double)ZRange / 2.0);

            SetCamera();
        }

        public Vec2 ProjectPoint(Vec3 v)
        {
            using (Context)
            {
                SetRenderMatrix();

                double* pModelViewMatrix = stackalloc double[16];
                double* pProjectionMatrix = stackalloc double[16];
                int* pViewport = stackalloc int[4];
                double wx, wy, wz;

                gl.GetDoublev(gl.MODELVIEW_MATRIX, (IntPtr)pModelViewMatrix);
                gl.GetDoublev(gl.PROJECTION_MATRIX, (IntPtr)pProjectionMatrix);
                gl.GetIntegerv(gl.VIEWPORT, (IntPtr)pViewport);

                gl.uProject(v.x, v.y, v.z, (IntPtr)pModelViewMatrix, (IntPtr)pProjectionMatrix, (IntPtr)pViewport, (IntPtr)(&wx), (IntPtr)(&wy), (IntPtr)(&wz));

                return new Vec2(wx, (double)pViewport[3] - wy);
            }
        }

        public Vec3 UnprojectPoint(Vec3 v)
        {
            using (Context)
            {
                SetRenderMatrix();

                double* pModelViewMatrix = stackalloc double[16];
                double* pProjectionMatrix = stackalloc double[16];
                int* pViewport = stackalloc int[4];
                double ox, oy, oz;

                gl.GetDoublev(gl.MODELVIEW_MATRIX, (IntPtr)pModelViewMatrix);
                gl.GetDoublev(gl.PROJECTION_MATRIX, (IntPtr)pProjectionMatrix);
                gl.GetIntegerv(gl.VIEWPORT, (IntPtr)pViewport);

                gl.uUnProject(v.x, (double)pViewport[3] -v.y, v.z, (IntPtr)pModelViewMatrix, (IntPtr)pProjectionMatrix, (IntPtr)pViewport, (IntPtr)(&ox), (IntPtr)(&oy), (IntPtr)(&oz));

                return new Vec3(ox, oy, oz);
            }
        }

        public bool GetXYZOnPlaneAt(Vec2 pt, Vec3 pnormal, Vec3 p, out Vec3 v)
        {
            using (Context)
            {
                SetRenderMatrix();

                double* pModelViewMatrix = stackalloc double[16];
                double* pProjectionMatrix = stackalloc double[16];
                int* pViewport = stackalloc int[4];
                double ox, oy, oz;

                gl.GetDoublev(gl.MODELVIEW_MATRIX, (IntPtr)pModelViewMatrix);
                gl.GetDoublev(gl.PROJECTION_MATRIX, (IntPtr)pProjectionMatrix);
                gl.GetIntegerv(gl.VIEWPORT, (IntPtr)pViewport);

                gl.uUnProject(pt.x, (double)pViewport[3] - pt.y, 0.0, (IntPtr)pModelViewMatrix, (IntPtr)pProjectionMatrix, (IntPtr)pViewport, (IntPtr)(&ox), (IntPtr)(&oy), (IntPtr)(&oz));
                Vec3 near = new Vec3(ox, oy, oz);

                gl.uUnProject(pt.x, (double)pViewport[3] - pt.y, 1.0, (IntPtr)pModelViewMatrix, (IntPtr)pProjectionMatrix, (IntPtr)pViewport, (IntPtr)(&ox), (IntPtr)(&oy), (IntPtr)(&oz));
                Vec3 far = new Vec3(ox, oy, oz);

                double d = pnormal * p;

                bool ret = Utils.PlaneLineXn(pnormal, d, near, far, out v);
                ret = ret && (Math.Abs(v.x) < (double)XRange) && (Math.Abs(v.y) < (double)YRange) && (Math.Abs(v.z) < (double)ZRange);
                if (ret)
                    Utils.JKRound(ref v);
                return ret;

            }
        }

        public void GetNearestGrid(ref Vec3 v)
        {
            double revgs = 1.0 / GridStep;

            double px = GridXVec * (v - GridPos);
            double py = GridYVec * (v - GridPos);

            px = Math.Round(px * revgs) * GridStep;
            py = Math.Round(py * revgs) * GridStep;

            v.x = GridPos.x + (px * GridXVec.x) + (py * GridYVec.x);
            v.y = GridPos.y + (px * GridXVec.y) + (py * GridYVec.y);
            v.z = GridPos.z + (px * GridXVec.z) + (py * GridYVec.z);
        }

        public bool GetGridAt(Vec2 pt, out Vec3 v)
        {
            bool ret = GetXYZOnPlaneAt(pt, GridNormal, GridPos, out v);
            if (ret)
                Utils.JKRound(ref v);
            return ret;
        }

        public void SetCamera()
        {
            gl.MatrixMode(gl.MODELVIEW);
            gl.LoadIdentity();

            gl.uLookAt(Vec3.Zero, -ZVec, YVec);

            gl.Translate(CamPos);
        }

        private const int SelectBufEntries = 2048;
        private IntPtr pSelectBuf = IntPtr.Zero;

        public void SetPointSize(double s)
        {
            using (Context)
            {
                gl.PointSize((float)s);
            }
        }

        Plane3 lplane = Plane3.Zero;
        Plane3 rplane = Plane3.Zero;
        Plane3 uplane = Plane3.Zero;
        Plane3 bplane = Plane3.Zero;

        public void BeginRectPick(Vec2 pt1, Vec2 pt2)
        {
            double d;
            if (pt1.x > pt2.x)
            {
                d = pt2.x;
                pt2.x = pt1.x;
                pt1.x = d;
            }
            if (pt1.y > pt2.y)
            {
                d = pt2.y;
                pt2.y = pt1.y;
                pt1.y = d;
            }

            using (Context)
            {
                Vec3 p1 = UnprojectPoint(pt1.At3DZ(0.0));
                Vec3 p2 = UnprojectPoint(pt1.At3DZ(1.0));
                Vec3 p3 = UnprojectPoint(new Vec3(pt1.x, pt2.y, 0.0));
                lplane = new Plane3(((p3 - p1) ^ (p2 - p1)).Unit, p1);

                p1 = UnprojectPoint(new Vec3(pt2.x, pt1.y, 0.0));
                p2 = UnprojectPoint(new Vec3(pt2.x, pt1.y, 1.0));
                p3 = UnprojectPoint(pt2.At3DZ(0.0));
                rplane = new Plane3(((p2 - p1) ^ (p3 - p1)).Unit, p1);

                p1 = UnprojectPoint(pt1.At3DZ(0.0));
                p2 = UnprojectPoint(pt1.At3DZ(1.0));
                p3 = UnprojectPoint(new Vec3(pt2.x, pt1.y, 0.0));
                uplane = new Plane3(((p2 - p1) ^ (p3 - p1)).Unit, p1);

                p1 = UnprojectPoint(new Vec3(pt1.x, pt2.y, 0.0));
                p2 = UnprojectPoint(new Vec3(pt1.x, pt2.y, 1.0));
                p3 = UnprojectPoint(pt2.At3DZ(0.0));
                bplane = new Plane3(((p3 - p1) ^ (p2 - p1)).Unit, p1);
            }
        }

        public void EndRectPick()
        {

        }

        public bool IsPolygonInRect(Polygon p)
        {
            foreach (Vertex v in p.vertices)
                if (!IsVertexInRect(v.v))
                    return false;

            return true;
        }

        public bool ArePolygonsInRect(IEnumerable<Polygon> ps)
        {
            foreach (Polygon p in ps)
                if (!IsPolygonInRect(p))
                    return false;

            return true;
        }

        public bool IsLineInRect(Vec3 v1, Vec3 v2)
        {
            return (IsVertexInRect(v1) && IsVertexInRect(v2));
        }

        public static bool IsInFront(Vec3 v, Plane3 p)
        {
            return (v - p.Point) * p.n > 0.0;
        }

        public bool IsVertexInRect(Vec3 v)
        {
            return IsInFront(v, lplane) && IsInFront(v, rplane) && IsInFront(v, uplane) && IsInFront(v, bplane);
        }

        public void BeginPick(Vec2 pt)
        {
            if (pSelectBuf != IntPtr.Zero)
            {
                Utils.PanMessage(Utils.MessageType.Warning, "BeginPick called when pick is already active");
            }

            using (Context)
            {
                pSelectBuf = Marshal.AllocHGlobal(sizeof(int) * SelectBufEntries);
                gl.SelectBuffer(sizeof(int) * SelectBufEntries, pSelectBuf);
                gl.RenderMode(gl.SELECT);

                gl.Disable(gl.CULL_FACE);

                gl.InitNames();
                gl.PushName(0);
                SetPickMatrix(pt);
            }
        }

        void SetPickMatrix(Vec2 pt)
        {
            using (Context)
            {
                int* pViewport = stackalloc int[4];
                gl.GetIntegerv(gl.VIEWPORT, (IntPtr)pViewport);

                gl.MatrixMode(gl.PROJECTION);
                gl.LoadIdentity();
                gl.uPickMatrix(pt.x, (double)pViewport[3] - pt.y, 4.0, 4.0, (IntPtr)pViewport);

                double dpx = (double)vpw / (double)ppunit * scale;
                double dpy = (double)vph / (double)vpw * dpx;
                gl.Ortho(-dpx / 2.0, dpx / 2.0, -dpy / 2.0, dpy / 2.0, -(double)ZRange / 2.0, (double)ZRange / 2.0);
                
                SetCamera();
            }
        }

        public void RenderPoly(Polygon p)
        {
            gl.Begin(gl.POLYGON);
            foreach (Vertex v in p.vertices)
                gl.Vertex(v.v);
            gl.End();
        }

        public void RenderLine(Vec3 v1, Vec3 v2)
        {
            gl.Begin(gl.LINES);
            gl.Vertex(v1);
            gl.Vertex(v2);
            gl.End();
        }

        public void PickPolygon(Polygon p, int id)
        {
            using (Context)
            {
                gl.LoadName((uint)id);
                RenderPoly(p);
            }
        }

        public void PickPolygons(IEnumerable<Polygon> ps, int id)
        {
            using (Context)
            {
                gl.LoadName((uint)id);
                foreach (Polygon poly in ps)
                    RenderPoly(poly);
            }
        }

        public void PickPolygons(IEnumerable<JKSurface> ps, int id)
        {
            using (Context)
            {
                gl.LoadName((uint)id);
                foreach (JKSurface poly in ps)
                    RenderPoly(poly);
            }
        }

        public void PickLine(Vec3 v1, Vec3 v2, int id)
        {
            using (Context)
            {
                gl.LoadName((uint)id);
                RenderLine(v1, v2);
            }
        }

        public void PickVertex(Vec3 v, int id)
        {
            using (Context)
            {
                gl.LoadName((uint)id);
                DrawVertex(v);
            }
        }

        public void EndPick()
        {
            if (pSelectBuf == IntPtr.Zero)
            {
                Utils.PanMessage(Utils.MessageType.Warning, "EndPick called when pick is not active");
                return;
            }

            using (Context)
            {
                gl.Flush();
                int hits = gl.RenderMode(gl.SELECT);
                Selected.Clear();
                List<double> zs = new List<double>();
                int* pi = (int*)pSelectBuf;

                for (int i = 0; i < hits; i++)
                {
                    int names = *(pi++);
                    double z = (double)*((uint*)pi) / (double)uint.MaxValue;
                    pi += 2;
                    for (int j = 0; j < names; j++)
                    {

                        {
                            int n = Selected.Count;
                            for (int x = 0; x < n; x++)
                                if (z < zs[x])
                                {
                                    n = x;
                                    break;
                                }

                            Selected.Insert(n, *pi);
                            zs.Insert(n, z);
                        }

                        pi++;
                    }
                }

                gl.RenderMode(gl.RENDER);
                gl.Disable(gl.CULL_FACE);

                Marshal.FreeHGlobal(pSelectBuf);
                pSelectBuf = IntPtr.Zero;
            }
        }

        public void BeginScene()
        {
            gl.Disable(gl.TEXTURE_2D);
            gl.PolygonMode(gl.FRONT_AND_BACK, gl.LINE);

            SetRenderMatrix();
        }

        public void EndScene()
        {
            gl.PolygonMode(gl.FRONT_AND_BACK, gl.FILL);

            /*Procedure TOGLRenderer.EndScene;
var oldpen,pen:Hpen;
begin
 glFlush;
 SwapBuffers(hdc);
 if crad<>0 then
 begin
  SelectObject(hdc,GetStockObject(NULL_BRUSH));
  pen:=CreatePen(PS_SOLID,1,ccolor);
  oldpen:=SelectObject(hdc,pen);
  Ellipse(hdc,ccenterx-crad,ccentery-crad,ccenterx+crad,ccentery+crad);
  SelectObject(hdc,oldpen);
  DeleteObject(pen);
 end;
end;       */
        }

        Color4 curcolor = Color4.white();

        public void SetColor(Color4 clr)
        {
            curcolor = clr;
            gl.Color(clr);
        }

        public double GridDot = 0.2;
        public double GridStep = 0.2;
        public double GridLine = 1.0;

        public enum CullMode
        {
            None,
            Front,
            Back
        }

        public void SetCulling(CullMode cm)
        {
            switch (cm)
            {
                case CullMode.None:
                    gl.Disable(gl.CULL_FACE);
                    break;

                case CullMode.Back:
                    gl.Enable(gl.CULL_FACE);
                    gl.CullFace(gl.BACK);
                    break;

                case CullMode.Front:
                    gl.Enable(gl.CULL_FACE);
                    gl.CullFace(gl.FRONT);
                    break;
            }
        }

        public void DrawPolygonsAt(IEnumerable<Poly3> polys, Vec3 pos, Orient orient)
        {
            Mat4 mx = Utils.CreateRotMatrix(orient.pitch, orient.yaw, orient.roll);
            foreach (Poly3 poly in polys)
            {
                gl.Begin(gl.POLYGON);
                foreach (Vec3 v in poly)
                    gl.Vertex(pos + v.Transformed(mx));
                gl.End();
            }
        }

        public void DrawPolygonsAt(Polygons polys, Vec3 pos, Orient orient)
        {
            DrawPolygonsAt(polys.Poly3, pos, orient);
        }

        public void DrawPolygons(IEnumerable<Poly3> polys)
        {
            foreach (Poly3 poly in polys)
                DrawPolygon(poly);
        }

        public void DrawPolygon(Poly3 poly)
        {
            gl.Begin(gl.POLYGON);
            foreach (Vec3 v in poly)
                gl.Vertex(v);
            gl.End();
        }

        public void DrawPolygon(Polygon poly)
        {
            gl.Begin(gl.POLYGON);
            foreach (Vertex v in poly.vertices)
                gl.Vertex(v.v);
            gl.End();
        }

        public void DrawPolygon(JKSurface surf)
        {
            DrawPolygon((Polygon)surf);
        }

        public void DrawPolygons(JKSurfaces surfs)
        {
            foreach (JKSurface surf in surfs)
                DrawPolygon(surf);
        }

        public void DrawVertex(Vec3 v)
        {
            gl.Begin(gl.POINTS);
            gl.Vertex(v);
            gl.End();
        }

        public void DrawCircle(Vec3 c, double rad)
        {
#if true
            Poly3 poly = new Poly3();
            for (double a = 0.0; a < 360.0; a += 10.0)
                poly.Add(c + (XVec * rad).Transformed(Mat4.RotationAroundAxis(ZVec, MathLib.ToRad(a))));

            DrawPolygon(poly);
#else
            Vec2 p = ProjectPoint(c);
            double r = (ProjectPoint(c + XVec * rad) - p).Magnitude;

            Poly2 poly = new Poly2();
            for (double a = 0.0; a < 360.0; a += 10.0)
                poly.Add(p + new Vec2(MathLib.ToRad(a)) * r);

            DrawPolygon(poly, curcolor);
#endif
        }

        public void DrawLine(Vec3 a, Vec3 b)
        {
            gl.Begin(gl.LINES);
            gl.Vertex(a);
            gl.Vertex(b);
            gl.End();
        }

        public void DrawVertices(JKVertices verts)
        {
            gl.Begin(gl.POINTS);
            foreach (JKVertex vert in verts)
                gl.Vertex(vert.v);
            gl.End();
        }

        public void DrawGrid()
        {
            int n = (int)Math.Round(GlobalVars.GridSize / GridDot);

            for (int i = -(n / 2 - 1); i <= (n / 2 - 1); i++)
                for (int j = -(n / 2 - 1); j <= (n / 2 - 1); j++)
                    DrawVertex(GridPos + ((double)i * GridDot * GridXVec) + ((double)j * GridDot * GridYVec));


            n = (int)Math.Round(GlobalVars.GridSize / GridLine);
            double gsize2 = GlobalVars.GridSize / 2.0;
            for (int i = -(n / 2); i <= (n / 2); i++)
                DrawLine(GridPos + ((double)i * GridLine * GridXVec) + (gsize2 * GridYVec),
                         GridPos + ((double)i * GridLine * GridXVec) + (-gsize2 * GridYVec));

            for (int i = -(n / 2); i <= (n / 2); i++)
                DrawLine(GridPos + (gsize2 * GridXVec) + ((double)i * GridLine * GridYVec),
                         GridPos + (-gsize2 * GridXVec) + ((double)i * GridLine * GridYVec));


            SetColor(GlobalVars.clGridX);
            DrawLine(GridPos, GridPos + gsize2 * GridXVec);

            SetColor(GlobalVars.clGridY);
            DrawLine(GridPos, GridPos + gsize2 * GridYVec);
        }

        protected override uint LoadTexture(string filename)
        {
            return 0;
        }

        public void SetX(Vec3 v)
        {
            XVec = v.Unit;
            YVec = ZVec ^ XVec;
            XVec = YVec ^ ZVec;
        }

        public void SetY(Vec3 v)
        {
            YVec = v.Unit;
            XVec = YVec ^ ZVec;
            YVec = ZVec ^ XVec;
        }

        public void SetZ(Vec3 v)
        {
            Vec3 g = v.Unit;
            if (MathLib.CompareScalar(g.Magnitude, 0.0) == 0)
                return;

            ZVec = g;
        }

        public void SetGridNormal(Vec3 v)
        {
            Vec3 g = v.Unit;
            if (MathLib.CompareScalar(g.Magnitude, 0.0) == 0)
                return;

            GridNormal = g;
        }

        public void SetGridXNormal(Vec3 v)
        {
            Vec3 g = v.Unit;
            if (MathLib.CompareScalar(g.Magnitude, 0.0) == 0)
                return;

            GridXVec = g;
            GridYVec = GridNormal ^ GridXVec;
        }

        public void SetGridYNormal(Vec3 v)
        {
            GridXVec = v.Unit;
            GridYVec = GridNormal ^ GridXVec;
        }
    }
}
