﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Smith.NED
{
    public static class Undo
    {
        public enum Change
        {
            Added,
            Deleted,
            Changed
        }

        [Flags]
        public enum SectorChange
        {
            Geo = 0x04,
            Val = 0x08,
            Both = Geo|Val
        }

        public class ChangeEntry
        {
            public int id;
            public UndoChange obj;

            public ChangeEntry(int _id, UndoChange _obj)
            {
                id = _id;
                obj = _obj;
            }
        }

        public class ChangeList : List<ChangeEntry>
        {
            public void AddObject(UndoChange uc, int objid)
            {
                Add(new ChangeEntry(objid, uc));
            }

            public int GetObjID(int n)
            {
                return this[n].id;
            }

            public UndoChange GetObj(int n)
            {
                return this[n].obj;
            }

            public bool ContainsObj(int objid)
            {
                return (FindObjID(objid) != -1);
            }

            public int FindObjID(int objid)
            {
                for (int x = 0; x < Count; x++)
                    if (this[x].id == objid)
                        return x;

                return -1;
            }
        }

        public class UndoRec
        {
            public string Name;
            public ChangeList scvchanges = new ChangeList();
            public ChangeList scgchanges = new ChangeList();
            public ChangeList thchanges = new ChangeList();
            public ChangeList ltchanges = new ChangeList();
            public ChangeList nodechanges = new ChangeList();

            public void ApplyUndo()
            {
                // undo sec geometry
                for (int i = 0; i < scgchanges.Count; i++)
                {
                    int id = scgchanges.GetObjID(i);
                    UndoChange uobj = scgchanges.GetObj(i);
                    uobj.Undo(id);
                }

                // undo sec values
                for (int i = 0; i < scvchanges.Count; i++)
                {
                    int id = scvchanges.GetObjID(i);
                    UndoChange uobj = scvchanges.GetObj(i);
                    uobj.Undo(id);
                }

                // undo lights
                for (int i = 0; i < ltchanges.Count; i++)
                {
                    int id = ltchanges.GetObjID(i);
                    UndoChange uobj = ltchanges.GetObj(i);
                    uobj.Undo(id);
                }

                // undo things
                for (int i = 0; i < thchanges.Count; i++)
                {
                    int id = thchanges.GetObjID(i);
                    UndoChange uobj = thchanges.GetObj(i);
                    uobj.Undo(id);
                }

                for (int i = 0; i < nodechanges.Count; i++)
                {
                    int id = nodechanges.GetObjID(i);
                    UndoChange uobj = nodechanges.GetObj(i);
                    uobj.Undo(id);
                }
            }
        }

        public abstract class UndoChange
        {
            public Change change;
            public byte[] data;

            public JKLevel Level
            {
                get
                {
                    return GlobalVars.Level;
                }
            }

            public abstract void SaveUndo(object obj, Change ch);
            public abstract void Undo(int objid);
        }

        public class LightUndo : UndoChange
        {
            public override void SaveUndo(object obj, Change ch)
            {
                change = ch;

                if (ch == Change.Added)
                    return;

                ZipBuffer zip = new ZipBuffer();

                (obj as JedLight).SerializeTo(zip);

                data = zip.Buffer;
            }

            public override void Undo(int objid)
            {
                int n = Level.GetLightByID(objid);

                if(change == Change.Added)
                {
                    if (n == -1)
                        return;

                    Level.lights.RemoveAt(n);
                    return;
                }

                JedLight lt;

                if (change == Change.Deleted)
                {
                    if (n != -1)
                        return;

                    lt = Level.NewLight();
                    Level.lights.Add(lt);
                }
                else
                {
                    if (n == -1)
                        return;

                    lt = Level.lights[n];
                }

                ZipBuffer zip = new ZipBuffer(data);
                lt.SerializeFrom(zip);

                GlobalVars.NedMain.LevelChanged(true);
            }
        }

        public class ThingUndo : UndoChange
        {
            public override void SaveUndo(object obj, Change ch)
            {
                change = ch;

                if (ch == Change.Added)
                    return;

                ZipBuffer zip = new ZipBuffer();

                (obj as JKThing).SerializeTo(zip);

                data = zip.Buffer;
            }

            public override void Undo(int objid)
            {
                int n = Level.GetThingByID(objid);
                switch (change)
                {
                    case Change.Added:
                        if (n == -1)
                            return;
                        {
                            UndoRec tmpur = Cur;
                            Cur = null;
                            Utils.DeleteThing(Level, n);
                            Cur = tmpur;
                        }
                        break;

                    case Change.Deleted:
                        if (n != -1)
                            return;
                        {
                            JKThing th = Level.NewThing();

                            ZipBuffer zip = new ZipBuffer(data);
                            int secnum;
                            th.SerializeFrom(zip, out secnum);
                            ResolveThingRefs(th, secnum);
                            Level.things.Add(th);
                            Level.RenumThings();
                            GlobalVars.NedMain.ThingAdded(th);
                        }
                        break;

                    case Change.Changed:
                        if (n == -1)
                            return;
                        {
                            JKThing th = Level.things[n];

                            ZipBuffer zip = new ZipBuffer(data);
                            int secnum;
                            th.SerializeFrom(zip, out secnum);
                            ResolveThingRefs(th, secnum);
                            GlobalVars.NedMain.ThingChanged(th);
                        }
                        break;
                }
            }
        }

        public static void ResolveThingRefs(JKThing th, int secnum)
        {
            if (secnum < 0 || secnum >= GlobalVars.Level.sectors.Count)
                th.sec = null;
            else
                th.sec = GlobalVars.Level.sectors[secnum];
        }

        public class SecValUndo : UndoChange
        {
            public int nsfs;

            public override void SaveUndo(object obj, Change ch)
            {
                change = ch;
                if (ch != Change.Changed && ch != Change.Deleted)
                    return;

                JKSector sec = obj as JKSector;
                nsfs = sec.surfaces.Count;

                ZipBuffer zip = new ZipBuffer();
                sec.SerializeValTo(zip);
                foreach (JKSurface surf in sec.surfaces)
                    surf.SerializeValTo(zip);
                data = zip.Buffer;
            }

            public override void Undo(int objid)
            {
                JKSector sec = null;
                if (change == Change.Changed || change == Change.Deleted)
                {
                    int n = Level.GetSectorByID(objid);
                    if (n == -1)
                        return;

                    sec = Level.sectors[n];
                }
                else
                    return;

                ZipBuffer zip = new ZipBuffer(data);

                sec.SerializeValFrom(zip);

                {
                    int n = Math.Min(sec.surfaces.Count, nsfs);
                    for (int i = 0; i < n; i++)
                    {
                        sec.surfaces[i].SerializeValFrom(zip);
                        sec.surfaces[i].adjoin = null;
                    }
                }

                {
                    foreach (JKSurface surf in sec.surfaces)
                    {
                        int sc = surf.mark >> 16;
                        int sf = surf.mark & 0xFFFF;
                        surf.adjoin = null;
                        if (sc == 65535)
                            continue;

                        sc = Level.GetSectorByID(surf.nadj);
                        if (sc == -1)
                            continue;

                        if (sc >= Level.sectors.Count)
                            continue;
                        JKSector asec = Level.sectors[sc];
                        if (sf >= asec.surfaces.Count)
                            continue;
                        JKSurface asurf = asec.surfaces[sf];

                        surf.adjoin = asurf;
                        asurf.adjoin = surf;
                    }
                }

                GlobalVars.NedMain.SectorChanged(sec, true);
            }
        }

        public class SecGeoUndo : UndoChange
        {
            public int nvxs;
            public int nsfs;

            public override void SaveUndo(object obj, Change ch)
            {
                change = ch;
                if (ch == Change.Deleted || ch == Change.Changed)
                {
                    JKSector sec = obj as JKSector;
                    sec.Renumber();
                    nsfs = sec.surfaces.Count;
                    nvxs = sec.vertices.Count;

                    ZipBuffer zip = new ZipBuffer();
                    sec.vertices.SerializeTo(zip);
                    foreach (JKSurface surf in sec.surfaces)
                    {
                        zip.WriteInt(surf.vertices.Count);
                        surf.SerializeVerticesTo(zip);
                    }
                    data = zip.Buffer;
                }
            }

            private void SetSector(JKSector sec)
            {
                UndoRec tmpur = Cur;
                Cur = null;

                while (sec.vertices.Count > nvxs)
                    sec.vertices.RemoveAt(sec.vertices.Count - 1);

                while (sec.vertices.Count < nvxs)
                    sec.NewVertex();

                ZipBuffer zip = new ZipBuffer(data);

                sec.vertices.SerializeFrom(zip);

                while (sec.surfaces.Count > nsfs)
                {
                    JKSurface surf = sec.surfaces[sec.surfaces.Count - 1];
                    Utils.RemoveSurfRefs(Level, surf);
                    sec.surfaces.Remove(surf);
                }

                while (sec.surfaces.Count < nsfs)
                {
                    JKSurface surf = sec.NewSurface();
                    sec.surfaces.Add(surf);
                }

                for (int i = 0; i < nsfs; i++)
                {
                    JKSurface surf = sec.surfaces[i];

                    int n = zip.ReadInt();

                    while (surf.vertices.Count > n)
                        surf.DeleteVertex(surf.vertices.Count - 1);

                    while (surf.vertices.Count < n)
                        surf.AddVertex(new JKVertex(sec));

                    surf.SerializeVerticesFrom(zip);
                }

                sec.Renumber();
                foreach (JKSurface surf in sec.surfaces)
                    surf.Recalc();

                Cur = tmpur;
            }

            public override void Undo(int objid)
            {
                int n = Level.GetSectorByID(objid);
                switch (change)
                {
                    case Change.Added:
                        if (n == -1)
                            return;
                        {
                            UndoRec tmpur = Cur;
                            Cur = null;
                            Utils.DeleteSector(Level, n);
                            Cur = tmpur;
                        }
                        break;

                    case Change.Deleted:
                        if (n != -1)
                            return;
                        {
                            JKSector sec = Level.NewSector();
                            sec.ID = objid;
                            SetSector(sec);
                            Level.sectors.Add(sec);
                            Level.RenumSecs();
                            GlobalVars.NedMain.SectorAdded(sec);
                        }
                        break;

                    case Change.Changed:
                        if (n == -1)
                            return;
                        {
                            JKSector sec = Level.sectors[n];
                            SetSector(sec);
                            GlobalVars.NedMain.SectorChanged(sec, true);
                        }
                        break;
                }
            }
        }

        public class UrqUndo : UndoChange
        {
            public override void SaveUndo(object obj, Change ch)
            {
                change = ch;
                if(change == Change.Deleted || change == Change.Changed)
                {
                    JKModel.HNode node = obj as JKModel.HNode;

                    ZipBuffer zip = new ZipBuffer();
                    node.SerializeTo(zip);
                    data = zip.Buffer;
                }
            }

            public override void Undo(int objid)
            {
                int n = Level.GetNodeByID(objid);

                if(change == Change.Added)
                {
                    if (n == -1)
                        return;

                    Level.hnodes.RemoveAt(n);
                } else {
                    JKModel.HNode node;
                    if (change == Change.Deleted)
                    {
                        if(n != -1)
                            return;

                        node = new JKModel.HNode();
                        Level.hnodes.Add(node);
                    }
                    else
                    {
                        if (n == -1)
                            return;

                        node = Level.hnodes[n];
                    }

                    ZipBuffer zip = new ZipBuffer(data);

                    node.SerializeFrom(zip);
                }

                GlobalVars.UrqForm.UpdateMeshOffsets();
                GlobalVars.UrqForm.Reload();
                GlobalVars.NedMain.LevelChanged(false);
            }
        }

        public static UndoRec Cur = null;
        public static List<UndoRec> Stack = new List<UndoRec>();

        public static void ApplyUndo()
        {
            if (Cur == null)
                return;

            Cur.ApplyUndo();

            Cur = null;
            Stack.RemoveAt(Stack.Count - 1);
            if (Stack.Count > 0)
                Cur = Stack[Stack.Count - 1];
        }

        public static void SaveSecUndo(JKSector sec, Change change, SectorChange how)
        {
            if (Cur == null)
                return;

            if((how & SectorChange.Geo)!=0)
            {
                if(!Cur.scgchanges.ContainsObj(sec.ID))
                {
                    UndoChange lch = new SecGeoUndo();
                    lch.SaveUndo(sec, change);
                    Cur.scgchanges.AddObject(lch, sec.ID);
                }
            }

            if ((how & SectorChange.Val) != 0)
            {
                if (!Cur.scvchanges.ContainsObj(sec.ID))
                {
                    UndoChange lch = new SecValUndo();
                    lch.SaveUndo(sec, change);
                    Cur.scvchanges.AddObject(lch, sec.ID);
                }
            }
        }

        public static void SaveThingUndo(JKThing th, Change change)
        {
            if (Cur == null)
                return;

            if (Cur.thchanges.ContainsObj(th.ID))
                return;

            ThingUndo tch = new ThingUndo();
            tch.SaveUndo(th, change);
            Cur.thchanges.AddObject(tch, th.ID);
        }

        public static void SaveNodeUndo(JKModel.HNode node, Change change)
        {
            if (Cur == null)
                return;

            if (Cur.nodechanges.ContainsObj(node.id))
                return;

            UrqUndo nch = new UrqUndo();
            nch.SaveUndo(node, change);
            Cur.nodechanges.AddObject(nch, node.id);
        }

        public static void SaveLightUndo(JedLight lt, Change change)
        {
            if (Cur == null)
                return;

            if (Cur.ltchanges.ContainsObj(lt.ID))
                return;

            LightUndo lch = new LightUndo();
            lch.SaveUndo(lt, change);
            Cur.ltchanges.AddObject(lch, lt.ID);
        }

        public static void StartUndoRec(string name)
        {
            if(!GlobalVars.UndoEnabled)
            {
                ClearUndoBuffer();
                return;
            }

            while (Stack.Count > 1000)
                Stack.RemoveAt(0);

            Cur = new UndoRec();
            Cur.Name = name;
            Stack.Add(Cur);
        }

        public static void ClearUndoBuffer()
        {
            Stack.Clear();
            Cur = null;
        }

        public static void ClearNodesUndo()
        {
            Cur = null;
            for (int i = Stack.Count - 1; i >= 0; i--)
                if (Stack[i].nodechanges.Count != 0)
                    Stack.RemoveAt(i);
        }
    }
}
