﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Smith.NED
{
    public enum VType
    {
        String,
        Int,
        Double,
        Flag,
        Pointer,
        Vector,
        Frame,
        Unknown
    }

    public enum COGType
    {
        Unknown,
        AI,
        COG,
        KEY,
        MAT,
        Msg,
        Model,
        Sector,
        Wav,
        Surface,
        Template,
        Thing,
        Int,
        Double,
        Vector
    }

    public enum AdjType
    {
        Unknown,
        MAT,
        COG,
        SND,
        PUP,
        SPR,
        PAR,
        Model,
        AI,
        Template,
        Frame
    }

    public class Value
    {
        public string name;
        public VType vType;
        public int intval;
        public double doubleval;
        public string stringval;

        public virtual void Assign(Value v)
        {
            name = v.name;
            vType = v.vType;
            intval = v.intval;
            doubleval = v.doubleval;
            stringval = v.stringval;
        }

        public virtual bool Val(string s)
        {
            switch (vType)
            {
                case VType.String:
                case VType.Unknown:
                    stringval = s;
                    return true;

                case VType.Vector:
                case VType.Frame:
                    stringval = s;
                    return true;

                case VType.Int:
                case VType.Pointer:
                    return int.TryParse(s, out intval);

                case VType.Double:
                    return double.TryParse(s, out doubleval);

                case VType.Flag:
                    try
                    {
                        intval = (int)JK.ParseFlags(s);
                        return true;
                    }
                    catch
                    {
                        return false;
                    }

                default:
                    return false;
            }
        }

        public void SetFrame(Vec3 v, Orient o)
        {
            if (!name.Equals("frame", StringComparison.InvariantCultureIgnoreCase))
                return;

            stringval = "(" + v.x.ToString("#0.000000") + "/" + v.y.ToString("#0.000000") + "/" + v.z.ToString("#0.000000") + ":" + o.pitch.ToString("#0.000000") + "/" + o.yaw.ToString("#0.000000") + "/" + o.roll.ToString("#0.000000") + ")";
        }

        public void GetFrame(out Vec3 v, out Orient o)
        {
            v = Vec3.Zero;
            o = Orient.Zero;

            if (string.IsNullOrEmpty(stringval))
                return;
            
            string[] parts = stringval.Split(new char[] { ' ', '\t', '(', ')', '/', ':' }, StringSplitOptions.RemoveEmptyEntries);
            if (parts.Length >= 3)
            {
                v.x = double.Parse(parts[0]);
                v.y = double.Parse(parts[1]);
                v.z = double.Parse(parts[2]);
            }

            if (parts.Length >= 6)
            {
                o.pitch = double.Parse(parts[3]);
                o.yaw = double.Parse(parts[4]);
                o.roll = double.Parse(parts[5]);
            }
        }

        public virtual string AsString
        {
            get
            {
                switch (vType)
                {
                    case VType.Unknown: goto case VType.String;
                    case VType.String:
                        return stringval;

                    case VType.Vector: goto case VType.Frame;
                    case VType.Frame:
                        return stringval;

                    case VType.Pointer: goto case VType.Int;
                    case VType.Int:
                        return intval.ToString();

                    case VType.Double:
                        return doubleval.ToString();

                    case VType.Flag:
                        return "0x" + intval.ToString("X");

                    default:
                        return string.Empty;
                }
            }
        }
    }

    public class Values : List<Value>
    {
        private static Dictionary<string, VType> _tplVTypes = null;
        public static Dictionary<string, VType> tplVTypes
        {
            get
            {
                if (_tplVTypes == null)
                {
                    _tplVTypes = new Dictionary<string, VType>(StringComparer.InvariantCultureIgnoreCase);
                    _tplVTypes.Add("frame", VType.Frame);
                }

                return _tplVTypes;
            }
        }

        private static Dictionary<string, AdjType> _tplNames = null;
        public static Dictionary<string, AdjType> tplNames
        {
            get
            {
                if (_tplNames == null)
                {
                    _tplNames = new Dictionary<string, AdjType>(StringComparer.InvariantCultureIgnoreCase);
                    _tplNames.Add("material", AdjType.MAT);
                    _tplNames.Add("cog", AdjType.COG);
                    _tplNames.Add("soundclass", AdjType.SND);
                    _tplNames.Add("puppet", AdjType.PUP);
                    _tplNames.Add("sprite", AdjType.SPR);
                    _tplNames.Add("particle", AdjType.PAR);
                    _tplNames.Add("model3d", AdjType.Model);
                    _tplNames.Add("aiclass", AdjType.AI);

                    _tplNames.Add("creatething", AdjType.Template);
                    _tplNames.Add("explode", AdjType.Template);
                    _tplNames.Add("fleshhit", AdjType.Template);
                    _tplNames.Add("weapon", AdjType.Template);
                    _tplNames.Add("weapon2", AdjType.Template);
                    _tplNames.Add("debris", AdjType.Template);
                    _tplNames.Add("trailthing", AdjType.Template);
                    _tplNames.Add("frame", AdjType.Frame);
                }

                return _tplNames;
            }
        }

        private static Dictionary<string, VType> _cogVTypes = null;
        public static Dictionary<string, VType> cogVTypes
        {
            get
            {
                if (_cogVTypes == null)
                {
                    _cogVTypes = new Dictionary<string, VType>(StringComparer.InvariantCultureIgnoreCase);
                    _cogVTypes.Add("flex", VType.Double);
                    _cogVTypes.Add("float", VType.Double);
                    _cogVTypes.Add("vector", VType.Vector);
                    _cogVTypes.Add("int", VType.Int);

                    _cogVTypes.Add("ai", VType.String);
                    _cogVTypes.Add("cog", VType.Int);
                    _cogVTypes.Add("keyframe", VType.String);
                    _cogVTypes.Add("material", VType.String);
                    _cogVTypes.Add("message", VType.String);
                    _cogVTypes.Add("model", VType.String);
                    _cogVTypes.Add("sector", VType.Pointer);
                    _cogVTypes.Add("sound", VType.String);
                    _cogVTypes.Add("surface", VType.Pointer);
                    _cogVTypes.Add("template", VType.String);
                    _cogVTypes.Add("thing", VType.Pointer);
                }

                return _cogVTypes;
            }
        }

        private static Dictionary<string, COGType> _cogTypes = null;
        public static Dictionary<string, COGType> cogTypes
        {
            get
            {
                if (_cogTypes == null)
                {
                    _cogTypes = new Dictionary<string, COGType>(StringComparer.InvariantCultureIgnoreCase);
                    _cogTypes.Add("ai", COGType.AI);
                    _cogTypes.Add("cog", COGType.COG);
                    _cogTypes.Add("keyframe", COGType.KEY);
                    _cogTypes.Add("material", COGType.MAT);
                    _cogTypes.Add("message", COGType.Msg);
                    _cogTypes.Add("model", COGType.Model);
                    _cogTypes.Add("sector", COGType.Sector);
                    _cogTypes.Add("sound", COGType.Wav);
                    _cogTypes.Add("surface", COGType.Surface);
                    _cogTypes.Add("template", COGType.Template);
                    _cogTypes.Add("thing", COGType.Thing);
                    _cogTypes.Add("int", COGType.Int);
                    _cogTypes.Add("float", COGType.Double);
                    _cogTypes.Add("flex", COGType.Double);
                    _cogTypes.Add("vector", COGType.Vector);
                }

                return _cogTypes;
            }
        }

        public static string GetCogTypeName(COGType ct, bool tryToUseFlex=true/*jed seems to save cog values as 'flex' rather than 'float' but our dictionary keys might be sorted differently*/)
        {
            foreach (string name in cogTypes.Keys)
                if (cogTypes[name] == ct)
                {
                    if (tryToUseFlex && name.Equals("float", StringComparison.InvariantCultureIgnoreCase))
                        continue;

                    return name;
                }

            return "unknown";
        }

        public static VType GetTPLVType(string name)
        {
            if(!tplVTypes.ContainsKey(name))
                return VType.String;

            return tplVTypes[name];
        }

        public static AdjType GetTPLType(string name)
        {
            if (!tplNames.ContainsKey(name))
                return AdjType.Unknown;

            return tplNames[name];
        }

        public static VType GetCOGVType(string name)
        {
            if (!cogVTypes.ContainsKey(name))
                return VType.Unknown;

            return cogVTypes[name];
        }

        public static COGType GetCOGType(string name)
        {
            if (!cogTypes.ContainsKey(name))
                return COGType.Unknown;

            return cogTypes[name];
        }
    }

    public class COGValue : Value
    {
        public COGType cogType;
        public bool local;
        public string desc;
        public uint mask;
        public object obj;

        public override bool Val(string s)
        {
            if (cogType == COGType.Sector || cogType == COGType.Surface || cogType == COGType.Thing)
                return int.TryParse(s, out intval);
            else
                return base.Val(s);
        }

        public bool JedVal(string s)
        {
            switch (cogType)
            {
                case COGType.Surface:
                    {
                        string[] parts = s.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                        if (parts.Length != 2)
                            return false;

                        int i, p;
                        if(!int.TryParse(parts[0], out i) ||
                           !int.TryParse(parts[1], out p))
                            return false;
                        
                        if(i < 0)
                        {
                            obj = null;
                            return true;
                        }

                        try
                        {
                            obj = GlobalVars.Level.sectors[i].surfaces[p];
                        }
                        catch
                        {
                            return false;
                        }

                        return true;
                    }

                case COGType.Sector:
                    {
                        int i;
                        if (!int.TryParse(s, out i))
                            return false;

                        if (i < 0)
                        {
                            obj = null;
                            return true;
                        }

                        if (i >= GlobalVars.Level.sectors.Count)
                            return false;

                        obj = GlobalVars.Level.sectors[i];
                        return true;
                    }

                case COGType.Thing:
                    {
                        int i;
                        if (!int.TryParse(s, out i))
                            return false;

                        if (i < 0)
                        {
                            obj = null;
                            return true;
                        }

                        if (i >= GlobalVars.Level.things.Count)
                            return false;

                        obj = GlobalVars.Level.things[i];
                        return true;
                    }

                default:
                    return Val(s);
            }
        }

        public string AsJedString
        {
            get
            {
                if (cogType == COGType.Surface)
                {
                    if (obj == null)
                        return "-1";

                    JKSurface surf = obj as JKSurface;
                    return surf.sector.num.ToString() + " " + surf.num.ToString();
                }

                return AsString;
            }
        }



        public override string AsString
        {
            get
            {
                if (cogType == COGType.Surface || cogType == COGType.Sector || cogType == COGType.Thing)
                {
                    if (obj == null)
                        return "-1";
                }

                switch (cogType)
                {
                    case COGType.Sector:
                        return (obj as JKSector).num.ToString();

                    case COGType.Surface:
                        JKSurface surf = obj as JKSurface;
                        return surf.sector.level.GetGlobalSFN(surf.sector.num, surf.sector.surfaces.IndexOf(surf)).ToString();

                    case COGType.Thing:
                        return (obj as JKThing).num.ToString();

                    default:
                        return base.AsString;
                }
            }
        }

        public void Assign(COGValue o)
        {
            name = o.name;
            cogType = o.cogType;
            vType = o.vType;
            desc = o.desc;
            local = o.local;
        }

        public static COGValue FromString(string ln)
        {
            COGValue cv = new COGValue();

            // take care of description first because it can have spaces and crap
            int desci = ln.IndexOf("desc", StringComparison.InvariantCultureIgnoreCase);
            if (desci != -1)
            {
                int descei = ln.IndexOf('=', desci);
                cv.desc = ln.Substring(descei + 1).Trim();

                ln = ln.Substring(0, desci);
            }

            // start by trimming up any whitespace to the left and right of equal signs
            // this can cover value assignments, linkids, flags...
            for (int i = 0; i < ln.Length; i++)
            {
                if (ln[i] == '=')
                {
                    string newleft = ln.Substring(0, i).TrimEnd(new char[] { ' ', '\t' });
                    string newright = ln.Substring(i+1).TrimStart(new char[] { ' ', '\t' });
                    
                    // set i to the new equal sign position. this will be incremented and then we'll go on to find the next one
                    i = newleft.Length;

                    // recombine
                    ln = newleft + "=" + newright;
                }
            }


            // split up the remaining garbage with spaces
            string[] parts = ln.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);

            // need at least 2 parts
            if (parts.Length < 2)
                return null;

            cv.cogType = Values.GetCOGType(parts[0]);
            cv.vType = Values.GetCOGVType(parts[0]);

            string[] nameparts = parts[1].Split(new char[] { '=' }, StringSplitOptions.RemoveEmptyEntries);
            cv.name = nameparts[0];
            if (nameparts.Length == 2)
                cv.Val(nameparts[1]);

            cv.local = false;

            // additional qualifiers?
            for (int i = 2; i < parts.Length; i++)
            {
                string[] whats = parts[i].Split(new char[] { '=' }, StringSplitOptions.RemoveEmptyEntries);
                switch (whats[0].ToLowerInvariant())
                {
                    case "local":
                        cv.local = true;
                        break;

                    case "mask":
                        // eat i guess?
                        break;

                    case "linkid":
                        // eat i guess?
                        break;

                    case "nolink":
                        // eat i guess?
                        break;

                    default:
                        // no idea what this is
                        break;
                }
            }

            return cv;
        }

        public void Resolve()
        {
            switch (cogType)
            {
                case COGType.Sector: obj = GlobalVars.Level.GetSectorN(intval); break;
                case COGType.Surface: obj = GlobalVars.Level.GetSurfaceN(intval); break;
                case COGType.Thing: obj = GlobalVars.Level.GetThingN(intval); break;
                default: return;
            }

            if (obj == null && intval != -1)
            {
                Utils.PanMessage(Utils.MessageType.Warning, "Cog parameter resolution failed: no " + Values.GetCogTypeName(cogType) + " num: " + intval.ToString());
            }
        }
    }

    public class COGValues : List<COGValue>
    {
        public int IndexOfName(string name)
        {
            for (int x = 0; x < Count; x++)
                if (this[x].name.Equals(name, StringComparison.InvariantCultureIgnoreCase))
                    return x;

            return -1;
        }
    }

    public class TemplateValue : Value
    {
        public AdjType adjType;

        public override void Assign(Value v)
        {
            base.Assign(v);

            if(v is TemplateValue)
                adjType = (v as TemplateValue).adjType;
        }

        public void SerializeTo(ZipUtil zip)
        {
            zip.WriteString(name);
            zip.WriteInt((int)adjType);
            zip.WriteInt((int)vType);
            zip.WriteString(AsString);
        }

        public static TemplateValue SerializeFrom(ZipUtil zip)
        {
            TemplateValue tv = new TemplateValue();
            tv.name = zip.ReadString();
            tv.adjType = (AdjType)zip.ReadInt();
            tv.vType = (VType)zip.ReadInt();
            tv.Val(zip.ReadString());
            return tv;
        }
    }

    public class TemplateValues : List<TemplateValue>
    {
        public void SerializeTo(ZipUtil zip)
        {
            zip.WriteInt(Count);
            foreach (TemplateValue tv in this)
                tv.SerializeTo(zip);
        }

        public void SerializeFrom(ZipUtil zip)
        {
            int num = zip.ReadInt();
            Clear();
            for (int x = 0; x < num; x++)
                Add(TemplateValue.SerializeFrom(zip));
        }

        public int IndexOfName(string name)
        {
            for (int x = 0; x < Count; x++)
                if (this[x].name.Equals(name, StringComparison.InvariantCultureIgnoreCase))
                    return x;

            return -1;
        }
    }
}
