﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Runtime.InteropServices;
using System.Threading;

namespace Smith.NED
{
    public partial class NedMain : GUI.BaseForm
    {
        public enum MapMode
        {
            Sector,
            Surface,
            Vertex,
            Thing,
            Edge,
            Light,
            Frame,
            Extra
        }

        public enum MouseMode
        {
            Select,
            Cleave,
            Drag,
            TranslateCam,
            RotateCam,
            TranslateGrid,
            RotateGrid,
            CreateSector,
            RectSelect
        }

        public const int AllLayers = -100;

        public enum ShiftTextureType
        {
            Left,
            Right,
            Up,
            Down
        }

        public enum RaiseObjectType
        {
            Up,
            Down
        }

        [Flags]
        public enum SectorChange
        {
            Vertices = 0x01,
            Values = 0x02,
            All = Vertices|Values
        }

        [Flags]
        public enum ToolAxis
        {
            TX = 0x01,
            X = 0x02,
            Y = 0x04,
            Z = 0x08,
            Grid = 0x10
        }

        public enum ThingView
        {
            Dots,
            Boxes,
            Wireframes
        }

        public enum MultiselectMode
        {
            Toggle,
            Add,
            Subtract
        }

        public class ExtraLine
        {
            public Vec3 v1;
            public Vec3 v2;
            public string Name;

        }

        public class ExtraVertex
        {
            public Vec3 v;
            public string Name;
        }

        public class ExtraPolygon
        {
            public Poly3 poly;
            public string Name;
        }

        public JKLevel Level
        {
            get
            {
                return GlobalVars.Level;
            }
        }

        //bool launchedAsMots = false;
        //string launchedPath = null;

        public static void Init(string[] cmdlineargs)
        {
            // load settings first since creating NED wants to know about window positions and stuff
            //Options.LoadSettings();



            string path = null;
            bool mots = false;


            // clean up surrounding quotes from all args
            string[] cmd = new string[cmdlineargs.Length];//Environment.GetCommandLineArgs();
            Array.Copy(cmdlineargs, cmd, cmdlineargs.Length);
            for (int x = 0; x < cmd.Length; x++)
                cmd[x] = cmd[x].Trim(new char[] { '\"' });

            // parse commandline args
            for (int ci = 1; ci < cmd.Length;)
            {
                switch (cmd[ci++].ToLowerInvariant())
                {
                    case "-path":
                        {
                            string tmp = cmd[ci++];

                            if (!string.IsNullOrEmpty(tmp) && tmp.Length >= 2 && tmp[1] == ':')
                                path = tmp;
                            //else
                            //    path = Path.Combine(jkpath, tmp);
                        }
                        break;

                    case "-mots":
                        mots = true;
                        break;
                }
            }

            

            NED.GlobalVars.CreateNED();

            Options.SetMOTS(mots);

            if (!string.IsNullOrEmpty(path))
            {
                Application.DoEvents();// evil but necessary
                GlobalVars.NedMain.performopen(path);
            }
        }

        public static void Shutdown()
        {
            Options.SaveSettings(true);

            NED.GlobalVars.DestroyNED();
        }

        public void RaiseObject(RaiseObjectType how)
        {
            Vec3 d;
            if (how == RaiseObjectType.Up)
                d = renderer.GridNormal * GlobalVars.PerpStep;
            else
                d = -renderer.GridNormal * GlobalVars.PerpStep;

            switch (map_mode)
            {
                case MapMode.Sector:
                    Undo.StartUndoRec("Translate sector(s)");
                    Utils.TranslateSectors(Level, scsel, Cur_SC, d);
                    break;

                case MapMode.Surface:
                    Undo.StartUndoRec("Translate surface(s)");
                    Utils.TranslateSurfaces(Level, sfsel, Cur_SC, Cur_SF, d);
                    break;

                case MapMode.Vertex:
                    Undo.StartUndoRec("Translate vertices");
                    Utils.TranslateVertices(Level, vxsel, Cur_SC, Cur_VX, d);
                    break;

                case MapMode.Edge:
                    Undo.StartUndoRec("Translate Edge(s)");
                    Utils.TranslateEdges(Level, edsel, Cur_SC, Cur_SF, Cur_ED, d);
                    break;

                case MapMode.Thing:
                    Undo.StartUndoRec("Translate thing(s)");
                    Utils.TranslateThings(Level, thsel, Cur_TH, d, GlobalVars.MoveFrames);
                    break;

                case MapMode.Frame:
                    Undo.StartUndoRec("Translate frame(s)");
                    if (Cur_FR == -1)
                        Utils.TranslateThingKeepFrame(Level, Cur_TH, d);
                    else
                        Utils.TranslateFrames(Level, frsel, Cur_TH, Cur_FR, d);
                    break;

                case MapMode.Light:
                    Undo.StartUndoRec("Translate light(s)");
                    Utils.TranslateLights(Level, ltsel, Cur_LT, d);
                    break;

                default:
                    return;
            }
        }

        public void ShiftTexture(ShiftTextureType how)
        {
            if (map_mode != MapMode.Surface)
                SetMapMode(MapMode.Surface);

            Vec2 d;
            switch (how)
            {
                case NedMain.ShiftTextureType.Left:
                    d = new Vec2(GlobalVars.TXStep, 0.0);
                    break;

                case NedMain.ShiftTextureType.Right:
                    d = new Vec2(-GlobalVars.TXStep, 0.0);
                    break;

                case NedMain.ShiftTextureType.Up:
                    d = new Vec2(0.0, -GlobalVars.TXStep);
                    break;

                case NedMain.ShiftTextureType.Down:
                    d = new Vec2(0.0, GlobalVars.TXStep);
                    break;

                default:
                    return;
            }

            Undo.StartUndoRec("Shift texture");

            int n = sfsel.AddSF(Cur_SC, Cur_SF);
            for (int i = 0; i < sfsel.Count; i++)
            {
                int sc, sf;
                sfsel.GetSCSF(n, out sc, out sf);

                JKSurface surf = Level.sectors[sc].surfaces[sf];
                {
                    Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Geo);
                    foreach (TexVertex tv in surf.texvertices)
                        tv.v += d;
                    SectorChanged(surf.sector, false);
                }
            }
            sfsel.DeleteN(n);
        }

        public void RotateTexture(ShiftTextureType how)
        {
            int fv = 0;
            switch (map_mode)
            {
                case MapMode.Edge:
                    fv = Cur_ED;
                    break;

                case MapMode.Surface:
                    fv = 0;
                    break;

                default:
                    SetMapMode(MapMode.Surface);
                    break;
            }

            double cosa, sina;
            switch (how)
            {
                case ShiftTextureType.Left:
                    cosa = Math.Cos(MathLib.ToRad(-GlobalVars.TXRotStep));
                    sina = Math.Sin(MathLib.ToRad(-GlobalVars.TXRotStep));
                    break;

                case ShiftTextureType.Right:
                    cosa = Math.Cos(MathLib.ToRad(GlobalVars.TXRotStep));
                    sina = Math.Sin(MathLib.ToRad(GlobalVars.TXRotStep));
                    break;

                default:
                    return;
            }

            Undo.StartUndoRec("Rotate texture");

            JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];

            Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Geo);

            Vec3 u, v;
            surf.CalcUVNormals(out u, out v);

            Vec3 nv = new Vec3(-sina * u.x + cosa * v.x,
                               -sina * u.y + cosa * v.y,
                               -sina * u.z + cosa * v.z);

            Vec3 nu = new Vec3(cosa * u.x + sina * v.x,
                                cosa * u.y + sina * v.y,
                                cosa * u.z + sina * v.z);

            surf.UpdateSurfUVData(nu, nv);
            surf.ArrangeTexture(fv, nu, nv);
            SectorChanged(surf.sector, false);
        }

        public void ScaleTexture(ShiftTextureType how)
        {
            if (map_mode != MapMode.Surface)
                SetMapMode(MapMode.Surface);

            Undo.StartUndoRec("Scale texture");

            int n = sfsel.AddSF(Cur_SC, Cur_SF);
            for (int i = 0; i < sfsel.Count; i++)
            {
                int sc, sf;
                sfsel.GetSCSF(i, out sc, out sf);
                {
                    JKSurface surf = Level.sectors[sc].surfaces[sf];
                    Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Geo);

                    switch (how)
                    {
                        case ShiftTextureType.Up:
                            surf.texScale *= GlobalVars.TXScaleStep;
                            break;

                        case ShiftTextureType.Down:
                            surf.texScale /= GlobalVars.TXScaleStep;
                            break;

                        default:
                            return;
                    }

                    Vec3 u, v;
                    surf.CalcUVNormals(out u, out v);
                    surf.ArrangeTexture(0, u, v);
                    SectorChanged(surf.sector, false);
                }
            }
            sfsel.DeleteN(n);
        }

        public void StraightenTexture(bool zero, bool rot90)
        {
            int n;
            switch (map_mode)
            {
                case MapMode.Surface:
                    Undo.StartUndoRec("Straighten surfaces");
                    n = sfsel.AddSF(Cur_SC, Cur_SF);
                    for (int i = 0; i < sfsel.Count; i++)
                    {
                        int sc, sf;
                        sfsel.GetSCSF(i, out sc, out sf);
                        {
                            JKSurface surf = Level.sectors[sc].surfaces[sf];
                            Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Geo);

                            Vec3 u, v;
                            if (rot90)
                            {
                                surf.CalcUVNormals(out v, out u);
                                v = -v;
                                surf.UpdateSurfUVData(u, v);
                            }
                            else
                                surf.CalcUVNormals(out u, out v);

                            if (zero)
                                surf.texvertices[0].v = Vec2.Zero;

                            surf.ArrangeTexture(0, u, v);
                            SectorChanged(surf.sector, false);
                        }
                    }
                    sfsel.DeleteN(n);
                    break;

                case MapMode.Edge:
                    Undo.StartUndoRec("Straighten surfaces");
                    n = edsel.AddED(Cur_SC, Cur_SF, Cur_ED);
                    for (int i = 0; i < edsel.Count; i++)
                    {
                        int sc, sf, ed;
                        edsel.GetSCSFED(i, out sc, out sf, out ed);
                        {
                            JKSurface surf = Level.sectors[sc].surfaces[sf];

                            Vec3 u, v;
                            if (rot90)
                            {
                                surf.CalcDefaultUVNormals(ed, out v, out u);
                                v = -v;
                                surf.UpdateSurfUVData(u, v);
                            }
                            else
                                surf.CalcDefaultUVNormals(ed, out u, out v);

                            if (zero)
                                surf.texvertices[0].v = Vec2.Zero;

                            surf.ArrangeTexture(ed, u, v);
                            SectorChanged(surf.sector, false);
                        }
                    }
                    edsel.DeleteN(n);
                    break;
            }
        }

        public void BringThingToSurf(int th, int sc, int sf)
        {
            Undo.StartUndoRec("Bring thing to Surface");
            JKSurface surf = Level.sectors[sc].surfaces[sf];
            Vec3 c = Utils.CalcSurfCenter(surf);
            JKThing thing = Level.things[th];

            Undo.SaveThingUndo(thing, Undo.Change.Changed);
            double l = -thing.box.Bottom;
            c += surf.normal * l;
            thing.pos = c;

            Vertex v1 = surf.vertices[0];
            Vertex v2 = surf.vertices[1];
            Vec3 zv = surf.normal;
            Vec3 yv = (v2.v - v1.v).Unit;
            Vec3 xv = yv ^ zv;

            thing.orient = Utils.GetJKPYR(xv, yv, zv);

            ThingChanged(thing);
        }

        public void AddThingAtSurf()
        {
            Undo.StartUndoRec("Add thing(s)");

            JKThing th = Level.NewThing();
            if (Cur_TH < Level.things.Count)
                th.Assign(Level.things[Cur_TH]);
            int n = Level.things.Count;
            Level.things.Add(th);
            ThingAdded(th);
            Level.RenumThings();

            BringThingToSurf(n, Cur_SC, Cur_SF);

            LayerThing(n);
            Undo.SaveThingUndo(th, Undo.Change.Added);
            SetCurTH(n);
        }

        public void AddThingAtXYZPYR(Vec3 p, Orient o)
        {
            Undo.StartUndoRec("Add thing(s)");

            JKThing th = Level.NewThing();
            if (Cur_TH < Level.things.Count)
                th.Assign(Level.things[Cur_TH]);
            int n = Level.things.Count;
            Level.things.Add(th);
            ThingAdded(th);
            Level.RenumThings();

            th.pos = p;
            th.orient = o;

            LayerThing(n);
            Undo.SaveThingUndo(th, Undo.Change.Added);
            SetCurTH(n);
        }

        JKSector stitch_sec = null;
        JKSurface stitch_surf = null;

        public void StartStitch()
        {
            if (map_mode != MapMode.Surface)
                return;

            stitch_sec = Level.sectors[Cur_SC];
            stitch_surf = Level.sectors[Cur_SC].surfaces[Cur_SF];
        }

        public void DoStitch()
        {
            if (map_mode != MapMode.Surface)
                return;

            if (stitch_sec == null || stitch_surf == null)
            {
                Utils.PanMessage(Utils.MessageType.Error, "No start surface was set for stitching");
                return;
            }

            if (Level.sectors.IndexOf(stitch_sec) < 0 || stitch_sec.surfaces.IndexOf(stitch_surf) < 0)
            {
                Utils.PanMessage(Utils.MessageType.Error, "The stitching start surface was deleted");
                return;
            }

            JKSurface ssurf = stitch_surf;
            JKSurface esurf = Level.sectors[Cur_SC].surfaces[Cur_SF];

            Undo.StartUndoRec("Stitch surfaces");
            Utils.StitchSurfaces(ssurf, esurf);
            if (!string.IsNullOrEmpty(ssurf.material))
                esurf.material = ssurf.material;

            SectorChanged(esurf.sector, false);
            StartStitch();

            // update UI
            GlobalVars.NedMain.EditObject();
        }

        private int LightCalcHalts = 0;

        public void HaltLightCalc()
        {
            if (LightCalcHalts == 0)
            {
                // we need to stop (synchronously)
                disableLightCalc = true;

                if (calcingLights && calcLightThread != null)
                {
                    PNEDStatus.Text = "Stopping light calc...";
                    statusStrip1.Update();

                    while (calcingLights)
                        System.Threading.Thread.Sleep(10);

                    PNEDStatus.Text = "Halted";
                    statusStrip1.Update();
                }
            }

            LightCalcHalts++;
        }

        public void ResumeLightCalc()
        {
            LightCalcHalts--;

            if (LightCalcHalts == 0)
            {
                // we can resume
                disableLightCalc = false;
            }
        }

        private bool? _AllSectors3DOPreview = null;
        public bool AllSectors3DOPreview
        {
            get
            {
                if (!_AllSectors3DOPreview.HasValue)
                {
                    _AllSectors3DOPreview = true;
                    foreach(JKSector sec in Level.sectors)
                        if ((sec.flags & Sector.Flag.Preview3DO) == 0)
                        {
                            _AllSectors3DOPreview = false;
                            break;
                        }
                }

                return _AllSectors3DOPreview.Value;
            }
        }

        public void NewProject()
        {
            HaltLightCalc();

            Level.Clear();

            JKSector sec = Level.NewSector();
            Level.sectors.Add(sec);

            sec.NewVertex(-1.0, 1.0, 0.0);
            sec.NewVertex(-1.0, -1.0, 0.0);
            sec.NewVertex(1.0, -1.0, 0.0);
            sec.NewVertex(1.0, 1.0, 0.0);

            sec.NewVertex(-1.0, 1.0, 1.0);
            sec.NewVertex(1.0, 1.0, 1.0);
            sec.NewVertex(1.0, -1.0, 1.0);
            sec.NewVertex(-1.0, -1.0, 1.0);

            JKSurface surf;

            surf = sec.NewSurface(); sec.surfaces.Add(surf);
            surf.surfaceFlags |= Surface.Flag.Floor;
            surf.AddVertices(sec.GetVertices(new int[] { 0, 1, 2, 3 }));

            surf = sec.NewSurface(); sec.surfaces.Add(surf);
            surf.AddVertices(sec.GetVertices(new int[] { 4, 5, 6, 7 }));

            surf = sec.NewSurface(); sec.surfaces.Add(surf);
            surf.AddVertices(sec.GetVertices(new int[] { 2, 6, 5, 3 }));

            surf = sec.NewSurface(); sec.surfaces.Add(surf);
            surf.AddVertices(sec.GetVertices(new int[] { 1, 7, 6, 2 }));

            surf = sec.NewSurface(); sec.surfaces.Add(surf);
            surf.AddVertices(sec.GetVertices(new int[] { 0, 4, 7, 1 }));

            surf = sec.NewSurface(); sec.surfaces.Add(surf);
            surf.AddVertices(sec.GetVertices(new int[] { 3, 5, 4, 0 }));


            JKThing thing = Level.NewThing(); Level.things.Add(thing);
            thing.pos = new Vec3(0.0, 0.0, 0.1);
            thing.name = "walkplayer";
            thing.sec = Level.sectors[0];

            JedLight light = Level.NewLight(); Level.lights.Add(light);
            light.Pos = new Vec3(0.0, 0.0, 0.5);

            foreach (JKSurface s in sec.surfaces)
                s.RecalcAll();

            sec.Renumber();

            Level.SetDefaultHeader();

            AskSaveAs = true;
            NewLevel = true;
            _AllSectors3DOPreview = null;
            LevelFile = "Untitled.jed";
            Level.AddMissingLayers();
            Level.masterCMP = string.Empty;
            Level.mots = isMots;

            lightsDirty = false;

            GlobalVars.NedMain?.ResumeLightCalc();
        }

        public bool AskSaveAs = true;
        public bool NewLevel = true;
        public string LevelFile = "Untitled.jed";
        public bool isMots = false;

        public enum OpenMethod
        {
            Open,
            Revert
        }

        void ResettingEditor()
        {
            MsgForm.Clear();
        }

        void SetLevelName()
        {
            Text = "Ned - " + LevelFile;

            if (AskSaveAs)
                GlobalVars.ProjectDir = string.Empty;
            else
                GlobalVars.ProjectDir = FileOperations.ExtractPath(LevelFile);

            if (FileOperations.ExtractExt(LevelFile).Equals(".jkl", StringComparison.InvariantCultureIgnoreCase))
            {
                int l = GlobalVars.ProjectDir.Length;
                if (l >= 4 && GlobalVars.ProjectDir.Substring(l - 4).Equals(@"jkl\", StringComparison.InvariantCultureIgnoreCase))
                    GlobalVars.ProjectDir = GlobalVars.ProjectDir.Substring(0, l - 4);
            }

            string projdir = GlobalVars.ProjectDir;
            int i = projdir.IndexOf('>');
            if (i != -1)
            {
                Log.Warning("UH OH.. THIS SHOULNDT HAPPENN");

                projdir = projdir.Substring(0, i);
                i = projdir.LastIndexOf('.');
                if (i != -1)
                    projdir = projdir.Substring(0, i);
            }
            if (!string.IsNullOrEmpty(projdir) && Directory.Exists(projdir))
                Directory.SetCurrentDirectory(projdir);
        }

        public void AddRecent(string fname)
        {
            List<string> recents = new List<string>(new string[] { GlobalVars.Recent1, GlobalVars.Recent2, GlobalVars.Recent3, GlobalVars.Recent4 });
            recents.Remove(fname);

            recents.Insert(0, fname);

            GlobalVars.Recent1 = recents[0];
            GlobalVars.Recent2 = recents[1];
            GlobalVars.Recent3 = recents[2];
            GlobalVars.Recent4 = recents[3];

            SyncRecents();
        }

        public void SyncRecents()
        {
            int im = FileMenu.DropDownItems.IndexOf(RecentBar);
            while (FileMenu.DropDownItems.Count > (im + 1))
                FileMenu.DropDownItems.RemoveAt(im + 1);

            foreach (string s in new string[] { GlobalVars.Recent1, GlobalVars.Recent2, GlobalVars.Recent3, GlobalVars.Recent4 })
            {
                if (string.IsNullOrEmpty(s))
                    continue;

                ToolStripMenuItem tsi = new ToolStripMenuItem();

                if (s.Length > 32)
                {
                    string dirname = FileOperations.ExtractPath(s);
                    dirname = dirname.Substring(0, Math.Min(28, dirname.Length));
                    tsi.Text = dirname + @"...\" + FileOperations.ExtractName(s);
                }
                else
                    tsi.Text = s;

                tsi.ToolTipText = s;
                tsi.Click += new EventHandler(Recent_Click);

                FileMenu.DropDownItems.Add(tsi);
            }
        }

        void Recent_Click(object sender, EventArgs e)
        {
            string filepath = (sender as ToolStripMenuItem).ToolTipText;
            if (string.IsNullOrEmpty(filepath) || (!FileOperations.IsInContainer(filepath) && !File.Exists(filepath)))
                return;

            if ((ModifierKeys & Keys.Control) != 0)
            {
                string str = $"-ned -path \"{filepath}\"";
                Clipboard.SetText(str);
                MessageBox.Show($"{str}\ncopied to clipboard");
                return;
            }

            if (!AskSave())
                return;

            OpenProject(filepath, OpenMethod.Open);
        }

        public NedMain()
        {
            NewProject();

#if false
            string[] lfms = Directory.GetFiles(@"c:\ned", "*.lfm");
            foreach (string lfm in lfms)
            {
                string txt = File.ReadAllText(lfm);

                for(int i=0; ;)
                {
                    i = txt.IndexOf("Glyph.Data", i, StringComparison.InvariantCultureIgnoreCase);
                    if (i == -1)
                        break;


                    int obji = txt.LastIndexOf("object", i);
                    obji += 7;
                    int col = txt.IndexOf(':', obji);
                    string objname = txt.Substring(obji, col - obji);


                    i = txt.IndexOf('{', i);
                    int end = txt.IndexOf('}', i);

                    string glyphcontent = txt.Substring(i + 1, end-i-1);
                    for(;;)
                    {
                        int c = glyphcontent.IndexOfAny(new char[] { (char)10, (char)13, ' ', '\t' });
                        if(c == -1)
                            break;
                        glyphcontent = glyphcontent.Remove(c, 1);
                    }

                    MemoryStream ms = new MemoryStream();
                    for (int x = 0; x < glyphcontent.Length; x += 2)
                    {
                        byte b = byte.Parse(glyphcontent.Substring(x, 2), System.Globalization.NumberStyles.HexNumber);
                        ms.WriteByte(b);
                    }

                    byte[] buf = new byte[ms.Position];
                    Array.Copy(ms.GetBuffer(), buf, ms.Position);

                    File.WriteAllBytes(@"c:\smith\glyphs\" + objname + ".bmp", buf);

                    i = end;
                }
            }
#endif


            InitializeComponent();


            BNSC.Image = ImportGlyph(Properties.Resources.BNSC);
            BNSF.Image = ImportGlyph(Properties.Resources.BNSF);
            BNED.Image = ImportGlyph(Properties.Resources.BNED);
            BNVX.Image = ImportGlyph(Properties.Resources.BNVX);
            BNTH.Image = ImportGlyph(Properties.Resources.BNTH);
            BNFR.Image = ImportGlyph(Properties.Resources.BNFR);
            BNLT.Image = ImportGlyph(Properties.Resources.BNLT);
            BNEX.Image = ImportGlyph(Properties.Resources.BNEX);
        }

        Bitmap ImportGlyph(Bitmap b)
        {
            Bitmap bmp = new Bitmap(b.Height, b.Height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
            for(int y=0; y<b.Height; y++)
                for (int x = 0; x < b.Height; x++)
                {
                    Color clr = b.GetPixel(x, y);

                    if (clr.R == 128 && clr.G == 128 && clr.B == 0)
                        clr = Color.Transparent;

                    bmp.SetPixel(x, y, clr);
                }
            return bmp;
        }

        bool updatestuff = false;

        public void UpdateItemEditor()
        {
            updatestuff = true;
        }

        public void VerifySelection()
        {
            if (!IsSelValid)
                SetMapMode(map_mode);
        }

        long lastTimestamp = 0;
        public void Process()
        {
            long curtimestamp = PerfTimer.Timestamp;
            double truedt = PerfTimer.TimeBetween(lastTimestamp, curtimestamp);//curtime.Subtract(lastProcess).TotalSeconds;
            if (truedt <= 0.0)
                return;

            double dt;
            if (truedt > (1.0 / 20.0))      // frame min..
                dt = 1.0 / 20.0;
            else
                dt = truedt;
            lastTimestamp = curtimestamp;


            VerifySelection();
            VerifyMultiSelection();

            if (updatestuff)
            {
                updatestuff = false;
                EditObject();
            }


            // teleport smith player?
            if (SmithMain.SmithRunning && GlobalVars.SmithSyncCameras)
            {
                if (NedFocused && GlobalVars.Preview3D.Visible && GlobalVars.Preview3D.renderer != null)
                {
                    if (Game.LocalPlayer != null)
                    {
                        Game.LocalPlayer.Teleport(GlobalVars.Preview3D.renderer.CamPos);
                        Game.LocalPlayer.SetAimPitch(-GlobalVars.Preview3D.renderer.CamOrient.pitch);
                        Game.LocalPlayer.SetAimYaw(-GlobalVars.Preview3D.renderer.CamOrient.yaw);
                    }

                    // neo mode while ned preview is up
                    if(!Program.MainForm.WasNeo.HasValue)
                        Program.MainForm.WasNeo = Program.MainForm.neo.Checked;
                    Program.MainForm.neo.Checked = true;
                }
                else
                {
                    // restore neo mode
                    if (Program.MainForm.WasNeo.HasValue)
                    {
                        Program.MainForm.neo.Checked = Program.MainForm.WasNeo.Value;
                        Program.MainForm.WasNeo = null;
                    }
                }
            }


            // ensure that thread either is or isnt running
            if (GlobalVars.CalculateLightsAuto)
            {
                wantCalcLight = true;
                if (calcLightThread == null)
                {
                    calcLightThread = new Thread(new ThreadStart(CalcLightThread));
                    calcLightThread.Start();
                }

                if (calcingLights && !disableLightCalc)
                {
                    PNEDStatus.Text = "Calculating light " + calcLightProgress.ToString() + "%...";
                }

                // is a dataset finished?
                if (lightsFinished)
                {
                    lightsFinished = false;

                    PNEDStatus.Text = "Idle";

                    //if (Level.sectors.Count == scs.Count)
                        //GlobalVars.Preview3D.ReloadLevel(true);
                    //else
                      //  for (int i = 0; i < scs.Count; i++)
                        //    SectorChanged(Level.sectors[scs.GetSC(i)]);

                    foreach (JKSector sec in Level.sectors)
                        SectorChanged(sec, false/*dont fire off another cycle*/);

                    LevelChanged(false/*dont fire another cycle*/);
                }
            }
            else
            {
                wantCalcLight = false;
                if (calcLightThread != null)
                {
                    // just forget about it.. it will die on its own
                    calcLightThread = null;
                }

                if (!string.IsNullOrEmpty(PNEDStatus.Text))
                    PNEDStatus.Text = string.Empty;
            }


            if(renderer != null)
                renderer.Process(dt);

            Redraw();
        }

        private volatile bool disableLightCalc = false;

        public bool WantLightCalcCancel
        {
            get
            {
                return disableLightCalc || lightsDirty;
            }
        }

        volatile bool calcingLights = false;

        void CalcLightThread()
        {
            while (wantCalcLight)
            {
                if (!NedRunning)
                {
                    wantCalcLight = false;
                    break;
                }

                if (!disableLightCalc)
                {
                    if (lightsDirty)
                    {
                        calcLightProgress = 0;
                        calcingLights = true;
                        lightsDirty = false;

                        SCMultiSel scs = new SCMultiSel();
                        for (int x = 0; x < Level.sectors.Count; x++)
                            scs.AddSC(x);

                        if (GlobalVars.NewLightCalc)
                            Utils.CalcLightingNew(Level, scs, true);
                        else
                            Utils.CalcLighting(Level, scs, true);

                        // if lights were re-dirtied, dont keep this result
                        if (lightsDirty)
                        {
                            calcLightProgress = 0;
                            lightsFinished = false;
                        }
                        else
                        {
                            calcLightProgress = 100;
                            lightsFinished = true;
                        }

                        calcingLights = false;
                    }
                }
                else
                {
                    //lightsDirty = true;// ensure we start again when allowed to resume    // we shouldnt have to do this since anything should still be able to set the dirty flag
                    lightsFinished = false;
                }

                System.Threading.Thread.Sleep(350);// check periodically
            }

            calcLightThread = null;
        }

        volatile Thread calcLightThread = null;
        volatile bool wantCalcLight = false;

        public volatile bool lightsDirty = false;
        public volatile bool lightsFinished = false;
        public volatile int calcLightProgress = 0;

        public void Redraw()
        {
            if (renderer == null)
                return;

            if (!Visible)
                return;

            using (renderer.Context)
            {
                bool smooth = true;// (ModifierKeys & Keys.Shift) == 0;

                if (smooth)
                {
                    gl.Enable(gl.LINE_SMOOTH);
                    gl.Enable(gl.POINT_SMOOTH);
                    gl.Enable(gl.POLYGON_SMOOTH);
                }
                else
                {
                    gl.Disable(gl.LINE_SMOOTH);
                    gl.Disable(gl.POINT_SMOOTH);
                    gl.Disable(gl.POLYGON_SMOOTH);
                }

                renderer.PushFBO(renderer.fbo, renderer, Width, Height, smooth ? 4 : 1);

                renderer.SetViewport(renderControl1.ClientRectangle.Width, renderControl1.ClientRectangle.Height);

                gl.ClearColor(GlobalVars.clMapBack);
                gl.Clear(gl.DEPTH_BUFFER_BIT | gl.COLOR_BUFFER_BIT);
                renderer.PushStencil(Renderer.Stencil.Off);
                renderer.PushBlend(Renderer.Blend.Normal);
                renderer.PushTextureEnable(true);


                //RenderWorld(visCam, Mat4.Identity, dt, truedt, camsec, false);
                //RenderGUI(visCam, dt, truedt, lastFramePerfReport);


                gl.Disable(gl.DEPTH_TEST);
                gl.BlendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);



                renderer.BeginScene();
                DrawInternal();
                renderer.EndScene();

                if(true)
                {
                    
                    gl.MatrixMode(gl.PROJECTION);
                    gl.LoadMatrixd(Mat4.Ortho(new Box2(renderControl1.ClientRectangle)));
                    gl.MatrixMode(gl.MODELVIEW);
                    gl.LoadIdentity();

                    /*Box2 rc = new Box2(renderControl1.ClientRectangle);
                    renderer.DrawLine(rc.FromUL(0.25, 0.25), rc.FromUL(0.7, 0.6), Color4.white());


                    gl.Begin(gl.POINTS);
                    gl.Vertex(rc.FromUL(0.1, 0.4));
                    gl.End();*/

                    DrawGUI();
                }

                RendererBase.FinishFBOStruct ffs = new RendererBase.FinishFBOStruct();
                ffs.resolveOverride = 0;
                renderer.PopFBO(ffs);

                renderer.PopBlend();
                renderer.PopTextureEnable();
                renderer.PopStencil();

                gl.SwapBuffers(renderer.hDC);
            }
        }

        void DrawGUI()
        {
            if (CleaveStarted || SCCreateStarted)
            {
                renderer.DrawLine(CleaveOrgXY, DrawToXY, Color4.red());
            }

            if (RectSelStarted)
            {
                Box2 rc = Box2.Expandable;
                rc.Expand(RectOrgXY);
                rc.Expand(RectLastXY);
                renderer.DrawRectangle(rc, Color4.red());
            }
        }

        void DrawThing(JKThing thing)
        {
            DrawThingAt(thing, thing.pos, thing.orient);
        }

        void DrawThingAt(JKThing thing, Vec3 pos, Orient orient)
        {
            renderer.DrawVertex(pos);

            switch (GlobalVars.ThingView)
            {
                case ThingView.Boxes:
                    if (thing.box.HasArea)
                        renderer.DrawPolygonsAt(thing.box.Faces, pos, orient);
                    break;

                case ThingView.Wireframes:
                    if (thing.model != null)
                    {
                        foreach (JKModel.Mesh mesh in thing.model.meshes)
                            renderer.DrawPolygonsAt(mesh.faces, pos, orient);
                    }
                    break;
            }
        }

        private void DrawInternal()
        {
            gl.PointSize(1.0f);
            renderer.SetColor(GlobalVars.clGrid);
            renderer.DrawGrid();

            if (map_mode == MapMode.Vertex)
            {
                gl.PointSize(4.0f);
                renderer.SetColor(GlobalVars.clVertex);
                foreach (JKSector sec in Level.sectors)
                    if (Toolbar.IsLayerVisible(sec.layer))
                        renderer.DrawVertices(sec.vertices);
                gl.PointSize(1.0f);
            }


            renderer.SetColor(GlobalVars.clMapGeoBack);
            renderer.SetCulling(Renderer.CullMode.Back);
            foreach (JKSector sec in Level.sectors)
                if (Toolbar.IsLayerVisible(sec.layer))
                    renderer.DrawPolygons(sec.surfaces);

            renderer.SetColor(GlobalVars.clMapGeo);
            renderer.SetCulling(Renderer.CullMode.Front);
            foreach (JKSector sec in Level.sectors)
                if (Toolbar.IsLayerVisible(sec.layer))
                    renderer.DrawPolygons(sec.surfaces);

            renderer.SetCulling(Renderer.CullMode.None);
            gl.PointSize(4.0f);

            if ((!ThingsHidden && !AllSectors3DOPreview) || map_mode == MapMode.Thing)
            {
                renderer.SetColor(GlobalVars.clThing);
                foreach (JKThing thing in Level.things)
                    if (Toolbar.IsLayerVisible(thing.layer))
                        DrawThing(thing);
            }

            if (!LightsHidden || map_mode == MapMode.Light)
            {
                renderer.SetColor(GlobalVars.clLight);
                foreach (JedLight light in Level.lights)
                    if (Toolbar.IsLayerVisible(light.layer))
                        renderer.DrawVertex(light.Pos);
            }

            if (map_mode == MapMode.Frame)
            {
                renderer.SetColor(GlobalVars.clFrame);
                renderer.SetPointSize(4.0);
                foreach (JKThing thing in Level.things)
                {
                    if (!Toolbar.IsLayerVisible(thing.layer))
                        continue;

                    renderer.DrawVertex(thing.pos);
                    foreach (TemplateValue tv in thing.vals)
                        DrawFrame(tv);
                }
            }

            renderer.SetColor(GlobalVars.clExtra);
            renderer.SetCulling(Renderer.CullMode.None);
            foreach (object obj in ExtraObjs)
            {
                if (obj is Vertex)
                    renderer.DrawVertex((obj as Vertex).v);
                else if (obj is ExtraLine)
                    renderer.DrawLine((obj as ExtraLine).v1, (obj as ExtraLine).v2);
                else if (obj is Polygon)
                    renderer.DrawPolygon(obj as Polygon);
            }

            // draw the 3d preview camera position
            if (GlobalVars.Preview3D.renderer != null)
                renderer.DrawCircle(GlobalVars.Preview3D.renderer.CamPos, 0.1);

            switch (map_mode)
            {
                case MapMode.Sector:
                    renderer.SetColor(GlobalVars.clMSelBack);
                    renderer.SetCulling(Renderer.CullMode.Back);
                    for (int s = 0; s < scsel.Count; s++)
                    {
                        JKSector sec = Level.sectors[scsel.GetSC(s)];
                        renderer.DrawPolygons(sec.surfaces);
                    }
                    renderer.SetColor(GlobalVars.clMSel);
                    renderer.SetCulling(Renderer.CullMode.Front);
                    for (int s = 0; s < scsel.Count; s++)
                    {
                        JKSector sec = Level.sectors[scsel.GetSC(s)];
                        renderer.DrawPolygons(sec.surfaces);
                    }
                    break;

                case MapMode.Surface:
                    renderer.SetColor(GlobalVars.clMSelBack);
                    renderer.SetCulling(Renderer.CullMode.Back);
                    for (int s = 0; s < sfsel.Count; s++)
                    {
                        int sc, sf;
                        sfsel.GetSCSF(s, out sc, out sf);
                        JKSurface surf = Level.sectors[sc].surfaces[sf];
                        renderer.DrawPolygon(surf);
                    }

                    renderer.SetColor(GlobalVars.clMSel);
                    renderer.SetCulling(Renderer.CullMode.Front);
                    for (int s = 0; s < sfsel.Count; s++)
                    {
                        int sc, sf;
                        sfsel.GetSCSF(s, out sc, out sf);
                        JKSurface surf = Level.sectors[sc].surfaces[sf];
                        renderer.DrawPolygon(surf);
                    }
                    break;

                case MapMode.Edge:
                    renderer.SetColor(GlobalVars.clMSelBack);
                    renderer.SetCulling(Renderer.CullMode.Back);
                    for (int s = 0; s < edsel.Count; s++)
                    {
                        int sc, sf, ed;
                        edsel.GetSCSFED(s, out sc, out sf, out ed);
                        JKSurface surf = Level.sectors[sc].surfaces[sf];
                        renderer.DrawLine(surf.vertices[ed].v, surf.vertices[surf.NextVX(ed)].v);
                    }

                    renderer.SetColor(GlobalVars.clMSel);
                    renderer.SetCulling(Renderer.CullMode.Front);
                    for (int s = 0; s < edsel.Count; s++)
                    {
                        int sc, sf, ed;
                        edsel.GetSCSFED(s, out sc, out sf, out ed);
                        JKSurface surf = Level.sectors[sc].surfaces[sf];
                        renderer.DrawLine(surf.vertices[ed].v, surf.vertices[surf.NextVX(ed)].v);
                    }
                    break;

                case MapMode.Vertex:
                    renderer.SetColor(GlobalVars.clMSel);
                    renderer.SetPointSize(4.0);
                    for (int s = 0; s < vxsel.Count; s++)
                    {
                        int sc, vx;
                        vxsel.GetSCVX(s, out sc, out vx);
                        renderer.DrawVertex(Level.sectors[sc].vertices[vx].v);
                    }
                    break;

                case MapMode.Thing:
                    renderer.SetColor(GlobalVars.clMSel);
                    for (int s = 0; s < thsel.Count; s++)
                        DrawThing(Level.things[thsel.GetTH(s)]);
                    break;

                case MapMode.Frame:
                    renderer.SetColor(GlobalVars.clMSel);
                    for (int s = 0; s < frsel.Count; s++)
                    {
                        int th, fr;
                        frsel.GetTHFR(s, out th, out fr);
                        if (fr == -1)
                            continue;

                        DrawFrame(Level.things[th].vals[fr]);
                    }
                    break;

                case MapMode.Light:
                    renderer.SetColor(GlobalVars.clMSel);
                    for (int s = 0; s < ltsel.Count; s++)
                        renderer.DrawVertex(Level.lights[ltsel.GetLT(s)].Pos);
                    break;
            }

            renderer.SetColor(GlobalVars.clMapSel);
            if (IsSelValid)
                switch (map_mode)
                {
                    case MapMode.Sector:
                        {
                            JKSector sec = Level.sectors[Cur_SC];

                            if (scsel.FindSC(Cur_SC) == -1)
                                renderer.SetColor(GlobalVars.clMapSelBack);
                            else
                                renderer.SetColor(GlobalVars.clSelMselBack);

                            renderer.SetCulling(Renderer.CullMode.Back);
                            renderer.DrawPolygons(sec.surfaces);

                            if (scsel.FindSC(Cur_SC) == -1)
                                renderer.SetColor(GlobalVars.clMapSel);
                            else
                                renderer.SetColor(GlobalVars.clSelMsel);

                            renderer.SetCulling(Renderer.CullMode.Front);
                            renderer.DrawPolygons(sec.surfaces);

                            if (sec.vertices.Count > 0)
                                renderer.DrawVertex(sec.vertices[0].v);
                        }
                        break;

                    case MapMode.Surface:
                        {
                            if (sfsel.FindSF(Cur_SC, Cur_SF) == -1)
                                renderer.SetColor(GlobalVars.clMapSelBack);
                            else
                                renderer.SetColor(GlobalVars.clSelMselBack);

                            renderer.SetCulling(Renderer.CullMode.Back);
                            renderer.DrawPolygon(Level.sectors[Cur_SC].surfaces[Cur_SF]);

                            if (sfsel.FindSF(Cur_SC, Cur_SF) == -1)
                                renderer.SetColor(GlobalVars.clMapSel);
                            else
                                renderer.SetColor(GlobalVars.clSelMsel);

                            renderer.SetCulling(Renderer.CullMode.Front);
                            renderer.DrawPolygon(Level.sectors[Cur_SC].surfaces[Cur_SF]);

                            if(true)
                            {
                                // draw surface normal
                                JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];
                                renderer.SetColor(Color4.green(0.5));
                                renderer.DrawLine(surf.Centroid, surf.Centroid + surf.normal * 0.15);

                                renderer.SetPointSize(4.0);
                                renderer.DrawVertex(surf.Centroid);
                            }
                        }
                        break;

                    case MapMode.Edge:
                        {
                            JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];

                            if (edsel.FindED(Cur_SC, Cur_SF, Cur_ED) == -1)
                                renderer.SetColor(GlobalVars.clMapSel);
                            else
                                renderer.SetColor(GlobalVars.clSelMsel);

                            Vec3 v1 = surf.vertices[Cur_ED].v;
                            Vec3 v2 = surf.vertices[surf.NextVX(Cur_ED)].v;
                            renderer.DrawLine(v1, v2);
                            renderer.SetPointSize(4.0);
                            renderer.DrawVertex(v1);
                            Vec3 d = v2 - v1;
                            v2 = v1 + d * 0.5;
                            Vec3 en = surf.normal ^ d;
                            v1 = v2 + en * 0.5;
                            renderer.DrawLine(v2, v1);
                        }
                        break;

                    case MapMode.Thing:
                        if (thsel.FindTH(Cur_TH) == -1)
                            renderer.SetColor(GlobalVars.clMapSel);
                        else
                            renderer.SetColor(GlobalVars.clSelMsel);

                        DrawSelThing(Level.things[Cur_TH]);
                        break;

                    case MapMode.Frame:
                        if (Cur_FR == -1)
                            renderer.DrawVertex(Level.things[Cur_TH].pos);
                        else
                        {
                            if (frsel.FindFR(Cur_TH, Cur_FR) == -1)
                                renderer.SetColor(GlobalVars.clMapSel);
                            else
                                renderer.SetColor(GlobalVars.clSelMsel);
                            DrawSelFrame(Cur_TH, Cur_FR);
                        }
                        break;

                    case MapMode.Vertex:
                        if (vxsel.FindVX(Cur_SC, Cur_VX) == -1)
                            renderer.SetColor(GlobalVars.clMapSel);
                        else
                            renderer.SetColor(GlobalVars.clSelMsel);

                        renderer.DrawVertex(Level.sectors[Cur_SC].vertices[Cur_VX].v);
                        break;

                    case MapMode.Light:
                        if (ltsel.FindLT(Cur_LT) == -1)
                            renderer.SetColor(GlobalVars.clMapSel);
                        else
                            renderer.SetColor(GlobalVars.clSelMsel);

                        if (Cur_LT < Level.lights.Count)
                        {
                            JedLight light = Level.lights[Cur_LT];
                            renderer.DrawVertex(light.Pos);
                            renderer.DrawCircle(light.Pos, light.range);
                        }
                        break;

                    case MapMode.Extra:
                        {
                            renderer.SetColor(GlobalVars.clMSel);

                            object obj = ExtraObjs[Cur_EX];
                            if (obj is Vertex)
                                renderer.DrawVertex((obj as Vertex).v);
                            else if (obj is ExtraLine)
                                renderer.DrawLine((obj as ExtraLine).v1, (obj as ExtraLine).v2);
                            else if (obj is Polygon)
                                renderer.DrawPolygon(obj as Polygon);
                        }
                        break;
                }
        }

        void DrawFrame(TemplateValue tv)
        {
            if (tv.adjType != AdjType.Frame)
                return;

            Vec3 fp;
            Orient fo;
            tv.GetFrame(out fp, out fo);
            renderer.DrawVertex(fp);
        }

        void DrawSelFrame(int th, int fr)
        {
            if (fr == -1)
                return;

            JKThing thing = Level.things[th];

            Vec3 fp;
            Orient fo;
            thing.vals[fr].GetFrame(out fp, out fo);

            renderer.DrawVertex(fp);

            renderer.SetCulling(Renderer.CullMode.None);
            DrawThingAt(thing, fp, fo);

            Vec3 d = new Vec3(0.0, 0.2, 0.0);
            Utils.RotateVector(ref d, fo.pitch, fo.yaw, fo.roll);
            renderer.DrawLine(fp, fp + d);
        }

        void DrawSelThing(JKThing thing)
        {
            renderer.SetCulling(Renderer.CullMode.None);
            DrawThing(thing);
            Vec3 d = new Vec3(0.0, 0.2, 0.0);
            Utils.RotateVector(ref d, thing.orient.pitch, thing.orient.yaw, thing.orient.roll);
            renderer.DrawLine(thing.pos, thing.pos + d);


#if false
            Vec3 spherePos = thing.pos;
            const double sphereRadius = 0.1;

            renderer.DrawSphere(spherePos, sphereRadius, Color4.yellow(0.5));


            int pushes = 0;
            const int maxPushes = 100;
            while(pushes < maxPushes)
            {

                //double shallowestPenetrationDist = double.MaxValue;
                //Vec3 shallowestOverlapPushVec = Vec3.Zero;
                List<Vec3> pushVecs = new List<Vec3>();

                foreach (JKSector sec in Level.sectors)
                {
                    foreach (JKSurface surf in sec.surfaces)
                    {
                        if (surf.adjoin != null && (surf.adjoinFlags & Adjoin.Flag.Passable) != 0)
                            continue;

                        //Plane3 plane = new Plane3(surf.normal, surf.vertices[0].v);
                        Poly3 poly = new Poly3();
                        foreach (Vertex v in surf.vertices)
                            poly.Add(v.v);

                        //Vec3 planePt = plane.NearestPointToPoint(spherePos);
                        //renderer.DrawVertex(planePt);

                        /*Tri3[] tris = poly.Triangulated;
                        foreach (Tri3 tri in tris)
                        {
                            renderer.DrawPolygon(tri.Polygon);

                            double hs = tri.Plane.Halfspace(spherePos);
                            if (hs <= 0.0)
                                hs = hs;
                        }*/


                        Vec3 closestPolyPt = poly.ClosestPointToPoint(spherePos);
                        //renderer.DrawVertex(closestPolyPt);

                        double sphereDistToPoly = (closestPolyPt - spherePos).Magnitude;
                        if (sphereDistToPoly > sphereRadius)
                            continue;

                        double neededDistToSep = sphereRadius - sphereDistToPoly + 0.01/*prevent actually resting on poly*/;

                        Vec3 awayVec = (spherePos - closestPolyPt).Unit * neededDistToSep;


                        /*if (neededDistToSep < shallowestPenetrationDist)
                        {
                            shallowestPenetrationDist = neededDistToSep;
                            shallowestOverlapPushVec = awayVec;
                        }*/
                        pushVecs.Add(awayVec);

                        //if(pushes == 1)
                        //    renderer.DrawSphere(spherePos + awayVec, sphereRadius, Color4.blue(0.3));

                        //if ((closestPolyPt - spherePos).Magnitude <= sphereRadius)
                        //    renderer.DrawLine(spherePos, closestPolyPt);
                    }
                }


                // no more overlap?
                //if (shallowestPenetrationDist == double.MaxValue)
                if(pushVecs.Count == 0)
                    break;


                Vec3 totalPush = Vec3.Zero;
                foreach (Vec3 pv in pushVecs)
                    totalPush += pv;


                // push by the least worst offender
                spherePos += totalPush / (double)pushVecs.Count;//shallowestOverlapPushVec;
                pushes++;

                //if(pushes == 1)
                //    renderer.DrawSphere(spherePos, sphereRadius, Color4.blue(0.3));
            }

            if (pushes >= maxPushes)
                renderer.DrawSphere(spherePos, sphereRadius, Color4.red(0.8));
            else
                renderer.DrawSphere(spherePos, sphereRadius, Color4.green(0.3));
#endif
        }

        private void NedMain_Shown(object sender, EventArgs e)
        {
            renderControl1.Init();

            //NewProject();

            // if paramstr(1)<>'' then OpenProject(Paramstr(1),op_open);

            //Toolbar.Show();
            GlobalVars.ItemEdit.Show();

            ResetEditor(true);
            Toolbar.SetDefaults();

            /*if (Program.LaunchSounds)
            {
                Audio.Channel c = Audio.PlaySound("i00ky22z.wav", Audio.DimensionMode._2D, false, true);
                if (c != null)
                {
                    c.Volume = 0.5;
                    c.Play();
                }
            }*/
        }

        private void New1_Click(object sender, EventArgs e)
        {
            if (!AskSave())
                return;

            ResettingEditor();

            PMsg.Text = string.Empty;
            Options.SetMOTS(false);
            NewProject();
            ResetEditor(true);
        }

        public bool AskSave()
        {
            if (!changed)
                return true;

            switch (MessageBox.Show("The project was changed. Save?", "NED", MessageBoxButtons.YesNoCancel))
            {
                case System.Windows.Forms.DialogResult.Yes:
                    return SaveProject();

                case System.Windows.Forms.DialogResult.No:
                    return true;

                default:
                    return false;
            }
        }

        public bool SaveProject()
        {
            return SysSaveProject(AskSaveAs);
        }

        public bool SysSaveProject(bool askSave)
        {
            string lastName = FileOperations.ExtractName(LevelFile);

            string fname;
            if (!askSave)
            {
                fname = FileOperations.ChangeFileExt(LevelFile, ".jed");
                FileOperations.BackUpFile(fname);
            }
            else
            {
                SaveFileDialog sfd = new SaveFileDialog();
                sfd.Filter = "JED Files (*.JED)|*.JED";
                sfd.Title = "Select a file to save to";
                sfd.FileName = FileOperations.ChangeFileExt(FileOperations.ExtractName(LevelFile), ".jed");
                if (sfd.ShowDialog(this) != System.Windows.Forms.DialogResult.OK)
                    return false;

                fname = sfd.FileName;
                string fpath = Path.GetDirectoryName(fname);

                /*string episodeFile = Path.Combine(fpath, "episode.jk");
                if (!File.Exists(episodeFile))
                {
                    string jklName = FileOperations.ChangeFileExt(FileOperations.ExtractName(fname), ".jkl");
                    using (StreamWriter sw = File.CreateText(episodeFile))
                    {
                        sw.WriteLine("\"New Level\"");
                        sw.WriteLine("# Title must be first line.  A translated version is looked up in jkStrings.uni");
                        sw.WriteLine("# Please use TABs for alignment!");
                        sw.WriteLine();
                        sw.WriteLine("TYPE  1  # single-player");
                        sw.WriteLine();
                        sw.WriteLine("SEQ   1");
                        sw.WriteLine();
                        sw.WriteLine("# <lightpow> and <darkpow> are the inventory bin numbers for that power.");
                        sw.WriteLine();
                        sw.WriteLine("# <line> <cd>  <level>  <type>   <file>         <lightpow>  <darkpow>   <gotoA>  <gotoB>");
                        sw.WriteLine();
                        sw.WriteLine("10:      1     1        LEVEL    " + jklName + "   0           0           -1    -1    # Level 1: NewLevel");
                        sw.WriteLine();
                        sw.WriteLine("end");
                    }
                }*/
            }

            // if saving .jed from an embedded JKL, maybe we should extract the whole thing?
            bool forceSaveTemplates = false;
            if (!string.IsNullOrEmpty(GlobalVars.ProjectGob))
            {
                forceSaveTemplates = true;  // definitely generate templates file..   this might be a little screwy since different JKLs can have different templates.. should probably be handled gracefully.. at some point.. by someone..

                string projectdir = Path.GetDirectoryName(fname);

                DialogResult dr = MessageBox.Show("It looks like you're saving a .jed project from a JKL embedded in a GOB. Would you like to extract " + Path.GetFileName(GlobalVars.ProjectGob) + " to the the output directory?\n" + projectdir, "Extract GOB", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
                if (dr == System.Windows.Forms.DialogResult.Cancel)
                    return false;

                if (dr == System.Windows.Forms.DialogResult.Yes)
                {
                    GOB gob = new GOB(false, GlobalVars.ProjectGob);

                    Dictionary<string, string> fileconflicts = new Dictionary<string, string>();

                    foreach (string gobfile in gob)
                    {
                        string outfile = Path.Combine(projectdir, gobfile);
                        string outpath = Path.GetDirectoryName(outfile);// gobfile could have subfolder so ensure to figure out the final directory name

                        if (!Directory.Exists(outpath))
                            Directory.CreateDirectory(outpath);

                        // if duplicate, keep track for us to let user pick what to do later
                        if (File.Exists(outfile))
                        {
                            fileconflicts.Add(gobfile, outfile);
                            continue;
                        }

                        using(MemoryStream mem = gob[gobfile])
                        {
                            if (mem.Length > 0)
                            {
                                byte[] buf = new byte[mem.Length];
                                mem.Read(buf, 0, buf.Length);
                                File.WriteAllBytes(outfile, buf);
                            }
                        }
                    }

                    if (fileconflicts.Count > 0)
                    {
                        DialogResult dr2 = MessageBox.Show("Found " + fileconflicts.Count.ToString() + " conflicts where file(s) already exist in project directory.\nDo you want to overwrite them with the GOB's data?", "Overwrite Project Files", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Warning);
                        if (dr2 == System.Windows.Forms.DialogResult.Cancel)
                        {
                            // a little late but whatever. i guess leave files there, who cares
                            return false;
                        }

                        if (dr2 == System.Windows.Forms.DialogResult.Yes)
                        {
                            foreach (string gobfile in fileconflicts.Keys)
                            {
                                // folders should already exist so we're good
                                using (MemoryStream ms = gob[gobfile])
                                {
                                    byte[] buf = new byte[ms.Length];
                                    ms.Read(buf, 0, buf.Length);
                                    File.WriteAllBytes(fileconflicts[gobfile], buf);
                                }
                            }
                        }
                    }
                }
            }

            SaveJEDTo(fname);

            GlobalVars.EpisodeEdit.UpdateLevelFile(lastName, fname);

            // update/create episode file
            GlobalVars.EpisodeEdit.Save();

            if(forceSaveTemplates)
                Templates.SaveToFile(System.IO.Path.Combine(GlobalVars.ProjectDir, Level.mots ? "mots.tpl" : "master.tpl"));

            return true;
        }

        public void ExtractProject(string gobpath, string dstpath, bool convertJKLToJED)
        {
            bool mots = gobpath.EndsWith(".goo", StringComparison.InvariantCultureIgnoreCase);

            GOB gob = new GOB(false, gobpath);

            Dictionary<string, string> outfiles = new Dictionary<string, string>();
            Dictionary<string, string> conflicts = new Dictionary<string, string>();

            foreach (string gobfile in gob)
            {
                string outfile = Path.Combine(dstpath, gobfile);

                // if duplicate, keep track for us to let user pick what to do later
                if (File.Exists(outfile))
                    conflicts.Add(gobfile, outfile);
                else
                    outfiles.Add(gobfile, outfile);
            }

            if (conflicts.Count > 0)
            {
                DialogResult dr = MessageBox.Show("Found " + conflicts.Count.ToString() + " conflicts where file(s) already exist in project directory.\nDo you want to overwrite them with the GOB's data?", "Overwrite Project Files", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Warning);
                if (dr == System.Windows.Forms.DialogResult.Cancel)
                    return;

                // if we want to override, move entries to nonconflicts
                if (dr == System.Windows.Forms.DialogResult.Yes)
                {
                    foreach (string gobfile in conflicts.Keys)
                        outfiles.Add(gobfile, conflicts[gobfile]);
                }
            }


            ProgressDialog.Reset(outfiles.Keys.Count, 0);
            ProgressDialog.Message("Extracting GOB");

            foreach (string gobfile in outfiles.Keys)
            {
                string outfile = outfiles[gobfile];
                string outpath = Path.GetDirectoryName(outfile);// gobfile could have subfolder so ensure to figure out the final directory name

                if (!Directory.Exists(outpath))
                    Directory.CreateDirectory(outpath);

                // folders should already exist so we're good
                using (MemoryStream ms = gob[gobfile])
                {
                    if (ms.Length > 0)
                    {
                        byte[] buf = new byte[ms.Length];
                        ms.Read(buf, 0, buf.Length);
                        File.WriteAllBytes(outfile, buf);
                    }
                }

                ProgressDialog.Step();
            }

            ProgressDialog.Cleanup();

            Templates.Clear();
            Templates.LoadFromFile(mots ? "mots.tpl" : "master.tpl");

            string oldprojectdir = GlobalVars.ProjectDir;
            GlobalVars.ProjectDir = dstpath;

            // load all JKLs
            string jklpath = Path.Combine(dstpath, "jkl");
            if (Directory.Exists(jklpath))
            {
                string[] jkls = Directory.GetFiles(jklpath, "*.jkl");
                if (jkls.Length > 0)
                {
                    // cant use progress dialog because JKLevel internally is

                    //ProgressDialog.Reset(jkls.Length, 0);
                    //ProgressDialog.Message("Converting JKL -> JED");

                    foreach (string jklfile in jkls)
                    {
                        JKLevel level = new JKLevel();
                        GlobalVars.LevelOverride = level;

                        Templates.AddFromJKL(jklfile);

                        level.LoadFromJKL(jklfile);

                        if(convertJKLToJED)
                            level.SaveToJED(Path.Combine(dstpath, Path.GetFileNameWithoutExtension(jklfile) + ".jed"));

                        GlobalVars.LevelOverride = null;

                        //ProgressDialog.Step();
                    }

                    //ProgressDialog.Cleanup();
                }
            }

            //SaveJEDTo(fname);

            // update/create episode file
            //GlobalVars.EpisodeEdit.Save();

            Templates.SaveToFile(System.IO.Path.Combine(dstpath, mots ? "mots.tpl" : "master.tpl"));


            GlobalVars.ProjectDir = oldprojectdir;



        }

        public void SaveJEDTo(string name)
        {
            Level.layerVisString = Toolbar.GetLVISString();
            Level.SaveToJED(name);
            AskSaveAs = false;
            NewLevel = false;
            changed = false;
            LevelFile = name;
            AddRecent(name);
            SetLevelName();
        }

        private void NewMOTSProject1_Click(object sender, EventArgs e)
        {
            if (!AskSave())
                return;

            ResettingEditor();

            PMsg.Text = string.Empty;
            Options.SetMOTS(true);
            NewProject();
            ResetEditor(true);
        }

        public Renderer renderer
        {
            get
            {
                return renderControl1.renderer;
            }
        }

        public int Cur_SC = 0;
        public int Cur_SF = 0;
        public int Cur_VX = 0;
        public int Cur_ED = 0;
        public int Cur_LT = 0;
        public int Cur_TH = 0;
        public int Cur_FR = 0;
        public int Cur_EX = 0;

        public SCMultiSel scsel = new SCMultiSel();
        public SFMultiSel sfsel = new SFMultiSel();
        public VXMultiSel vxsel = new VXMultiSel();
        public THMultiSel thsel = new THMultiSel();
        public LTMultiSel ltsel = new LTMultiSel();
        public EDMultiSel edsel = new EDMultiSel();
        public FRMultiSel frsel = new FRMultiSel();
        public THMultiSel exsel = new THMultiSel();

        public List<object> ExtraObjs = new List<object>();
        public string ExtraObjsName = string.Empty;

        public int AddExtraVertex(Vec3 a, string name)
        {
            ExtraVertex v = new ExtraVertex();
            v.v = a;
            v.Name = name;
            int i = ExtraObjs.Count;
            ExtraObjs.Add(v);
            return i;
        }

        public void ClearExtraObjs()
        {
            ExtraObjs.Clear();
            Cur_EX = -1;
            ExtraObjsName = string.Empty;
        }

        //bool wasActive = false;

#if true
        // havin a hell of a time getting C# with "child" windows to behave like they did in JED.
        // if we make sub-dialogs like item editor as child windows, then they can never go behind the main window.
        // if we make them separate forms, then they lose Z order when the application is restored.
        // this wndproc code below at least gets all sub-dialogs to come back to visibility, but the Z order is not preserved.
        // even more annoyingly, the ActiveForm regardless on activated/deactivated, regardless if we call base.WndProc first or after,
        // ActiveForm always seems to be null.  perhaps we need to take a PreMsgFilter approach...to see if ActiveForm is valid before windows does too much.

        protected override void WndProc(ref Message m)
        {
            if (m.Msg == 0x06/*activate*/)
            {
                if (((uint)m.WParam & 0xFFFF) == 0x02/*clickactive*/)
                {
                    handleFirstClickOnActivated = true;
                }
            }

            base.WndProc(ref m);

            if (m.Msg == 0x1C/*activateapp*/)
            {
                if (m.WParam != IntPtr.Zero)
                {
                    //activated
                    foreach (Form f in Application.OpenForms)
                        if (f != this)
                            f.BringToFront();

                    BringToFront();
                }
                else
                {
                    //deactivated

                }
            }
        }
#else
        [DllImport("user32.dll")]
        static extern IntPtr GetForegroundWindow();

        protected override void OnActivated(EventArgs e)
        {
            base.OnActivated(e);

            IntPtr oldForegroundWindow = GetForegroundWindow();

            // if we dont own the previous window then we'll want to bring all of our stuff to the top
            foreach(Form f in Application.OpenForms)
                if (f.Handle == oldForegroundWindow)
                {
                    // we own this window.. do nothing
                    return;
                }

            // bring all windows to the forefront
            foreach (Form f in Application.OpenForms)
                f.BringToFront();
        }

        protected override void OnDeactivate(EventArgs e)
        {
            base.OnDeactivate(e);

            
        }
#endif

        public void CalcAngle(ref double vangle)
        {
            if (map_mode != MapMode.Surface)
            {
                MessageBox.Show("You must be in surface mode");
                return;
            }

            int n = sfsel.AddSF(Cur_SC, Cur_SF);
            if (sfsel.Count != 2)
            {
                sfsel.DeleteN(n);
                MessageBox.Show("You must have 2 surfaces selected");
                return;
            }

            JKSurface surf1, surf2;
            if (sfsel.FindSF(Cur_SC, Cur_SF) == 0)
            {
                int sc, sf;
                sfsel.GetSCSF(0, out sc, out sf);
                surf1 = Level.sectors[sc].surfaces[sf];
                sfsel.GetSCSF(1, out sc, out sf);
                surf2 = Level.sectors[sc].surfaces[sf];
            }
            else
            {
                int sc, sf;
                sfsel.GetSCSF(0, out sc, out sf);
                surf2 = Level.sectors[sc].surfaces[sf];
                sfsel.GetSCSF(1, out sc, out sf);
                surf1 = Level.sectors[sc].surfaces[sf];
            }

            sfsel.Clear();

            double d = surf1.normal * surf2.normal;
            vangle = 180.0 - MathLib.ToDeg(Math.Acos(d));

            Vec3 vec = (surf1.normal ^ surf2.normal).Unit;
            if (vec.Magnitude == 0.0)
            {
                MessageBox.Show("The surfaces are already coplanar");
                return;
            }

            Vec3 lvec = surf2.normal ^ vec;// shouldn't we be normalizing here?

            Vec3 pA = surf1.vertices[0].v;
            Vec3 pB = surf2.vertices[0].v;
            Vec3 p;
            if (!Utils.PlaneLineXnNew(surf1.normal, pA, pB, pB + lvec, out p))
                Utils.PanMessage(Utils.MessageType.Warning, "Something something dark side");

            renderer.GridPos = p;
            renderer.SetGridNormal(vec);
            renderer.SetGridXNormal(lvec);
        }

        public void FlipObject(ToolAxis how)
        {
            switch (map_mode)
            {
                case MapMode.Sector:
                    {
                        Undo.StartUndoRec("Flip sector(s)");
                        int i = scsel.AddSC(Cur_SC);
                        if (how == ToolAxis.Grid)
                            Utils.FlipSectorsOverPlane(Level, scsel, renderer.GridNormal, renderer.GridPos);
                        else
                        {
                            Vec3 c = Utils.CalcSecCenter(Level.sectors[Cur_SC]);
                            Undo.StartUndoRec("Flip Sectors");
                            Utils.FlipSectors(Level, scsel, c, how);
                        }
                        scsel.DeleteN(i);
                    }
                    break;

                case MapMode.Thing:
                    {
                        Undo.StartUndoRec("Flip thing(s)");

                        int i = thsel.AddTH(Cur_TH);
                        if (how == ToolAxis.Grid)
                            Utils.FlipThingsOverPlane(Level, thsel, renderer.GridNormal, renderer.GridPos);
                        else
                            Utils.FlipThings(Level, thsel, Level.things[Cur_TH].pos, how);
                        thsel.DeleteN(i);
                    }
                    break;

                case MapMode.Light:
                    {
                        Undo.StartUndoRec("Flip light(s)");

                        int i = ltsel.AddLT(Cur_LT);
                        Vec3 c = Level.lights[Cur_LT].Pos;
                        Vec3 vec;
                        switch (how)
                        {
                            case ToolAxis.X: vec = Vec3.PosX; break;
                            case ToolAxis.Y: vec = Vec3.PosY; break;
                            case ToolAxis.Z: vec = Vec3.PosZ; break;
                            case ToolAxis.Grid: vec = renderer.GridNormal; c = renderer.GridPos; break;
                            default:
                                return;
                        }
                        Utils.FlipLightsOverPlane(Level, ltsel, vec, c);
                        ltsel.DeleteN(i);
                    }
                    break;

                case MapMode.Frame:
                    {
                        Undo.StartUndoRec("Flip frame(s)");

                        int i = frsel.AddFR(Cur_TH, Cur_FR);
                        Vec3 c;
                        if (Cur_FR < 0)
                            c = Level.things[Cur_TH].pos;
                        else
                        {
                            Orient dummy;
                            Level.things[Cur_TH].vals[Cur_FR].GetFrame(out c, out dummy);
                        }

                        Vec3 vec;
                        switch (how)
                        {
                            case ToolAxis.X: vec = Vec3.PosX; break;
                            case ToolAxis.Y: vec = Vec3.PosY; break;
                            case ToolAxis.Z: vec = Vec3.PosZ; break;
                            case ToolAxis.Grid: vec = renderer.GridNormal; c = renderer.GridPos; break;
                            default:
                                return;
                        }
                        Utils.FlipFramesOverPlane(Level, frsel, vec, c);
                        frsel.DeleteN(i);
                    }
                    break;
            }
        }

        public void ScaleObject(double sfactor, ToolAxis how)
        {
            Utils.ScaleData sd = new Utils.ScaleData();
            sd.sfactor = sfactor;
            if ((how & ToolAxis.X) != 0)
                sd.how = Utils.ScaleDataMode.X;
            else if ((how & ToolAxis.Y) != 0)
                sd.how = Utils.ScaleDataMode.Y;
            else if ((how & ToolAxis.Z) != 0)
                sd.how = Utils.ScaleDataMode.Z;
            else if ((how & ToolAxis.Grid) != 0)
            {
                sd.how = Utils.ScaleDataMode.Vector;
                sd.vec = renderer.GridNormal;
            }
            else
                sd.how = Utils.ScaleDataMode.XYZ;

            switch (map_mode)
            {
                case MapMode.Sector:
                    {
                        Undo.StartUndoRec("Scale sector(s)");

                        sd.c = Utils.CalcSecCenter(Level.sectors[Cur_SC]);

                        int i = scsel.AddSC(Cur_SC);
                        Utils.ScaleSectors(Level, scsel, sd, (how & ToolAxis.TX) != 0);
                        scsel.DeleteN(i);
                    }
                    break;

                case MapMode.Thing:
                    {
                        Undo.StartUndoRec("Scale thing(s)");

                        sd.c = Level.things[Cur_TH].pos;

                        int i = thsel.AddTH(Cur_TH);
                        Utils.ScaleThings(Level, thsel, sd);
                        thsel.DeleteN(i);

                    }
                    break;

                case MapMode.Light:
                    {
                        Undo.StartUndoRec("Scale light(s)");

                        sd.c = Level.lights[Cur_LT].Pos;

                        int i = ltsel.AddLT(Cur_LT);
                        Utils.ScaleLights(Level, ltsel, sd);
                        ltsel.DeleteN(i);
                    }
                    break;

                case MapMode.Frame:
                    {
                        Undo.StartUndoRec("Scale light(s)");

                        int i = frsel.AddFR(Cur_TH, Cur_FR);
                        if (Cur_FR == -1)
                            sd.c = Level.things[Cur_TH].pos;
                        else
                        {
                            Orient dummy;
                            Level.things[Cur_TH].vals[Cur_FR].GetFrame(out sd.c, out dummy);
                        }
                        Utils.ScaleFrames(Level, frsel, sd);
                        frsel.DeleteN(i);
                    }
                    break;
            }
        }

        public void TranslateObject(Vec3 d)
        {
            switch (map_mode)
            {
                case MapMode.Sector:
                    {
                        Undo.StartUndoRec("Translate sectors");

                        JKSector sec = Level.sectors[Cur_SC];
                        Utils.TranslateSectors(Level, scsel, Cur_SC, d);
                    }
                    break;

                case MapMode.Surface:
                    {
                        Undo.StartUndoRec("Translate surfaces");

                        JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];
                        Utils.TranslateSurfaces(Level, sfsel, Cur_SC, Cur_SF, d);
                    }
                    break;
            }
        }

        private void AddTHValue(JKThing th, string name, string vs)
        {
            TemplateValue tv = new TemplateValue();
            th.vals.Add(tv);
            tv.name = name;
            tv.vType = Values.GetTPLVType(tv.name);
            tv.adjType = Values.GetTPLType(tv.name);
            tv.Val(vs);
        }

        public void MakeDoor()
        {
            if (map_mode != MapMode.Thing)
            {
                MessageBox.Show("You must be in Thing mode");
                return;
            }

            JKThing th = Level.things[Cur_TH];

            int i = th.vals.IndexOfName("numframes");
            if (i != -1)
            {
                MessageBox.Show("The thing appears to be a door already");
                return;
            }

            double d = th.box.Top - th.box.Bottom;
            if (MathLib.CompareScalar(d, 0.0) == 0)
            {
                MessageBox.Show("The thing doesn't have the correct bounding box");
                return;
            }

            Undo.StartUndoRec("Make door");
            Undo.SaveThingUndo(th, Undo.Change.Changed);

            th.vals.Clear();

            AddTHValue(th, "thingflags", "0x408");
            AddTHValue(th, "numframes", "2");
            Vec3 v = th.pos;
            AddTHValue(th, "frame", "(" + Utils.dstr(v.x, 6) + "/" + Utils.dstr(v.y, 6) + "/" + Utils.dstr(v.z, 6) + ":" +
                                          Utils.dstr(th.orient.pitch, 6) + "/" + Utils.dstr(th.orient.yaw, 6) + "/" + Utils.dstr(th.orient.roll, 6) + ")");
            v = new Vec3(0.0, 0.0, d/*magically delicious*/);
            Utils.RotateVector(ref v, th.orient.pitch, th.orient.yaw, th.orient.roll);
            AddTHValue(th, "frame", "(" + Utils.dstr(v.x, 6) + "/" + Utils.dstr(v.y, 6) + "/" + Utils.dstr(v.z, 6) + ":" +
                                          Utils.dstr(th.orient.pitch, 6) + "/" + Utils.dstr(th.orient.yaw, 6) + "/" + Utils.dstr(th.orient.roll, 6) + ")");

            CogFile cf = new CogFile();
            JKCOG cg = new JKCOG();
            cg.name = "00_door.cog";
            Level.cogs.Add(cg);
            cf.LoadNoLocals(cg.name);
            for (int x = 0; x < cf.Count; x++)
            {
                COGValue cv = new COGValue();
                cv.Assign(cf[x]);
                cg.vals.Add(cv);
                if (x == 0)
                    cv.Val(th.num.ToString());
                else
                    cv.Val(cf[i].AsString);
                cv.Resolve();
            }

            GlobalVars.CogForm.RefreshList();
        }

        public void RotateObject(double angle, ToolAxis axis)
        {
            
            Vec3 vec;
            Vec3? c = null;
            switch (axis)
            {
                case ToolAxis.X: vec = Vec3.PosX; break;
                case ToolAxis.Y: vec = Vec3.PosY; break;
                case ToolAxis.Z: vec = Vec3.PosZ; break;
                case ToolAxis.Grid: vec = renderer.GridNormal; c = renderer.GridPos; break;
                default: 
                    return;
            }

            switch (map_mode)
            {
                case MapMode.Sector:
                    {
                        JKSector sec = Level.sectors[Cur_SC];

                        Undo.StartUndoRec("Rotate sector(s)");
                        int i = scsel.AddSC(Cur_SC);
                        Utils.RotateSectors(Level, scsel, vec, c ?? Utils.CalcSecCenter(sec), angle);
                        scsel.DeleteN(i);
                    }
                    break;

                case MapMode.Thing:
                    {
                        JKThing th = Level.things[Cur_TH];

                        Undo.StartUndoRec("Rotate thing(s)");
                        int i = thsel.AddTH(Cur_TH);
                        Utils.RotateThings(Level, thsel, vec, c ?? th.pos, angle);
                        thsel.DeleteN(i);
                    }
                    break;

                case MapMode.Light:
                    {
                        JedLight lt = Level.lights[Cur_LT];

                        Undo.StartUndoRec("Rotate light(s)");
                        int i = ltsel.AddLT(Cur_LT);
                        Utils.RotateLights(Level, ltsel, vec, c ?? lt.Pos, angle);
                        ltsel.DeleteN(i);
                    }
                    break;

                case MapMode.Frame:
                    {
                        Vec3 altC;
                        if (Cur_FR < 0)
                            altC = Level.things[Cur_TH].pos;
                        else
                        {
                            Orient dummy;
                            Level.things[Cur_TH].vals[Cur_FR].GetFrame(out altC, out dummy);
                        }

                        Undo.StartUndoRec("Rotate frame(s)");
                        int i = frsel.AddFR(Cur_TH, Cur_FR);
                        Utils.RotateFrames(Level, frsel, vec, c ?? altC, angle);
                        frsel.DeleteN(i);
                    }
                    break;
            }
        }

        public void ResetEditor(bool DefaultParams)
        {
            Undo.ClearUndoBuffer();
            ClearExtraObjs();
            GlobalVars.UrqForm.Reload();

            if (DefaultParams)
            {
                renderer.CamPos = Vec3.Zero;
                renderer.SetZ(Vec3.PosZ);
                renderer.SetX(Vec3.PosX);
                camOrient = Orient.Zero;
                renderer.GridPos = Vec3.Zero;
                renderer.SetGridNormal(Vec3.PosZ);
                renderer.SetGridXNormal(Vec3.PosX);
                renderer.scale = 1.0;
                
                Cur_SC = 0;
                Cur_SF = 0;
                Cur_TH = 0;
                Cur_VX = 0;
                Cur_LT = 0;
                Toolbar.CBGridStep.Text = GlobalVars.DefGridStep.ToString();
                Toolbar.CBGridLine.Text = GlobalVars.DefGridLine.ToString();
                Toolbar.CBGridDot.Text = GlobalVars.DefGridDot.ToString();
                Toolbar.CBGridSize.Text = GlobalVars.DefGridSize.ToString();
                Toolbar.CBGrid.Checked = true;
            }

            if (DefaultParams)
            {
                SetMapMode(MapMode.Sector);
                Toolbar.SetLVis(Level.layerVisString);
                SetThingView(GlobalVars.ThingView);
                SetMSelMode(GlobalVars.DefMSelMode);
                scsel.Clear();
                sfsel.Clear();
                edsel.Clear();
                vxsel.Clear();
                thsel.Clear();
                ltsel.Clear();
            }
            else
            {
                Toolbar.RefreshLayers(false);
                SetMapMode(map_mode);
                VerifyMultiSelection();
            }

            SetLevelName();

            GlobalVars.CogForm.RefreshList();

            LoadTemplates();

            LoadPlugins();

            changed = false;
            GlobalVars.Preview3D.ShowPreview();//ReloadLevel();
            SyncRecents();

            SetMotsIndicator();

            // ensure "All" is selected
            cogForm1.SelectGroupAll();

            GlobalVars.EpisodeEdit.RefreshEpisode();

            foreach (JKThing thing in Level.things)
                UpdateThingData(thing);

            SaveTimer.Enabled = (GlobalVars.AutoSave && GlobalVars.SaveInterval > 0);
            if(GlobalVars.SaveInterval > 0)
                SaveTimer.Interval = GlobalVars.SaveInterval * 60 * 1000;
            SetHideLights(false);
            SetHideThings(false);
        }


        void ToSCMsel(MapMode mode)
        {
            switch (mode)
            {
                case MapMode.Surface:
                    scsel.Clear();
                    for (int i = 0; i < sfsel.Count; i++)
                    {
                        int sc, a;
                        sfsel.GetSCSF(i, out sc, out a);
                        scsel.AddSC(sc);
                    }
                    break;

                case MapMode.Edge:
                    scsel.Clear();
                    for (int i = 0; i < edsel.Count; i++)
                    {
                        int sc, a, b;
                        edsel.GetSCSFED(i, out sc, out a, out b);
                        scsel.AddSC(sc);
                    }
                    break;

                case MapMode.Vertex:
                    scsel.Clear();
                    for (int i = 0; i < vxsel.Count; i++)
                    {
                        int sc, a;
                        vxsel.GetSCVX(i, out sc, out a);
                        scsel.AddSC(sc);
                    }
                    break;
            }
        }

        void ToSFMsel(MapMode mode)
        {
            switch (mode)
            {
                case MapMode.Sector:
                    sfsel.Clear();
                    for (int i = 0; i < scsel.Count; i++)
                    {
                        int sc = scsel.GetSC(i);
                        JKSector sec = Level.sectors[sc];
                        for (int sf = 0; sf < sec.surfaces.Count; sf++)
                            sfsel.AddSF(sc, sf);
                    }
                    break;

                case MapMode.Edge:
                    sfsel.Clear();
                    for (int i = 0; i < edsel.Count; i++)
                    {
                        int sc, sf, b;
                        edsel.GetSCSFED(i, out sc, out sf, out b);
                        sfsel.AddSF(sc, sf);
                    }
                    break;

                case MapMode.Vertex:
                    sfsel.Clear();
                    break;
            }
        }

        void ToEDMsel(MapMode mode)
        {
            switch (mode)
            {
                case MapMode.Sector:
                    edsel.Clear();
                    for (int i = 0; i < scsel.Count; i++)
                    {
                        int sc = scsel.GetSC(i);
                        JKSector sec = Level.sectors[sc];
                        for (int sf = 0; sf < sec.surfaces.Count; sf++)
                        {
                            for (int ed = 0; ed < sec.surfaces[sf].vertices.Count; ed++)
                                edsel.AddED(sc, sf, ed);
                        }
                    }
                    break;

                case MapMode.Surface:
                    edsel.Clear();
                    for (int i = 0; i < sfsel.Count; i++)
                    {
                        int sc, sf;
                        sfsel.GetSCSF(i, out sc, out sf);
                        JKSurface surf = Level.sectors[sc].surfaces[sf];
                        for (int ed = 0; ed < surf.vertices.Count; ed++)
                            edsel.AddED(sc, sf, ed);
                    }
                    break;

                case MapMode.Vertex:
                    edsel.Clear();
                    break;
            }
        }

        void ToVXMsel(MapMode mode)
        {
            switch (mode)
            {
                case MapMode.Sector:
                    vxsel.Clear();
                    for (int i = 0; i < scsel.Count; i++)
                    {
                        int sc = scsel.GetSC(i);
                        JKSector sec = Level.sectors[sc];
                        for (int vx = 0; vx < sec.vertices.Count; vx++)
                            vxsel.AddVX(sc, vx);
                    }
                    break;

                case MapMode.Surface:
                    vxsel.Clear();
                    for (int i = 0; i < sfsel.Count; i++)
                    {
                        int sc, sf;
                        sfsel.GetSCSF(i, out sc, out sf);
                        JKSurface surf = Level.sectors[sc].surfaces[sf];
                        for (int vx = 0; vx < surf.vertices.Count; vx++)
                            vxsel.AddVX(sc, surf.vertices[vx].num);
                    }
                    break;

                case MapMode.Edge:
                    vxsel.Clear();
                    for (int i = 0; i < edsel.Count; i++)
                    {
                        int sc, sf, ed;
                        edsel.GetSCSFED(i, out sc, out sf, out ed);
                        JKSurface surf = Level.sectors[sc].surfaces[sf];
                        vxsel.AddVX(sc, surf.vertices[ed].num);
                        vxsel.AddVX(sc, surf.vertices[surf.NextVX(ed)].num);
                    }
                    break;
            }
        }

        void ToFRMsel(MapMode mode)
        {
            switch (mode)
            {
                case MapMode.Thing:
                    frsel.Clear();
                    for (int i = 0; i < thsel.Count; i++)
                    {
                        int th = thsel.GetTH(i);
                        JKThing thing = Level.things[th];
                        for (int fr = 0; fr < thing.vals.Count; fr++)
                            if (thing.vals[fr].adjType == AdjType.Frame)
                                frsel.AddFR(th, fr);
                    }
                    break;
            }
        }

        void ToTHMsel(MapMode mode)
        {
            switch (mode)
            {
                case MapMode.Frame:
                    thsel.Clear();
                    for (int i = 0; i < frsel.Count; i++)
                    {
                        int th, a;
                        frsel.GetTHFR(i, out th, out a);
                        thsel.AddTH(th);
                    }
                    break;
            }
        }

        public void SetMapMode(MapMode mm)
        {
            MapMode oldmode = map_mode;
            map_mode = mm;
            switch (mm)
            {
                case MapMode.Sector:
                    ToSCMsel(oldmode);
                    SetCurSC(Cur_SC);
                    break;

                case MapMode.Surface:
                    ToSFMsel(oldmode);
                    SetCurSF(Cur_SC, Cur_SF);
                    break;

                case MapMode.Vertex:
                    ToVXMsel(oldmode);
                    SetCurVX(Cur_SC, Cur_VX);
                    break;

                case MapMode.Thing:
                    ToTHMsel(oldmode);
                    SetCurTH(Cur_TH);
                    break;

                case MapMode.Frame:
                    ToFRMsel(oldmode);
                    SetCurFR(Cur_TH, Cur_FR);
                    break;

                case MapMode.Edge:
                    ToEDMsel(oldmode);
                    SetCurED(Cur_SC, Cur_SF, Cur_ED);
                    break;

                case MapMode.Light:
                    SetCurLT(Cur_LT);
                    break;

                case MapMode.Extra:
                    SetCurEX(Cur_EX);
                    break;

                default:
                    GlobalVars.ItemEdit.Hide();
                    return;
            }

            BNSC.Checked = mm == MapMode.Sector;
            BNSF.Checked = mm == MapMode.Surface;
            BNVX.Checked = mm == MapMode.Vertex;
            BNED.Checked = mm == MapMode.Edge;
            BNTH.Checked = mm == MapMode.Thing;
            BNFR.Checked = mm == MapMode.Frame;
            BNLT.Checked = mm == MapMode.Light;
            // extra?
            Invalidate();
        }

        public void SetCurSC(int sc)
        {
            Cur_SC = sc;
            if (Cur_SC >= Level.sectors.Count)
                Cur_SC = 0;
            if (Cur_SC < 0)
                Cur_SC = Level.sectors.Count - 1;
            EditObject();
        }

        public void SetCurSF(int sc, int sf)
        {
            Cur_SC = sc;
            Cur_SF = sf;
            if (sc >= Level.sectors.Count)
                Cur_SC = 0;
            if (Cur_SC < 0)
                Cur_SC = Level.sectors.Count - 1;

            JKSector sec = Level.sectors[Cur_SC];
            if (Cur_SF < 0)
                Cur_SF = sec.surfaces.Count - 1;
            if (Cur_SF >= sec.surfaces.Count)
                Cur_SF = 0;

            EditObject();
        }

        public void SetCurVX(int sc, int vx)
        {
            Cur_SC = sc;
            Cur_VX = vx;
            if (sc >= Level.sectors.Count)
                Cur_SC = 0;
            if (sc < 0)
                Cur_SC = Level.sectors.Count - 1;

            JKSector sec = Level.sectors[Cur_SC];
            if (vx < 0)
                Cur_VX = sec.vertices.Count - 1;
            if (vx >= sec.vertices.Count)
                Cur_VX = 0;

            EditObject();
        }

        public void SetCurED(int sc, int sf, int ed)
        {
            Cur_SC = sc;
            Cur_SF = sf;
            Cur_ED = ed;
            if (sc >= Level.sectors.Count)
                Cur_SC = 0;
            if (sc < 0)
                Cur_SC = Level.sectors.Count - 1;

            JKSector sec = Level.sectors[Cur_SC];
            if (sf < 0)
                Cur_SF = sec.surfaces.Count - 1;
            if (sf >= sec.surfaces.Count)
                Cur_SF = 0;

            JKSurface surf = sec.surfaces[Cur_SF];
            if (surf.vertices.Count == 0)
                return;

            if (ed == -2)
                Cur_ED = surf.vertices.Count - 1;
            else if (ed < 0)
                SetCurED(Cur_SC, Cur_SF - 1, -2);
            if (ed >= surf.vertices.Count)
                SetCurED(Cur_SC, Cur_SF + 1, 0);

            EditObject();
        }

        public void SetCurTH(int th)
        {
            Cur_TH = th;
            if (th >= Level.things.Count)
                Cur_TH = 0;
            if (th < 0)
                Cur_TH = Level.things.Count - 1;

            EditObject();
        }

        public void SetCurFR(int th, int fr)
        {
            Cur_TH = th;
            if (th >= Level.things.Count)
                Cur_TH = 0;
            if (th < 0)
                Cur_TH = Level.things.Count - 1;

            JKThing thing = Level.things[Cur_TH];

            if (fr < -1)
                Cur_FR = thing.vals.Count - 1;
            if (fr >= thing.vals.Count)
                Cur_FR = -1;

            EditObject();
        }

        public void SetCurLT(int lt)
        {
            Cur_LT = lt;
            if (lt >= Level.lights.Count)
                Cur_LT = 0;
            if (lt < 0)
                Cur_LT = Level.lights.Count - 1;

            EditObject();
        }

        public void SetCurEX(int ex)
        {
            Cur_EX = ex;
            if (ex >= ExtraObjs.Count)
                Cur_EX = 0;
            if (ex < 0)
                Cur_EX = ExtraObjs.Count - 1;

            if (OnExtraSelect != null)
                OnExtraSelect(ExtraObjs[Cur_EX]);

            EditObject();
        }

        public void EditObject()
        {
            if (!IsSelValid)
                return;

            switch (map_mode)
            {
                case MapMode.Sector: GlobalVars.ItemEdit.LoadSector(Level.sectors[Cur_SC]); break;
                case MapMode.Surface: GlobalVars.ItemEdit.LoadSurface(Level.sectors[Cur_SC].surfaces[Cur_SF]); break;
                case MapMode.Vertex: GlobalVars.ItemEdit.LoadVertex(Level.sectors[Cur_SC].vertices[Cur_VX]); break;
                case MapMode.Thing: GlobalVars.ItemEdit.LoadThing(Level.things[Cur_TH]); break;
                case MapMode.Light: GlobalVars.ItemEdit.LoadLight(Cur_LT); break;
                case MapMode.Edge: GlobalVars.ItemEdit.LoadEdge(Level.sectors[Cur_SC].surfaces[Cur_SF], Cur_ED); break;
                case MapMode.Frame: GlobalVars.ItemEdit.LoadFrame(Cur_TH, Cur_FR); break;
                case MapMode.Extra: GlobalVars.ItemEdit.LoadExtra(Cur_EX); break;
            }
        }

        public bool IsSelValid
        {
            get
            {
                switch (map_mode)
                {
                    case MapMode.Sector:
                        return (Cur_SC >= 0 && Cur_SC < Level.sectors.Count);

                    case MapMode.Surface:
                        if (Cur_SC < 0 || Cur_SC >= Level.sectors.Count)
                            return false;

                        return (Cur_SF >= 0 && Cur_SF < Level.sectors[Cur_SC].surfaces.Count);

                    case MapMode.Edge:
                        if (Cur_SC < 0 || Cur_SC >= Level.sectors.Count)
                            return false;

                        if (Cur_SF < 0 || Cur_SF >= Level.sectors[Cur_SC].surfaces.Count)
                            return false;

                        return (Cur_ED >= 0 && Cur_ED < Level.sectors[Cur_SC].surfaces[Cur_SF].vertices.Count);

                    case MapMode.Vertex:
                        if (Cur_SC < 0 || Cur_SC >= Level.sectors.Count)
                            return false;

                        return (Cur_SC >= 0 && Cur_VX < Level.sectors[Cur_SC].vertices.Count);

                    case MapMode.Thing:
                        return (Cur_TH >= 0 && Cur_TH < Level.things.Count);

                    case MapMode.Frame:
                        if (Cur_TH < 0 || Cur_TH >= Level.things.Count)
                            return false;

                        return (Cur_FR >= -1 && Cur_FR < Level.things[Cur_TH].vals.Count);

                    case MapMode.Light:
                        return (Cur_LT >= 0 && Cur_LT < Level.lights.Count);
                        
                    case MapMode.Extra:
                        return (Cur_EX >= 0 && Cur_EX < ExtraObjs.Count);

                    default:
                        return false;
                }
            }
        }

        public delegate void NotifyEvent(object obj);
        public delegate void ExtraMoveEvent(object obj, bool continued);

        public NotifyEvent OnExtraSelect = null;
        public ExtraMoveEvent OnExtraMove = null;

        public void FireOnExtraMove(object obj, bool continued)
        {
            if (OnExtraMove != null)
                OnExtraMove(obj, continued);
        }

        public MapMode map_mode = MapMode.Sector;

        public bool IsSCValid(int sc)
        {
            return (sc >= 0 && sc < Level.sectors.Count);
        }

        public bool IsSFValid(int sc, int sf)
        {
            if (sc < 0 || sc >= Level.sectors.Count)
                return false;

            return (sf >= 0 && sf < Level.sectors[sc].surfaces.Count);
        }

        public bool IsEDValid(int sc, int sf, int ed)
        {
            if (sc < 0 || sc >= Level.sectors.Count)
                return false;

            if (sf < 0 || sf >= Level.sectors[sc].surfaces.Count)
                return false;

            return (ed >= 0 && ed < Level.sectors[sc].surfaces[sf].vertices.Count);
        }

        public bool IsVXValid(int sc, int vx)
        {
            if (sc < 0 || sc >= Level.sectors.Count)
                return false;

            return (vx >= 0 && vx < Level.sectors[sc].vertices.Count);
        }

        public bool IsTHValid(int th)
        {
            return (th >= 0 && th < Level.things.Count);
        }

        public bool IsLTValid(int lt)
        {
            return (lt >= 0 && lt < Level.lights.Count);
        }

        public bool IsFRValid(int th, int fr)
        {
            if (th < 0 || th >= Level.things.Count)
                return false;

            return (fr >= -1 && fr < Level.things[th].vals.Count);// wtf lol
        }

        public void VerifyMultiSelection()
        {
            switch (map_mode)
            {
                case MapMode.Sector:
                    for (int i = scsel.Count - 1; i >= 0; i--)
                        if (!IsSCValid(scsel[i]))
                            scsel.DeleteN(i);
                    break;

                case MapMode.Surface:
                    for (int i = sfsel.Count - 1; i >= 0; i--)
                    {
                        int sc, sf;
                        sfsel.GetSCSF(i, out sc, out sf);
                        if (!IsSFValid(sc, sf))
                            sfsel.DeleteN(i);
                    }
                    break;

                case MapMode.Edge:
                    for (int i = edsel.Count - 1; i >= 0; i--)
                    {
                        int sc, sf, ed;
                        edsel.GetSCSFED(i, out sc, out sf, out ed);
                        if (!IsEDValid(sc, sf, ed))
                            edsel.DeleteN(i);
                    }
                    break;

                case MapMode.Vertex:
                    for (int i = vxsel.Count - 1; i >= 0; i--)
                    {
                        int sc, vx;
                        vxsel.GetSCVX(i, out sc, out vx);
                        if (!IsVXValid(sc, vx))
                            vxsel.DeleteN(i);
                    }
                    break;

                case MapMode.Thing:
                    for (int i = thsel.Count - 1; i >= 0; i--)
                        if (!IsTHValid(thsel.GetTH(i)))
                            thsel.DeleteN(i);
                    break;

                case MapMode.Frame:
                    for (int i = frsel.Count - 1; i >= 0; i--)
                    {
                        int n, fr;
                        frsel.GetTHFR(i, out n, out fr);
                        if (!IsFRValid(n, fr))
                            frsel.DeleteN(i);
                    }
                    break;

                case MapMode.Light:
                    for (int i = ltsel.Count - 1; i >= 0; i--)
                        if (!IsLTValid(ltsel.GetLT(i)))
                            ltsel.DeleteN(i);
                    break;
            }
        }

        public void SetMSelMode(MultiselectMode mode)
        {
            string mname;
            switch (mode)
            {
                case MultiselectMode.Toggle:
                    mname = "Toggle";
                    miToggle.Checked = true;
                    PMsel.Text = "*";
                    break;

                case MultiselectMode.Add:
                    mname = "Add";
                    miAdd.Checked = true;
                    PMsel.Text = "+";
                    break;

                case MultiselectMode.Subtract:
                    mname = "Subtract";
                    miSubtract.Checked = true;
                    PMsel.Text = "-";
                    break;

                default:
                    return;
            }

            msel_mode = mode;
            PMsel.ToolTipText = "Current multiselction mode is "+mname+" Click to change";
        }

        public MultiselectMode msel_mode = MultiselectMode.Toggle;

        public bool LightsHidden = false;

        public void SetHideLights(bool hide)
        {
            LightsHidden = hide;
            if (LightsHidden)
                HideLights.Text = "Show &Lights";
            else
                HideLights.Text = "Hide &Lights";
            Invalidate();
        }

        public bool ThingsHidden = false;

        public void SetHideThings(bool hide)
        {
            ThingsHidden = hide;
            if (ThingsHidden)
                HideThings.Text = "Show &Things";
            else
                HideThings.Text = "Hide &Things";
            Invalidate();
        }

        public bool ShowWeirdAdjoins = false;

        public void SetShowWeirdAdjoins(bool show)
        {
            ShowWeirdAdjoins = show;

            showWeirdAdjoinsIn3DPreviewToolStripMenuItem.Checked = ShowWeirdAdjoins;

            if(GlobalVars.Preview3D != null)
                GlobalVars.Preview3D.Redraw();
        }

        public void LoadPlugins()
        {
            Log.Error("LoadPlugins not implemented");
        }

        public void LoadTemplates()
        {
            if (!string.IsNullOrEmpty(GlobalVars.ProjectGob))
            {
                // not very efficient in performance but efficient in that we can reuse some simth code
                using(BaseFile bf = FileOperations.OpenGameFile(LevelFileJKL))
                {
                    MemoryStream ms = new MemoryStream(bf.Read(bf.Length));

                    Tokenizer t = new Tokenizer(ms);

                    if (t.SeekToLine("section: templates"))
                    //Log.Warning("Couldn't find section: templates");
                    //else
                    {
                        //Log.Info("Parsing section: templates");

                        t.EatToken(2);
                        int numTemplates = t.ReadInt();
                        //JK.AddTemplates(Templates.ToArray());
                        //Templates.AddRange(JK.ParseTemplates(t, numTemplates));
                        Templates.Clear();
                        foreach(Template src in JK.ParseTemplates(t, numTemplates))
                            Templates.All.Add(JKTemplate.FromSmithTemplate(src));
                    }                    
                }

                return;
            }

            // normal jed behavior
            Templates.LoadFromFile(GlobalVars.IsMOTS ? "mots.tpl" : "master.tpl");
        }

        public string ShortJKLName
        {
            get
            {
                return FileOperations.ChangeFileExt(FileOperations.ExtractName(LevelFile), ".jkl");
            }
        }

        public void SetThingView(ThingView how)
        {
            GlobalVars.ThingView = how;
            switch (how)
            {
                case ThingView.Dots: Dots.Checked = true; Boxes.Checked = false; Wireframes.Checked = false; break;
                case ThingView.Boxes: Dots.Checked = false; Boxes.Checked = true; Wireframes.Checked = false; break;
                case ThingView.Wireframes:
                    {
                        Dots.Checked = false;
                        Boxes.Checked = false;
                        Wireframes.Checked = true;
                        ProgressDialog.Reset(Level.things.Count, 0);
                        ProgressDialog.Message("Loading 3DOs...");
                        foreach (JKThing thing in Level.things)
                        {
                            UpdateThingData(thing);
                            ProgressDialog.Step();
                        }
                        ProgressDialog.Cleanup();
                    }
                    break;
            }
            Invalidate();
        }

        public void UpdateThingData(JKThing thing)
        {
            if (GlobalVars.ThingView != ThingView.Wireframes)
            {
                int itpl = Templates.IndexOfName(thing.name);
                if (itpl == -1)
                    thing.box = Box3.Empty;
                else
                    thing.box = Templates.Get(itpl).box;

                if (GlobalVars.Preview3D.Visible && GlobalVars.P3DThings)
                    LoadThing3DO(thing, true);
            }
            else
            {
                LoadThing3DO(thing, true);
            }
        }

        public void LoadThing3DO(JKThing thing, bool force)
        {
            if (!force && thing.model != null)
                return;

            int itpl = Templates.IndexOfName(thing.name);
            if (itpl == -1)
            {
                thing.model = null;
                return;
            }

            TemplateValue tval = Templates.GetTPLField(thing.name, "model3d");
            if (tval == null)
            {
                thing.model = null;
                return;
            }

            thing.model = JKModel.Load3DO(tval.AsString);

            if (thing.model != null)
                thing.box = thing.model.GetBBox();
        }

        public MapSettings Toolbar
        {
            get
            {
                return mapSettings1;
            }
        }

        public bool changed = false;

        private void NedMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            e.Cancel = !AskSave();
            if (e.Cancel)
                return;

            if (SmithMain.SmithRunning)
            {
                Program.MainForm.Close();
            }

            Audio.Channel chan = null;
            /*if (Program.LaunchSounds)
            {
                chan = Audio.PlaySound("i00ky40z.wav", Audio.DimensionMode._2D, false, true);
                if (chan != null)
                {
                    chan.Volume = 0.5;
                    chan.Play();
                }
            }*/

            renderControl1.Shutdown();

            NedRunning = false;

            disableLightCalc = true;

            // allow all sub-forms to close.. we'll re-clear this when launching ned
            SubForm.AllowClose = true;

            if (chan != null)
            {
                while (chan.IsPlaying)
                    System.Threading.Thread.Sleep(100);
            }
        }

        private void PNProjType_DoubleClick(object sender, EventArgs e)
        {
            Options.SetMOTS(!GlobalVars.IsMOTS);
            SetMotsIndicator();
            LoadTemplates();
        }

        public void SetMotsIndicator()
        {
            if (GlobalVars.IsMOTS)
            {
                PNProjType.Text = "MOTS";
                PNProjType.ToolTipText = "Current project is MOTS project (double-click to change)";
            }
            else
            {
                PNProjType.Text = "JK";
                PNProjType.ToolTipText = "Current project is JK project (double-click to change)";
            }
        }

        private void SaveTimer_Tick(object sender, EventArgs e)
        {
            if (string.IsNullOrEmpty(GlobalVars.ProjectDir))
                return;

            if (!Path.GetExtension(LevelFile).Equals(".jed", StringComparison.InvariantCultureIgnoreCase))
                return;

            string autosaveDir = Path.Combine(GlobalVars.ProjectDir, "autosave");
            Directory.CreateDirectory(autosaveDir);

            string fname = Path.Combine(autosaveDir, Path.GetFileName(LevelFile));
            Level.SaveToJED(fname);
            Utils.PanMessage(Utils.MessageType.Info, "Project autosaved to " + fname);
        }


        public void performopen(string file)
        {
            // CHECK FOR "OPEN" and call OpenProject
            foreach (string ext in new string[] { ".jkl", ".jed", ".lev" })
                if (file.EndsWith(ext, StringComparison.InvariantCultureIgnoreCase))
                {
                    OpenProject(file, OpenMethod.Open);
                    return;
                }

            // else call ImportProject
            ImportProject(file);
        }


        private void nedopenmenu(object sender, EventArgs e)
        {
            if (!AskSave())
                return;

            string file = string.Empty;


            if (!FileDialogs.Execute("Most things|*.JKL;*.JED;*.GOB;*.GOO;*.lev;*.gob;*.asc;*.3do;*.3ds|Jed files(*.JED)|*.JED|JK Files (*goo;*.gob;*.jkl)|*.GOO;*.GOB;*.JKL|DF Files(*.LEV;*.GOB)|*.LEV;*.GOB|3D Studio ASC|*.asc|JK 3DOs|*.3do;*.gob|Autodesk 3DS|*.3ds", ref file))
                return;

            performopen(file);          
        }

        private void OpenProject(string fname, OpenMethod how)
        {
            string ext = FileOperations.ExtractExt(fname).ToUpper();

            ResettingEditor();

            PMsg.Text = string.Empty;
            _AllSectors3DOPreview = null;

            if (ext == ".JKL")
            {
                LevelFile = fname;
                Level.LoadFromJKL(fname);
                NewLevel = false;
                AskSaveAs = Utils.IsInContainer(fname);
                SetLevelName();
            }
            else if (ext == ".JED")
            {
                LevelFile = fname;
                Level.LoadFromJED(fname);
                NewLevel = false;
                AskSaveAs = Utils.IsInContainer(fname);
                SetLevelName();
            }

            Options.SetMOTS(Level.mots);
            Level.JKLPostLoad();
            AddRecent(fname);
            ResetEditor(how != OpenMethod.Revert);
            Invalidate();

            lightsDirty = false;
        }

        private void ImportProject(string file)
        {
            ResettingEditor();

            LevelFile = "Untitled.jed";
            _AllSectors3DOPreview = null;

            bool wantToAutoSaveProject = false;
            bool wantToAutoExport3DO = false;

            string s = FileOperations.ExtractExt(file).ToLowerInvariant();
            if (s == ".3do")
            {
                LevelFile = file;
                Import3DO(file);
            }
            else if (s == ".asc")
            {
                MessageBox.Show("Import ASC not supported currently");
                //Level.ImportASC(ofd.FileName);
            }
            else if (s == ".3ds")
            {
                LevelFile = file;
                if (!Import3DS(file))
                {
                    NewProject();
                    return;
                }

                wantToAutoSaveProject = true;
                wantToAutoExport3DO = true;
            }
            else
            {
                MessageBox.Show("Import DF files not supported currently");
                /*scale = 40.0;
                s = "40";
                if(DFImport.ShowDialog() != */
            }

            PMsg.Text = string.Empty;
            NewLevel = true;
            ResetEditor(true);
            Level.mots = isMots;
            // presumably this is only needed for importing DF levels.. so i'll leave it out for now
            /*  for i:=0 to Level.Things.count-1 do
  begin
   th:=Level.Things[i];
   th.z:=th.z-th.bbox.z1;
   nsec:=FindSectorForThing(th);
   if nsec<>-1 then
    th.Sec:=Level.Sectors[nsec];
   if th.sec=nil then th.layer:=Level.AddLayer('Orphan things')
   else
   begin
    s:=Level.GetLayerName(th.sec.layer);
    s:='Things'+Copy(s,6,length(s));
    th.layer:=Level.AddLayer(s);
   end;
  end; */

            Toolbar.LoadLayers();


            if (wantToAutoSaveProject)
                SysSaveProject(false);

            if (wantToAutoExport3DO)
            {
                string out3dofile = Path.Combine(Path.Combine(GlobalVars.ProjectDir, "3do"), Path.GetFileNameWithoutExtension(LevelFile) + ".3do");
                if (MessageBox.Show($"Would you like to export {out3dofile} automatically?", "Autoexport model", MessageBoxButtons.OKCancel) == DialogResult.OK)
                {
                    // auto select all sectors
                    for (int x = 0; x < Level.sectors.Count; x++)
                        scsel.AddSC(x);

                    // invoke export 3do, without dialog


                    ExportModel(out3dofile, ExportModelMode.JK3do);


                    // unselect all sectors
                    scsel.Clear();
                }
            }
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            DoKeyDown(e);
        }

        public void CancelMouseMode()
        {
            switch (mouse_mode)
            {
                case MouseMode.Cleave:
                    CleaveStarted = false;
                    break;

                case MouseMode.RectSelect:
                    RectSelStarted = false;
                    break;
            }

            SetMouseMode(MouseMode.Select);
        }

        public void ClearMultiSelection()
        {
            switch (map_mode)
            {
                case MapMode.Sector: scsel.Clear(); break;
                case MapMode.Surface: sfsel.Clear(); break;
                case MapMode.Edge: edsel.Clear(); break;
                case MapMode.Vertex: vxsel.Clear(); break;
                case MapMode.Thing: thsel.Clear(); break;
                case MapMode.Frame: frsel.Clear(); break;
                case MapMode.Light: ltsel.Clear(); break;
            }
        }

        public Orient camOrient = Orient.Zero;

        void RotAxis(ref Vec3 x, ref Vec3 y, int dir)
        {
            if (dir > 0)
                RotAxisAngle(ref x, ref y, 5.0);
            else if (dir < 0)
                RotAxisAngle(ref x, ref y, -5.0);
        }

        void RotAxisAngle(ref Vec3 x, ref Vec3 y, double ang)
        {
            double ca = Math.Cos(MathLib.ToRad(ang));
            double sa = Math.Sin(MathLib.ToRad(ang));

            Vec3 nx = new Vec3(ca * x.x + sa * y.x,
                               ca * x.y + sa * y.y,
                               ca * x.z + sa * y.z);

            Vec3 ny = new Vec3(-sa * x.x + ca * y.x,
                               -sa * x.y + ca * y.y,
                               -sa * x.z + ca * y.z);

            x = nx;
            y = ny;
        }


        public void SetRendFromPYR()
        {
            /* CreateRotMatrix(mx,pch,yaw,rol);
 MultVM3(mx,vec.dx,vec.dy,vec.dz);   */

            Vec3 x = Vec3.PosX;
            Utils.RotateVector(ref x, camOrient.pitch, 0.0, 0.0);
            Utils.RotateVector(ref x, 0.0, camOrient.roll/*WTF*/, 0.0);
            Utils.RotateVector(ref x, 0.0, 0.0, camOrient.yaw/*WTF*/);

            Vec3 z = Vec3.PosZ;
            Utils.RotateVector(ref z, camOrient.pitch, 0.0, 0.0);
            Utils.RotateVector(ref z, 0.0, camOrient.roll/*WTF*/, 0.0);
            Utils.RotateVector(ref z, 0.0, 0.0, camOrient.yaw/*WTF*/);

            renderer.SetZ(z);
            renderer.SetX(x);

            // not sure if this is right.. hope so

            // yeah no it doesnt seem right
            //renderer.SetZ(Vec3.PosY.Transformed(camOrient.Matrix.Inverse));
            //renderer.SetX(Vec3.PosX.Transformed(camOrient.Matrix.Inverse));

            /* SetVec(x,1,0,0);
 SetVec(y,0,1,0);
 SetVec(z,0,0,1);
 {PCH,ROL,YAW}
 RotateVector(x,rPCH,0,0); RotateVector(x,0,rROL,0); RotateVector(x,0,0,rYAW);
 RotateVector(y,rPCH,0,0); RotateVector(y,0,rROL,0); RotateVector(y,0,0,rYAW);
 RotateVector(z,rPCH,0,0); RotateVector(z,0,rROL,0); RotateVector(z,0,0,rYAW);
 {So PCH - x , YAW - y, ROL - z}
 Renderer.SetZ(z.dx,z.dy,z.dz);
 Renderer.SetX(x.dx,x.dy,x.dz);      */
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            DoMouseMove(e);
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            DoMouseDown(e);
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            DoMouseUp(e);
        }

        Vec2 lastSelXY = Vec2.Zero;

        public int csel = -1;

        void Do_SelectAt(Vec2 pt, bool fore)
        {
            if (renderer == null)
                return;

            using (renderer.Context)
            {
                if (lastSelXY == pt && renderer.Selected.Count != 0)
                {
                    if (fore)
                    {
                        if (csel > renderer.Selected.Count - 1)
                            csel = 0;
                        else
                            csel++;
                    }
                    else
                    {
                        if (csel == 0)
                            csel = renderer.Selected.Count - 1;
                        else
                            csel--;
                    }
                }
                else
                    csel = -1;

                if (csel == -1)
                {
                    renderer.BeginPick(pt);
                    renderer.SetPointSize(4.0);
                    switch (map_mode)
                    {
                        case MapMode.Sector:
                            for (int s = 0; s < Level.sectors.Count; s++)
                            {
                                JKSector sec = Level.sectors[s];
                                if (Toolbar.IsLayerVisible(sec.layer))
                                    renderer.PickPolygons(sec.surfaces, s);
                            }
                            break;

                        case MapMode.Surface:
                            for (int s = 0; s < Level.sectors.Count; s++)
                            {
                                JKSector sec = Level.sectors[s];
                                if (Toolbar.IsLayerVisible(sec.layer))
                                    for (int sf = 0; sf < sec.surfaces.Count; sf++)
                                        renderer.PickPolygon(sec.surfaces[sf], s + sf * 65536);
                            }
                            break;

                        case MapMode.Edge:
                            for (int s = 0; s < Level.sectors.Count; s++)
                            {
                                JKSector sec = Level.sectors[s];
                                if (Toolbar.IsLayerVisible(sec.layer))
                                    for (int sf = 0; sf < sec.surfaces.Count; sf++)
                                    {
                                        JKSurface surf = sec.surfaces[sf];
                                        for (int ed = 0; ed < surf.vertices.Count; ed++)
                                        {
                                            Vec3 v1 = surf.vertices[ed].v;
                                            Vec3 v2 = surf.vertices[surf.NextVX(ed)].v;
                                            renderer.PickLine(v1, v2, s * 65536 + ed * 1024 + sf);
                                        }
                                    }
                            }
                            break;

                        case MapMode.Thing:
                            for (int th = 0; th < Level.things.Count; th++)
                            {
                                JKThing thing = Level.things[th];
                                if (Toolbar.IsLayerVisible(thing.layer))
                                    renderer.PickVertex(thing.pos, th);
                            }
                            break;

                        case MapMode.Frame:
                            for (int th = 0; th < Level.things.Count; th++)
                            {
                                JKThing thing = Level.things[th];
                                if (!Toolbar.IsLayerVisible(thing.layer))
                                    continue;

                                renderer.PickVertex(thing.pos, th * 65536 + 65535);
                                for (int fr = 0; fr < thing.vals.Count; fr++)
                                {
                                    TemplateValue val = thing.vals[fr];
                                    if (val.adjType != AdjType.Frame)
                                        continue;

                                    Vec3 fv;
                                    Orient fo;
                                    val.GetFrame(out fv, out fo);
                                    renderer.PickVertex(fv, th * 65536 + fr);
                                }
                            }
                            break;

                        case MapMode.Vertex:
                            for (int s = 0; s < Level.sectors.Count; s++)
                            {
                                JKSector sec = Level.sectors[s];
                                if (Toolbar.IsLayerVisible(sec.layer))
                                    for (int vx = 0; vx < sec.vertices.Count; vx++)
                                        renderer.PickVertex(sec.vertices[vx].v, s + vx * 65536);
                            }
                            break;

                        case MapMode.Light:
                            for (int th = 0; th < Level.lights.Count; th++)
                            {
                                JedLight light = Level.lights[th];
                                if (Toolbar.IsLayerVisible(light.layer))
                                    renderer.PickVertex(light.Pos, th);
                            }
                            break;

                        case MapMode.Extra:
                            for (int s = 0; s < ExtraObjs.Count; s++)
                            {
                                object obj = ExtraObjs[s];
                                if (obj is Vertex)
                                    renderer.PickVertex((obj as Vertex).v, s);
                                else if (obj is ExtraLine)
                                    renderer.PickLine((obj as ExtraLine).v1, (obj as ExtraLine).v2, s);
                                else if (obj is Polygon)
                                    renderer.PickPolygon(obj as Polygon, s);
                            }
                            break;
                    }

                    renderer.EndPick();
                    if (renderer.Selected.Count != 0)
                        csel = 0;
                }

                
                if(renderer.Selected.Count != 0)
                {
                    switch (map_mode)
                    {
                        case MapMode.Sector:
                            SetCurSC(renderer.Selected[csel]);
                            break;

                        case MapMode.Surface:
                            Cur_SC = renderer.Selected[csel] % 65536;
                            Cur_SF = renderer.Selected[csel] / 65536;
                            SetCurSF(Cur_SC, Cur_SF);
                            break;

                        case MapMode.Edge:
                            {
                                int vx = renderer.Selected[csel];
                                int s = vx / 65536;
                                int sf = vx % 1024;
                                int ed = (vx % 65536) / 1024;
                                SetCurED(s, sf, ed);
                            }
                            break;

                        case MapMode.Thing:
                            SetCurTH(renderer.Selected[csel]);
                            break;

                        case MapMode.Frame:
                            Cur_FR = renderer.Selected[csel] % 65536;
                            Cur_TH = renderer.Selected[csel] / 65536;
                            if (Cur_FR == 65535)
                                Cur_FR = -1;
                            SetCurFR(Cur_TH, Cur_FR);
                            break;

                        case MapMode.Vertex:
                            Cur_SC = renderer.Selected[csel] % 65536;
                            Cur_VX = renderer.Selected[csel] / 65536;
                            SetCurVX(Cur_SC, Cur_VX);
                            break;

                        case MapMode.Light:
                            SetCurLT(renderer.Selected[csel]);
                            break;

                        case MapMode.Extra:
                            SetCurEX(renderer.Selected[csel]);
                            break;
                    }
                }

                lastSelXY = pt;
            }
        }

        public void Do_SelSC(int sc)
        {
            int i;
            switch (msel_mode)
            {
                case MultiselectMode.Toggle:
                    i = scsel.AddSC(sc);
                    scsel.DeleteN(i ^ MultiSel.NBit);
                    break;

                case MultiselectMode.Add:
                    scsel.AddSC(sc);
                    break;

                case MultiselectMode.Subtract:
                    i = scsel.FindSC(sc);
                    if (i != -1)
                        scsel.DeleteN(i);
                    break;
            }
        }

        public void Do_SelSF(int sc, int sf)
        {
            int i;
            switch (msel_mode)
            {
                case MultiselectMode.Toggle:
                    i = sfsel.AddSF(sc, sf);
                    sfsel.DeleteN(i ^ MultiSel.NBit);
                    break;

                case MultiselectMode.Add:
                    sfsel.AddSF(sc, sf);
                    break;

                case MultiselectMode.Subtract:
                    i = sfsel.FindSF(sc, sf);
                    if (i != -1)
                        sfsel.DeleteN(i);
                    break;
            }
        }

        public void Do_SelVX(int sc, int vx)
        {
            int i;
            switch (msel_mode)
            {
                case MultiselectMode.Toggle:
                    i = vxsel.AddVX(sc, vx);
                    vxsel.DeleteN(i ^ MultiSel.NBit);
                    break;

                case MultiselectMode.Add:
                    vxsel.AddVX(sc, vx);
                    break;

                case MultiselectMode.Subtract:
                    i = vxsel.FindVX(sc, vx);
                    if (i != -1)
                        vxsel.DeleteN(i);
                    break;
            }
        }

        public void Do_SelED(int sc, int sf, int ed)
        {
            int i;
            switch (msel_mode)
            {
                case MultiselectMode.Toggle:
                    i = edsel.AddED(sc, sf, ed);
                    edsel.DeleteN(i ^ MultiSel.NBit);
                    break;

                case MultiselectMode.Add:
                    edsel.AddED(sc, sf, ed);
                    break;

                case MultiselectMode.Subtract:
                    i = edsel.FindED(sc, sf, ed);
                    if (i != -1)
                        edsel.DeleteN(i);
                    break;
            }
        }

        public void Do_SelTH(int th)
        {
            int i;
            switch (msel_mode)
            {
                case MultiselectMode.Toggle:
                    i = thsel.AddTH(th);
                    thsel.DeleteN(i ^ MultiSel.NBit);
                    break;

                case MultiselectMode.Add:
                    thsel.AddTH(th);
                    break;

                case MultiselectMode.Subtract:
                    i = thsel.FindTH(th);
                    if (i != -1)
                        thsel.DeleteN(i);
                    break;
            }
        }

        public void Do_SelFR(int th, int fr)
        {
            int i;
            switch (msel_mode)
            {
                case MultiselectMode.Toggle:
                    i = frsel.AddFR(th, fr);
                    frsel.DeleteN(i ^ MultiSel.NBit);
                    break;

                case MultiselectMode.Add:
                    frsel.AddFR(th, fr);
                    break;
                    
                case MultiselectMode.Subtract:
                    i = frsel.FindFR(th, fr);
                    if (i != -1)
                        frsel.DeleteN(i);
                    break;
            }
        }

        public void Do_SelLT(int lt)
        {
            int i;
            switch (msel_mode)
            {
                case MultiselectMode.Toggle:
                    i = ltsel.AddLT(lt);
                    ltsel.DeleteN(i ^ MultiSel.NBit);
                    break;

                case MultiselectMode.Add:
                    ltsel.AddLT(lt);
                    break;

                case MultiselectMode.Subtract:
                    i = ltsel.FindLT(lt);
                    if (i != -1)
                        ltsel.DeleteN(i);
                    break;
            }
        }

        public void Do_MultiSelect()
        {
            switch (map_mode)
            {
                case MapMode.Sector: Do_SelSC(Cur_SC); break;
                case MapMode.Surface: Do_SelSF(Cur_SC, Cur_SF); break;
                case MapMode.Edge: Do_SelED(Cur_SC, Cur_SF, Cur_ED); break;
                case MapMode.Vertex: Do_SelVX(Cur_SC, Cur_VX); break;
                case MapMode.Thing: Do_SelTH(Cur_TH); break;
                case MapMode.Frame: Do_SelFR(Cur_TH, Cur_FR); break;
                case MapMode.Light: Do_SelLT(Cur_LT); break;
            }
        }

        bool SCCreateStarted = false;

        Vec3 CleaveOrg = Vec3.Zero;

        public bool SnapToGrid = true;

        public void Do_EndCreateSC(Vec2 pt, bool snapToVertex)
        {
            try
            {
                if (!SCCreateStarted)
                    return;

                Vec3 v;
                if (!renderer.GetGridAt(pt, out v))
                    return;

                if (snapToVertex)
                    Utils.GetNearestVX(Level, ref v, 0.2);
                else
                    if (SnapToGrid)
                        renderer.GetNearestGrid(ref v);

                Vec3 nx = v - CleaveOrg;
                double vl = nx.Magnitude;

                JKSector sc = Toolbar.GetNewShape();

                if (vl < 0.1)
                    nx = renderer.GridXVec;
                else
                {
                    Box3 box = Utils.FindBBox(sc);
                    vl = box.Right - box.Left;
                    nx /= vl;
                }

                vl = nx.Magnitude;

                Vec3 nz = renderer.GridNormal * vl;

                Vec3 ny = (nz ^ nx).Unit * vl;

                JKSector sec = Level.sectors[Cur_SC];
                JKSector nsc = Level.NewSector();
                Utils.DuplicateSector(sc, nsc, Vec3.Zero, nx, ny, nz, v);
                nsc.Assign(sec);
                Level.sectors.Add(nsc);
                Level.RenumSecs();

                Undo.StartUndoRec("Create sector");
                Undo.SaveSecUndo(nsc, Undo.Change.Added, Undo.SectorChange.Both);

                SectorAdded(nsc);
                SetCurSC(Level.sectors.Count - 1);
            }
            finally
            {
                SetMouseMode(MouseMode.Select);
                SCCreateStarted = false;
            }
        }

        public void SectorAdded(JKSector sec)
        {
            _AllSectors3DOPreview = null;

            GlobalVars.Preview3D.AddSector(sec);
            LevelChanged(true);
        }

        public void SectorChanged(JKSector sec, bool lights)
        {
            _AllSectors3DOPreview = null;

            GlobalVars.Preview3D.UpdateSector(sec);
            LevelChanged(lights);
        }

        public void SectorDeleted(JKSector sec)
        {
            _AllSectors3DOPreview = null;

            GlobalVars.Preview3D.DeleteSector(sec);
            LevelChanged(true);
        }

        public void ThingChanged(JKThing th)
        {
            UpdateThingData(th);
            GlobalVars.Preview3D.UpdateThing(th);
            LevelChanged(false);
        }

        public void ThingAdded(JKThing th)
        {
            UpdateThingData(th);
            GlobalVars.Preview3D.AddThing(th);
            LevelChanged(false);
        }

        public void ThingDeleted(JKThing th)
        {
            GlobalVars.Preview3D.DeleteThing(th);
            LevelChanged(false);
        }

        public void LevelChanged(bool lights)
        {
            if (lights)
                lightsDirty = true;

            changed = true;
        }

        Vec2 CleaveOrgXY = Vec2.Zero;
        bool CleaveStarted = false;

        Vec2 LastXY = Vec2.Zero;
        Vec2 DrawToXY = Vec2.Zero;

        public void Do_StartCleave(Vec2 pt, bool snaptovertex)
        {
            if (CleaveStarted)
                Redraw();

            CleaveOrgXY = pt;

            if (!renderer.GetGridAt(CleaveOrgXY, out CleaveOrg))
                return;

            if (snaptovertex)
                Utils.GetNearestVX(Level, ref CleaveOrg, 0.2);
            else if (SnapToGrid)
                renderer.GetNearestGrid(ref CleaveOrg);

            CleaveOrgXY = renderer.ProjectPoint(CleaveOrg);

            LastXY = CleaveOrgXY;
            DrawToXY = LastXY;
            SetMouseMode(MouseMode.Cleave);
            CleaveStarted = true;
        }

        public void Do_StartCreateSC(Vec2 pt, bool snaptovertex)
        {
            if (SCCreateStarted)
                Redraw();

            CleaveOrgXY = pt;

            if (!renderer.GetGridAt(CleaveOrgXY, out CleaveOrg))
                return;

            if (snaptovertex)
                Utils.GetNearestVX(Level, ref CleaveOrg, 0.2);
            else if (SnapToGrid)
                renderer.GetNearestGrid(ref CleaveOrg);

            CleaveOrgXY = renderer.ProjectPoint(CleaveOrg);

            LastXY = CleaveOrgXY;
            DrawToXY = LastXY;
            SetMouseMode(MouseMode.CreateSector);
            SCCreateStarted = true;
        }

        public void DoMouseDown(MouseEventArgs e)
        {
            // dont use the location passed in since we have this whole mess between NedMain form client rectangle VS rendercontrol client rectangle
            Vec2 pt = new Vec2(renderControl1.PointToClient(Cursor.Position));

            if (e.Button == MouseButtons.Middle)
            {
                Vec2? mpt = GetMousePos();
                if (mpt.HasValue)
                    Do_StartRotateCam(mpt.Value);

                return;
            }

            switch (mouse_mode)
            {
                case MouseMode.Select:
                    break;

                case MouseMode.Cleave:
                    Do_StartCleave(pt, (ModifierKeys & Keys.Shift) != 0);
                    break;

                case MouseMode.CreateSector:
                    Do_StartCreateSC(pt, (ModifierKeys & Keys.Shift) != 0);
                    break;
            }
        }

        public void DoMouseUp(MouseEventArgs e)
        {
            // dont use the location passed in since we have this whole mess between NedMain form client rectangle VS rendercontrol client rectangle
            Vec2 pt = new Vec2(renderControl1.PointToClient(Cursor.Position));

            if (e.Button == MouseButtons.Middle)
            {
                SetMouseMode(MouseMode.Select);
                return;
            }

            switch (mouse_mode)
            {
                case MouseMode.Select:
                    Do_SelectAt(pt, (ModifierKeys & Keys.Alt) != 0);
                    if ((ModifierKeys & Keys.Shift) != 0)
                        Do_MultiSelect();
                    break;

                case MouseMode.CreateSector:
                    Do_EndCreateSC(pt, (ModifierKeys & Keys.Shift) != 0);
                    break;

                case MouseMode.Drag:
                    Do_EndDrag(pt);
                    break;

                case MouseMode.Cleave:
                    Do_EndCleave(pt, (ModifierKeys & Keys.Shift) != 0);
                    break;

                case MouseMode.RectSelect:
                    Do_EndRectSelect(pt);
                    break;
            }
        }

        void Do_EndRectSelect(Vec2 pt)
        {
            renderer.BeginRectPick(RectOrgXY, pt);
            switch (map_mode)
            {
                case MapMode.Sector:
                    {
                        for (int i = 0; i < Level.sectors.Count; i++)
                        {
                            JKSector sec = Level.sectors[i];
                            if (!Toolbar.IsLayerVisible(sec.layer))
                                continue;

                            bool allin = true;
                            foreach(JKVertex vert in sec.vertices)
                                if (!renderer.IsVertexInRect(vert.v))
                                {
                                    allin = false;
                                    break;
                                }

                            if (allin)
                                Do_SelSC(i);
                        }
                    }
                    break;

                case MapMode.Surface:
                    {
                        for (int i = 0; i < Level.sectors.Count; i++)
                        {
                            JKSector sec = Level.sectors[i];
                            if (!Toolbar.IsLayerVisible(sec.layer))
                                continue;

                            for (int sf = 0; sf < sec.surfaces.Count; sf++)
                                if (renderer.IsPolygonInRect(sec.surfaces[sf]))
                                    Do_SelSF(i, sf);
                        }
                    }
                    break;

                case MapMode.Edge:
                    {
                        for (int i = 0; i < Level.sectors.Count; i++)
                        {
                            JKSector sec = Level.sectors[i];
                            if (!Toolbar.IsLayerVisible(sec.layer))
                                continue;

                            for (int sf = 0; sf < sec.surfaces.Count; sf++)
                            {
                                JKSurface surf = sec.surfaces[sf];
                                for (int ed = 0; ed < surf.vertices.Count; ed++)
                                    if (renderer.IsLineInRect(surf.vertices[ed].v, surf.vertices[surf.NextVX(ed)].v))
                                        Do_SelED(i, sf, ed);
                            }
                        }
                    }
                    break;

                case MapMode.Vertex:
                    {
                        for (int i = 0; i < Level.sectors.Count; i++)
                        {
                            JKSector sec = Level.sectors[i];
                            if (!Toolbar.IsLayerVisible(sec.layer))
                                continue;

                            for (int vx = 0; vx < sec.vertices.Count; vx++)
                                if (renderer.IsVertexInRect(sec.vertices[vx].v))
                                    Do_SelVX(i, vx);
                        }
                    }
                    break;

                case MapMode.Thing:
                    {
                        for (int i = 0; i < Level.things.Count; i++)
                        {
                            JKThing thing = Level.things[i];
                            if (!Toolbar.IsLayerVisible(thing.layer))
                                continue;

                            if (renderer.IsVertexInRect(thing.pos))
                                Do_SelTH(i);
                        }
                    }
                    break;

                case MapMode.Light:
                    {
                        for (int i = 0; i < Level.lights.Count; i++)
                        {
                            JedLight light = Level.lights[i];
                            if (!Toolbar.IsLayerVisible(light.layer))
                                continue;

                            if (renderer.IsVertexInRect(light.Pos))
                                Do_SelLT(i);
                        }
                    }
                    break;

                case MapMode.Frame:
                    {
                        for (int i = 0; i < Level.things.Count; i++)
                        {
                            JKThing thing = Level.things[i];
                            if (!Toolbar.IsLayerVisible(thing.layer))
                                continue;

                            for (int fr = 0; fr < thing.vals.Count; fr++)
                            {
                                TemplateValue val = thing.vals[fr];
                                if (val.adjType != AdjType.Frame)
                                    continue;

                                Vec3 fp;
                                Orient fo;
                                val.GetFrame(out fp, out fo);
                                if (renderer.IsVertexInRect(fp))
                                    Do_SelFR(i, fr);
                            }
                        }
                    }
                    break;
            }
            renderer.EndRectPick();
            RectSelStarted = false;
            SetMouseMode(MouseMode.Select);
        }

        void Do_EndDrag(Vec2 pt)
        {
            SetMouseMode(MouseMode.Select);
        }

        void Do_EndCleave(Vec2 pt, bool snaptovertex)
        {
            if (!CleaveStarted)
                return;

            Vec3 v2;
            if (!renderer.GetGridAt(pt, out v2))
                return;

            if (snaptovertex)
                Utils.GetNearestVX(Level, ref v2, 0.2);
            else if (SnapToGrid)
                renderer.GetNearestGrid(ref v2);

            Vec3 v1 = CleaveOrg;

            if (MathLib.CompareScalar((v2 - v1).Magnitude, 0.0) == 0)
                return;

            Vec3 cnormal = (renderer.GridNormal ^ (v2 - v1)).Unit;
            //bool cleaved = false;
            CleaveBy(cnormal, v1);

            SetMouseMode(MouseMode.Select);
            CleaveStarted = false;
        }

        public void CleaveBy(Vec3 norm, Vec3 v)
        {
            switch (map_mode)
            {
                case MapMode.Sector:
                    Undo.StartUndoRec("Cleave sector(s)");
                    scsel.AddSC(Cur_SC);
                    for (int i = 0; i < scsel.Count; i++)
                        Utils.CleaveSector(Level.sectors[scsel.GetSC(i)], norm, v);
                    scsel.Clear();
                    break;

                case MapMode.Surface:
                    Undo.StartUndoRec("Cleave surface(s)");
                    sfsel.AddSF(Cur_SC, Cur_SF);
                    for (int i = 0; i < sfsel.Count; i++)
                    {
                        int nsc, nsf;
                        sfsel.GetSCSF(i, out nsc, out nsf);
                        Utils.CleaveSurface(Level.sectors[nsc].surfaces[nsf], norm, v);
                    }
                    sfsel.Clear();
                    break;

                case MapMode.Edge:
                    Undo.StartUndoRec("Cleave Edge");
                    Utils.CleaveEdge(Level.sectors[Cur_SC].surfaces[Cur_SF], Cur_ED, norm, v);
                    break;
            }
        }

        Vec3 DragOrg = Vec3.Zero;
        Vec3 DragOrgPos = Vec3.Zero;

        private void Do_StartDrag(Vec2 pt)
        {
            Vec3 f;
            if (!renderer.GetGridAt(pt, out f))
                return;

            SetMouseMode(MouseMode.Drag);
            DragOrg = f;

            DragOrgPos = GetCurObjRefXYZ();

            switch (map_mode)
            {
                case MapMode.Sector: Undo.StartUndoRec("Drag sector(s)"); break;
                case MapMode.Surface: Undo.StartUndoRec("Drag surface(s)"); break;
                case MapMode.Edge: Undo.StartUndoRec("Drag edge(s)"); break;
                case MapMode.Vertex: Undo.StartUndoRec("Drag vertices"); break;
                case MapMode.Light: Undo.StartUndoRec("Drag light(s)"); break;
                case MapMode.Thing: Undo.StartUndoRec("Drag thing(s)"); break;
                case MapMode.Frame: Undo.StartUndoRec("Drag frame(s)"); break;
                case MapMode.Extra:
                    if (OnExtraMove != null)
                        OnExtraMove(ExtraObjs[Cur_EX], false);
                    break;
            }
        }

        public Vec3 GetCurObjRefXYZ()
        {
            switch (map_mode)
            {
                case MapMode.Sector:
                    return Level.sectors[Cur_SC].vertices[0].v;

                case MapMode.Surface:
                    return Level.sectors[Cur_SC].surfaces[Cur_SF].vertices[0].v;

                case MapMode.Vertex:
                    return Level.sectors[Cur_SC].vertices[Cur_VX].v;

                case MapMode.Edge:
                    return Level.sectors[Cur_SC].surfaces[Cur_SF].vertices[Cur_ED].v;

                case MapMode.Thing:
                    return Level.things[Cur_TH].pos;

                case MapMode.Frame:
                    if (Cur_FR == -1)
                        return Level.things[Cur_TH].pos;
                    else
                    {
                        Vec3 r;
                        Orient o;
                        Level.things[Cur_TH].vals[Cur_FR].GetFrame(out r, out o);
                        return r;
                    }

                case MapMode.Light:
                    return Level.lights[Cur_LT].Pos;

                case MapMode.Extra:
                    {
                        object ex = ExtraObjs[Cur_EX];
                        if (ex is Vertex)
                            return (ex as Vertex).v;
                        if (ex is ExtraLine)
                            return (ex as ExtraLine).v1;
                        if (ex is Polygon)
                            return (ex as Polygon).vertices[0].v;
                    }
                    break;
            }

            return Vec3.Zero;
        }

        bool RectSelStarted = false;
        Vec2 RectOrgXY = Vec2.Zero;
        Vec2 RectLastXY = Vec2.Zero;
        public void Do_StartRectSelect(Vec2 pt)
        {
            RectSelStarted = true;
            RectOrgXY = pt;
            RectLastXY = RectOrgXY;
            SetMouseMode(MouseMode.RectSelect);
        }

        public void DoMouseMove(MouseEventArgs e)
        {
            // dont use the location passed in since we have this whole mess between NedMain form client rectangle VS rendercontrol client rectangle
            Vec2 pt = new Vec2(renderControl1.PointToClient(Cursor.Position));

            switch (mouse_mode)
            {
                case MouseMode.Select:
                    if (e.Button == System.Windows.Forms.MouseButtons.Left && (ModifierKeys & Keys.Control) != 0)
                        Do_StartDrag(pt);
                    else if (e.Button == System.Windows.Forms.MouseButtons.Left && (ModifierKeys & Keys.Alt) != 0)
                        Do_StartRectSelect(pt);

                    snapToY = false;
                    break;
                case MouseMode.Drag: Do_ProceedDrag(pt, (ModifierKeys & Keys.Shift) != 0, (ModifierKeys & Keys.Alt) != 0, snapToY); break;
                case MouseMode.Cleave: Do_ProceedCleave(pt, (ModifierKeys & Keys.Shift) != 0); break;
                case MouseMode.CreateSector: Do_ProceedCreateSC(pt, (ModifierKeys & Keys.Shift) != 0); break;
                case MouseMode.TranslateGrid: Do_TranslateGrid(pt); break;
                case MouseMode.TranslateCam: Do_TranslateCam(pt); break;
                case MouseMode.RotateCam: Do_RotateCam(pt); break;
                case MouseMode.RectSelect: Do_ProceedRectSelect(pt); break;
                    // need to implement allthe other crap
            }

            Vec3 g;
            if (renderer.GetGridAt(pt, out g))
            {
                PXYZ.Text = "X,Y,Z: " + g.x.ToString("#0.0000") + ", " + g.y.ToString("#0.0000") + ", " + g.z.ToString("#0.0000");
            }
        }

        public void Do_TranslateGrid(Vec2 pt)
        {
            Vec3 n;
            if (!renderer.GetXYZOnPlaneAt(pt, renderer.ZVec, renderer.CamPos, out n))
                return;

            double a = 1.0 / GlobalVars.GridMoveStep;
            n.x = Math.Floor(n.x * a) / a;
            n.y = Math.Floor(n.y * a) / a;
            n.z = Math.Floor(n.z * a) / a;

            renderer.GridPos = n;
        }

        public void Do_TranslateCam(Vec2 pt)
        {
            Vec3 n;
            if (!renderer.GetXYZOnPlaneAt(pt, renderer.ZVec, renderer.CamPos, out n))
                return;

            renderer.CamPos = n;

            Do_StartTranslateCam(pt);
        }

        Vec2 CamOrgXY = Vec2.Zero;

        public void Do_StartTranslateCam(Vec2 pt)
        {
            CamOrgXY = pt;
        }

        public void Do_ProceedCreateSC(Vec2 pt, bool snaptovertex)
        {
            if (!SCCreateStarted)
                return;

            Vec3 f;
            if (!renderer.GetGridAt(pt, out f))
                return;

            if (snaptovertex)
                Utils.GetNearestVX(Level, ref f, 0.2);
            else if (SnapToGrid)
                renderer.GetNearestGrid(ref f);

            DrawToXY = renderer.ProjectPoint(f);
        }

        public void Do_ProceedDrag(Vec2 pt, bool snaptovx, bool snaptoaxis, bool yaxis)
        {
            Vec3 f;
            if (!renderer.GetGridAt(pt, out f))
                return;

            Vec3 r = DragOrgPos;

            Vec3 d = GetDXYZ(r, f, snaptovx, snaptoaxis, yaxis);

            r = GetCurObjRefXYZ();

            d = DragOrgPos - r + d;

            switch (map_mode)
            {
                case MapMode.Sector: Utils.TranslateSectors(Level, scsel, Cur_SC, d); break;
                case MapMode.Surface: Utils.TranslateSurfaces(Level, sfsel, Cur_SC, Cur_SF, d); break;
                case MapMode.Vertex: Utils.TranslateVertices(Level, vxsel, Cur_SC, Cur_VX, d); break;
                case MapMode.Edge: Utils.TranslateEdges(Level, edsel, Cur_SC, Cur_SF, Cur_ED, d); break;
                case MapMode.Thing: Utils.TranslateThings(Level, thsel, Cur_TH, d, GlobalVars.MoveFrames); break;
                case MapMode.Frame:
                    if (Cur_FR == -1)
                        Utils.TranslateThingKeepFrame(Level, Cur_TH, d);
                    else
                        Utils.TranslateFrames(Level, frsel, Cur_TH, Cur_FR, d);
                    break;
                case MapMode.Light: Utils.TranslateLights(Level, ltsel, Cur_LT, d); break;
                case MapMode.Extra: TranslateExtras(Cur_EX, d); break;
            }
        }

        public bool LayerThing(int n)
        {
            JKThing thing = Level.things[n];
            int nsec = Utils.FindSectorForThing(thing);

            if (nsec == -1)
                thing.sec = null;
            else
                thing.sec = Level.sectors[nsec];
            ThingChanged(thing);
            return (nsec != -1);
        }

        /*int*/
        void TranslateExtras(int curex, Vec3 d)
        {
            object ex = ExtraObjs[curex];
            if (ex is Vertex)
                (ex as Vertex).v += d;
            if (ex is ExtraLine)
            {
                (ex as ExtraLine).v1 += d;
                (ex as ExtraLine).v2 += d;
            }
            if (ex is Polygon)
            {
                foreach (Vertex v in (ex as Polygon).vertices)
                    v.v += d;
            }

            if (OnExtraMove != null)
                OnExtraMove(ex, true);
        }

        public Vec3 GetDXYZ(Vec3 r, Vec3 f, bool snaptovx, bool snaptoaxis, bool yaxis)
        {
            Vec3 d = f - DragOrg;

            const double dragCloseEnough = 0.2;

            if (snaptoaxis)
            {
                Vec2 ar = new Vec2();
                ar.x = renderer.GridXVec.x * d.x + renderer.GridXVec.y * d.y + renderer.GridXVec.z * d.z;
                ar.y = renderer.GridYVec.x * d.x + renderer.GridYVec.y * d.y + renderer.GridYVec.z * d.z;

                if (yaxis)
                {
                    d.x = renderer.GridYVec.x * ar.y;
                    d.y = renderer.GridYVec.y * ar.y;
                    d.z = renderer.GridYVec.z * ar.y;
                }
                else
                {
                    d.x = renderer.GridXVec.x * ar.x;
                    d.y = renderer.GridXVec.y * ar.x;
                    d.z = renderer.GridXVec.z * ar.x;
                }

                if (SnapToGrid)
                {
                    double sd = renderer.GridNormal * renderer.GridPos;
                    Vec3 ar3;
                    Utils.PlaneLineXn(renderer.GridNormal, sd, r + d, r + d + renderer.GridNormal, out ar3);
                    Vec3 a = ar3;
                    renderer.GetNearestGrid(ref a);
                    d = a - ar3 + d;
                }

                return d;
            }

            if (!snaptovx)
            {
                if (!SnapToGrid)
                    return d;

                double sd = renderer.GridNormal * renderer.GridPos;
                Vec3 ar3;
                Utils.PlaneLineXn(renderer.GridNormal, sd, r + d, r + d + renderer.GridNormal, out ar3);
                Vec3 a = ar3;
                renderer.GetNearestGrid(ref a);
                d = a - ar3 + d;
            }
            else
            {
                Vec3 a = r + d;
                if (!Utils.GetNearestVX(Level, ref a, dragCloseEnough))
                    return d;

                d = a - r;
            }

            return d;
        }

        public void Do_ProceedRectSelect(Vec2 pt)
        {
            RectLastXY = pt;
        }

        public void Do_ProceedCleave(Vec2 pt, bool snaptovertex)
        {
            if (!CleaveStarted)
                return;

            Vec3 v;
            if (!renderer.GetGridAt(pt, out v))
                return;

            if (snaptovertex)
                Utils.GetNearestVX(Level, ref v, 0.2);
            else if (SnapToGrid)
                renderer.GetNearestGrid(ref v);

            DrawToXY = renderer.ProjectPoint(v);
            LastXY = DrawToXY;
        }

        bool snapToY = false;

        public Vec2? GetMousePos()
        {
            Point pt = renderControl1.PointToClient(Cursor.Position);
            if (!ClientRectangle.Contains(pt))
                return null;

            return new Vec2(pt);
        }

        public MouseMode mouse_mode = MouseMode.Select;

        private static Cursor _SaberCursor = null;
        public static Cursor SaberCursor
        {
            get
            {
                if (_SaberCursor == null)
                    _SaberCursor = new Cursor(new MemoryStream(Properties.Resources.saber));

                return _SaberCursor;
            }
        }

        private static Cursor _SingleDragCursor = null;
        public static Cursor SingleDragCursor
        {
            get
            {
                if (_SingleDragCursor == null)
                    _SingleDragCursor = new Cursor(new MemoryStream(Properties.Resources.dragsingle));

                return _SingleDragCursor;
            }
        }

        private static Cursor _MultiDragCursor = null;
        public static Cursor MultiDragCursor
        {
            get
            {
                if (_MultiDragCursor == null)
                    _MultiDragCursor = new Cursor(new MemoryStream(Properties.Resources.dragmulti));

                return _MultiDragCursor;
            }
        }

        public void SetMouseMode(MouseMode mm)
        {
            switch (mm)
            {
                case MouseMode.Select: Cursor = Cursors.Default; break;
                case MouseMode.Cleave: Cursor = SaberCursor; break;
                case MouseMode.Drag: Cursor = SingleDragCursor; break;
                case MouseMode.TranslateCam: Cursor = Cursors.SizeAll; break;
                case MouseMode.RotateCam: Cursor = Cursors.Default; break;
                case MouseMode.TranslateGrid: Cursor = Cursors.SizeAll; break;
                case MouseMode.RotateGrid: Cursor = Cursors.Default; break;
                case MouseMode.CreateSector: Cursor = Cursors.Default; break;
                case MouseMode.RectSelect: Cursor = Cursors.Default; break;
            }
            mouse_mode = mm;
        }

        Vec2 GOrgXY = Vec2.Zero;
        Orient GOrg = Orient.Zero;
        Vec3 OrgX = Vec3.PosX;
        Vec3 OrgY = Vec3.PosY;
        Vec3 OrgZ = Vec3.PosZ;

        public void Do_StartRotateCam(Vec2 pt)
        {
            if (mouse_mode == MouseMode.RotateCam)
                return;

            GOrgXY = pt;

            GOrg = camOrient;

            OrgX = renderer.XVec;
            OrgY = renderer.YVec;
            OrgZ = renderer.ZVec;

            SetMouseMode(MouseMode.RotateCam);
        }

        public void Do_RotateCam(Vec2 pt)
        {
            Vec2 d = pt - GOrgXY;

            if (GlobalVars.MapRot == GlobalVars.MapRotation.Old)
            {
                camOrient.roll = GOrg.roll + d.x;
                camOrient.pitch = GOrg.pitch + d.y;
                SetRendFromPYR();
            }
            else
            {
                Vec3 nx = OrgX;
                Vec3 ny = OrgY;
                Vec3 nz = OrgZ;

                RotAxisAngle(ref nx, ref ny, d.x);
                RotAxisAngle(ref ny, ref nz, d.y);

                renderer.SetZ(nz);
                renderer.SetX(nx);
            }
        }



        protected override void OnPreviewKeyDown(PreviewKeyDownEventArgs e)
        {
            DoPreviewKeyDown(e);
        }

        public void DoPreviewKeyDown(PreviewKeyDownEventArgs e)
        {
            // since we brought a couple tool dialogs into side panel, we removed the menu items.. this also means removing the shortcut key.
            // so we'll let a couple slip through to be implemented manually in DoKeyDown
            if((new List<Keys>(new Keys[] {
                Keys.Control, Keys.Alt, Keys.Shift,
                Keys.F1, Keys.F2, Keys.F3, /*Keys.F4,*/ Keys.F5, Keys.F6, /*Keys.F7,*/ /*Keys.F8,*/ /*Keys.F9,*/
                /*Keys.F10,*/ Keys.F11, Keys.F12
            })).Contains(e.KeyCode))
                return;

            if ((ModifierKeys & (Keys.Shift | Keys.Control | Keys.Alt)) != 0)
                return;

            e.IsInputKey = true;
        }

        protected override void OnKeyUp(KeyEventArgs e)
        {
            DoKeyUp(e);
        }

        public void DoKeyUp(KeyEventArgs e)
        {
            if (e.KeyCode == Keys.G ||
                e.KeyCode == Keys.R ||
                e.KeyCode == Keys.Space)
                SetMouseMode(MouseMode.Select);
        }

        public void NextObject()
        {
            switch (map_mode)
            {
                case MapMode.Sector:
                    {
                        int fi = Cur_SC;
                        int n = fi;
                        do
                        {
                            if (n >= (Level.sectors.Count-1))
                                n = 0;
                            else
                                n++;
                            if (n == fi)
                                return;
                        } while (!Toolbar.IsLayerVisible(Level.sectors[n].layer));

                        SetCurSC(n);
                    }
                    break;

                case MapMode.Surface:
                    SetCurSF(Cur_SC, Cur_SF + 1);
                    break;

                case MapMode.Vertex:
                    SetCurVX(Cur_SC, Cur_VX + 1);
                    break;

                case MapMode.Edge:
                    SetCurED(Cur_SC, Cur_SF, Cur_ED + 1);
                    break;

                case MapMode.Thing:
                    {
                        int fi = Cur_TH;
                        int n = fi;
                        do
                        {
                            if(n >= (Level.things.Count-1))
                                n = 0;
                            else
                                n++;
                            if (n == fi)
                                return;
                        } while (!Toolbar.IsLayerVisible(Level.things[n].layer));

                        SetCurTH(n);
                    }
                    break;

                case MapMode.Frame:
                    {
                        JKThing th = Level.things[Cur_TH];
                        Cur_FR = th.NextFrame(Cur_FR);
                        SetCurFR(Cur_TH, Cur_FR);
                    }
                    break;

                case MapMode.Light:
                    {
                        int fi = Cur_LT;
                        int n = fi;
                        do
                        {
                            if (n >= (Level.lights.Count-1))
                                n = 0;
                            else
                                n++;
                            if (n == fi)
                                return;
                        } while (!Toolbar.IsLayerVisible(Level.lights[n].layer));

                        SetCurLT(n);
                    }
                    break;

                case MapMode.Extra:
                    //SetCurLT(Cur_EX + 1); // jed actually has this
                    SetCurEX(Cur_EX + 1);// i assume its supposed to be this
                    break;
            }
        }

        public void PreviousObject()
        {
            switch (map_mode)
            {
                case MapMode.Sector:
                    {
                        int fi = Cur_SC;
                        int n = fi;
                        do
                        {
                            if (n < 1)
                                n = Level.sectors.Count - 1;
                            else
                                n--;
                            if (n == fi)
                                return;
                        } while (!Toolbar.IsLayerVisible(Level.sectors[n].layer));

                        SetCurSC(n);
                    }
                    break;

                case MapMode.Surface:
                    SetCurSF(Cur_SC, Cur_SF - 1);
                    break;

                case MapMode.Vertex:
                    SetCurVX(Cur_SC, Cur_VX - 1);
                    break;

                case MapMode.Edge:
                    SetCurED(Cur_SC, Cur_SF, Cur_ED - 1);
                    break;

                case MapMode.Thing:
                    {
                        int fi = Cur_TH;
                        int n = fi;
                        do
                        {
                            if (n < 1)
                                n = Level.things.Count - 1;
                            else
                                n--;
                            if (n == fi)
                                return;
                        } while (!Toolbar.IsLayerVisible(Level.things[n].layer));

                        SetCurTH(n);
                    }
                    break;

                case MapMode.Frame:
                    {
                        JKThing th = Level.things[Cur_TH];
                        if (Cur_FR == -1)
                            Cur_FR = th.vals.Count - 1;
                        else
                            Cur_FR = th.PrevFrame(Cur_FR);
                        SetCurFR(Cur_TH, Cur_FR);
                    }
                    break;

                case MapMode.Light:
                    {
                        int fi = Cur_LT;
                        int n = fi;
                        do
                        {
                            if (n < 1)
                                n = Level.lights.Count - 1;
                            else
                                n--;
                            if (n == fi)
                                return;
                        } while (!Toolbar.IsLayerVisible(Level.lights[n].layer));

                        SetCurLT(n);
                    }
                    break;

                case MapMode.Extra:
                    SetCurEX(Cur_EX - 1);
                    break;
            }
        }

        public void SnapGridToObject()
        {
            Vec3 norm = renderer.GridNormal;
            Vec3 xnorm = renderer.GridXVec;
            Vec3 r = renderer.GridPos;

            switch (map_mode)
            {
                case MapMode.Sector:
                    r = Level.sectors[Cur_SC].vertices[0].v;
                    break;

                case MapMode.Surface:
                    {
                        JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];

                        Vec3 v1 = surf.vertices[0].v;
                        Vec3 v2 = surf.vertices[1].v;
                        r = v1;
                        norm = surf.normal;
                        xnorm = v2 - v1;
                    }
                    break;

                case MapMode.Edge:
                    {
                        JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];

                        Vec3 v1 = surf.vertices[Cur_ED].v;
                        Vec3 v2 = surf.vertices[surf.NextVX(Cur_ED)].v;
                        r = v1;

                        norm = surf.normal;
                        xnorm = v2 - v1;
                    }
                    break;

                case MapMode.Vertex:
                    {
                        int i = vxsel.AddVX(Cur_SC, Cur_VX);
                        if (vxsel.Count == 3)
                        {
                            int sc, vx;
                            vxsel.GetSCVX(0, out sc, out vx);
                            Vec3 v1 = Level.sectors[sc].vertices[vx].v;
                            vxsel.GetSCVX(1, out sc, out vx);
                            Vec3 v2 = Level.sectors[sc].vertices[vx].v;
                            xnorm = v2 - v1;
                            vxsel.GetSCVX(2, out sc, out vx);
                            v1 = Level.sectors[sc].vertices[vx].v;
                            Vec3 av = v1 - v2;
                            norm = xnorm ^ av;
                        }

                        vxsel.DeleteN(i);
                        r = Level.sectors[Cur_SC].vertices[Cur_VX].v;
                    }
                    break;

                case MapMode.Thing:
                    {
                        r = Level.things[Cur_TH].pos;
                    }
                    break;

                case MapMode.Frame:
                    {
                        if (Cur_FR == -1)
                            r = Level.things[Cur_TH].pos;
                        else
                        {
                            Orient fo;
                            Level.things[Cur_TH].vals[Cur_FR].GetFrame(out r, out fo);
                        }
                    }
                    break;

                case MapMode.Light:
                    {
                        r = Level.lights[Cur_LT].Pos;
                    }
                    break;
                }

            renderer.GridPos = r;
            renderer.SetGridNormal(norm);
            renderer.SetGridXNormal(xnorm);
        }

        public void DoKeyDown(KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Menu || e.KeyCode == Keys.LMenu || e.KeyCode == Keys.RMenu || e.KeyCode == Keys.Alt)
                return;

            if (e.KeyCode == Keys.F4)
            {
                Toolbar.Show();
                renderControl1.Focus();
                return;
            }

            if (e.KeyCode == Keys.F10)
            {
                tabControl3.SelectedTab = tabConsistencyCheck;
                GlobalVars.ConsistencyChecker.Check();
                return;
            }

            if (e.KeyCode == Keys.F7)
            {
                tabControl3.SelectedTab = tabCogs;
                return;
            }

            if (e.KeyCode == Keys.F9)
            {
                GlobalVars.ToolForm.Show();
                renderControl1.Focus();
                return;
            }

            if (e.KeyCode == Keys.F8)
            {
                GlobalVars.EpisodeEdit.Show();
                return;
            }

            /*if (e.KeyCode == Keys.Enter)
            {
                itemEdit1.Show();
                return;
            }*/

            if (ModifierKeys == (Keys.Control | Keys.Alt))
            {
                if(e.KeyCode == Keys.Z)
                    snapToY = !snapToY;
            }

            if (ModifierKeys == Keys.Shift)
            {
                if (GlobalVars.MapRot == GlobalVars.MapRotation.Old)
                {
                    bool updatecam = true;
                    switch (e.KeyCode)
                    {
                        case Keys.D1: camOrient = new Orient(0.0, 0.0, 0.0); renderer.SetGridNormal(Vec3.PosZ); renderer.SetGridXNormal(Vec3.PosX); break;
                        case Keys.D2: camOrient = new Orient(-90.0, 180.0, 0.0); renderer.SetGridNormal(Vec3.PosY); renderer.SetGridXNormal(Vec3.NegX); break;
                        case Keys.D3: camOrient = new Orient(90.0, 0.0, 90.0); renderer.SetGridNormal(Vec3.PosX); renderer.SetGridXNormal(Vec3.PosY); break;
                        case Keys.D4: camOrient = new Orient(0.0, 180.0, 0.0); renderer.SetGridNormal(Vec3.NegZ); renderer.SetGridXNormal(Vec3.NegX); break;
                        case Keys.D5: camOrient = new Orient(90.0, 0.0, 0.0); renderer.SetGridNormal(Vec3.NegY); renderer.SetGridXNormal(Vec3.PosX); break;
                        case Keys.D6: camOrient = new Orient(0.0, -90.0, -90.0); renderer.SetGridNormal(Vec3.NegX); renderer.SetGridXNormal(Vec3.NegY); break;
                        case Keys.Right: camOrient.yaw -= 5.0; break;
                        case Keys.Left: camOrient.yaw += 5.0; break;
                        case Keys.Up: camOrient.pitch -= 5.0; break;
                        case Keys.Down: camOrient.pitch += 5.0; break;
                        case Keys.Prior: camOrient.roll -= 5.0; break;
                        case Keys.Next: camOrient.roll += 5.0; break;

                        default: updatecam = false; break;
                    }

                    if(updatecam)
                        SetRendFromPYR();

                }
                else
                {
                    bool updatevecs = false;

                    switch (e.KeyCode)
                    {
                        case Keys.D1: renderer.SetZ(Vec3.PosZ); renderer.SetX(Vec3.PosX); renderer.SetGridNormal(Vec3.PosZ); renderer.SetGridXNormal(Vec3.PosX); break;
                        case Keys.D2: renderer.SetZ(Vec3.PosY); renderer.SetX(Vec3.NegX); renderer.SetGridNormal(Vec3.PosY); renderer.SetGridXNormal(Vec3.NegX); break;
                        case Keys.D3: renderer.SetZ(Vec3.PosX); renderer.SetX(Vec3.PosY); renderer.SetGridNormal(Vec3.PosX); renderer.SetGridXNormal(Vec3.PosY); break;
                        case Keys.D4: renderer.SetZ(Vec3.NegZ); renderer.SetX(Vec3.NegX); renderer.SetGridNormal(Vec3.NegZ); renderer.SetGridXNormal(Vec3.NegX); break;
                        case Keys.D5: renderer.SetZ(Vec3.NegY); renderer.SetX(Vec3.PosX); renderer.SetGridNormal(Vec3.NegY); renderer.SetGridXNormal(Vec3.PosX); break;
                        case Keys.D6: renderer.SetZ(Vec3.NegX); renderer.SetX(Vec3.NegY); renderer.SetGridNormal(Vec3.NegX); renderer.SetGridXNormal(Vec3.NegY); break;
                        case Keys.Right: RotAxis(ref renderer.ZVec, ref renderer.XVec, -1); updatevecs = true; break;
                        case Keys.Left: RotAxis(ref renderer.ZVec, ref renderer.XVec, 1); updatevecs = true; break;
                        case Keys.Up: RotAxis(ref renderer.YVec, ref renderer.ZVec, -1); updatevecs = true; break;
                        case Keys.Down: RotAxis(ref renderer.YVec, ref renderer.ZVec, 1); updatevecs = true; break;
                        case Keys.Prior: RotAxis(ref renderer.XVec, ref renderer.YVec, -1); updatevecs = true; break;
                        case Keys.Next: RotAxis(ref renderer.XVec, ref renderer.YVec, 1); updatevecs = true; break;
                    }

                    if (updatevecs)
                    {
                        renderer.SetZ(renderer.ZVec);
                        renderer.SetX(renderer.XVec);
                        return;
                    }

                }



                switch (e.KeyCode)
                {
                    case Keys.F:
                        Goto_Adjoin();
                        break;

                    case Keys.X:
                        if (map_mode == MapMode.Surface)
                        {
                            JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];
                            if (surf.adjoin == null)
                            {
                                string s = surf.ExtrudeSize.ToString();
                                if (Utils.InputBox("Extrude", "By:", ref s) == DialogResult.OK)
                                {
                                    double d;
                                    if (double.TryParse(s, out d))
                                    {
                                        if (d <= 0.0)
                                            MessageBox.Show($"Invalid Value: {s}");
                                        else
                                        {
                                            Undo.StartUndoRec("Extrude surface");

                                            Utils.ExtrudeSurface(surf, d);
                                            SetCurSF(Cur_SC, Cur_SF);
                                        }
                                    }
                                }
                            }
                        }
                        break;

                    case Keys.T:
                        SetMapMode(MapMode.Frame);
                        break;

                    case Keys.C:
                        {
                            Vec2? mpt = GetMousePos();
                            if (mpt.HasValue)
                            {
                                //renderer.ZVec

                                Vec3 v;
                                if (renderer.GetXYZOnPlaneAt(mpt.Value, renderer.ZVec, renderer.CamPos, out v))
                                {
                                    renderer.CamPos = -v;
                                    renderer.scale /= 1.5;
                                }
                            }
                        }
                        break;

                    case Keys.A:
                        if (map_mode == MapMode.Thing)
                            LayerThings();
                        break;

                    case Keys.G:
                        {
                            Vec3 a = renderer.GridYVec;
                            Vec3 b = renderer.GridXVec;
                            renderer.SetGridNormal(b);
                            renderer.SetGridXNormal(a);
                        }
                        break;

                    case Keys.S:
                    case Keys.W:
                        SnapGridToObject();
                        break;

                    case Keys.Oemcomma:
                        ShiftTexture(ShiftTextureType.Up);
                        break;

                    case Keys.OemPeriod:
                        ShiftTexture(ShiftTextureType.Down);
                        break;

                    case Keys.OemQuestion:
                        StraightenTexture(false, true);
                        break;

                    case Keys.Insert:
                        DoStitch();
                        break;
                }

                /*if Shift=[ssShift] then
 begin

  Ord('N'): NextObjectInSurface;
  Ord('P'): PreviousObjectInSurface;

 end;
  Key:=0;
  Invalidate;
  exit;
 end;*/
                return;
            }

            if (ModifierKeys == Keys.Alt)
            {
                Vec3 n = renderer.GridNormal;
                Vec3 xn = renderer.GridXVec;

                switch (e.KeyCode)
                {
                    case Keys.A:
                        switch (map_mode)
                        {
                            case MapMode.Sector:
                                {
                                    int i = scsel.AddSC(Cur_SC);
                                    if (scsel.Count == 2)
                                    {
                                        Undo.StartUndoRec("Unadjoin sectors");
                                        if (Utils.UnadjoinSectors(Level.sectors[scsel.GetSC(0)], Level.sectors[scsel.GetSC(1)]))
                                            Utils.PanMessage(Utils.MessageType.Info, "Sectors unadjoined");
                                    }
                                    else
                                        MessageBox.Show("You must have 2 sectors selected");
                                    scsel.DeleteN(i);
                                }
                                break;

                            case MapMode.Surface:
                                {
                                    int x = sfsel.AddSF(Cur_SC, Cur_SF);
                                    int nsel = 0;
                                    Undo.StartUndoRec("Unadjoin surfaces");
                                    for (int i = 0; i < sfsel.Count; i++)
                                    {
                                        int nsc, nsf;
                                        sfsel.GetSCSF(i, out nsc, out nsf);
                                        if (Utils.Unadjoin(Level.sectors[nsc].surfaces[nsf]))
                                            nsel++;
                                    }
                                    Utils.PanMessage(Utils.MessageType.Info, nsel.ToString() + " adjoins removed");
                                    SetCurSF(Cur_SC, Cur_SF);
                                    sfsel.DeleteN(x);
                                }
                                break;
                        }
                        break;

                    case Keys.C://   Ord('C'): With Renderer do CleaveBy(gnormal,GridX,GridY,GridZ); {Cleave by grid plane}
                        {
                            CleaveBy(renderer.GridNormal, renderer.GridPos);
                        }
                        break;


                    case Keys.I:

                        if (map_mode == MapMode.Surface)
                        {
                            int i = sfsel.AddSF(Cur_SC, Cur_SF);

                            Utils.FlipSurfaces(Level, sfsel);

                            sfsel.DeleteN(i);
                        }

                        break;
                }


                    /*
     if shift=[ssAlt] then
     begin
      n:=Renderer.gnormal;
      xn:=Renderer.gxnormal;

      Ord('B'): case map_mode of
                 MM_SF:begin
                        i:=sfsel.AddSF(cur_SC,cur_SF);
                        sec:=BuildSector(level,sfsel);
                        if sec=nil then sfsel.DeleteN(i) else
                        begin
                         sfsel.clear;
                         SetMapMode(MM_SC);
                         SetCurSC(sec.num);
                       end;
                      end;
                 MM_VX:begin
                        i:=vxsel.AddVX(cur_SC,cur_VX);
                        surf:=BuildSurface(level,vxsel);
                        if surf=nil then vxsel.DeleteN(i) else
                        begin
                         vxsel.clear;
                         SetMapMode(MM_SF);
                         SetCurSF(surf.sector.num,surf.num);
                       end;
                      end;
                end;


      VK_HOME: begin StraightenTexture(false,false); exit; end;
      VK_DELETE: if Map_mode=MM_SF then DeleteSurface(Level,cur_SC,cur_SF);
      188: begin ScaleTexture(st_down); exit; end; {<}
      190: begin RotateTexture(st_up); exit; end; {>}

     end;

      Renderer.SetGridNormal(n.dx,n.dy,n.dz);
      Renderer.SetGridXNormal(xn.dx,xn.dy,xn.dz);
      Invalidate;
      exit;
     end;*/

                        return;
            }

            if (ModifierKeys == (Keys.Shift | Keys.Control))
            {
                /*if shift=[ssShift,ssCtrl] then
 begin
 case key of
  191,VK_HOME: StraightenTexture(true,true);
 end;
  exit;
 end;*/

                return;
            }

            if (ModifierKeys == Keys.Control)
            {
                /*switch (e.KeyCode)
                {
                    case Keys.W:
                        // CTRL+W  map material to wall

                        break;
                }*/
                /*if shift=[ssCtrl] then
 begin
 case key of
  VK_INSERT: begin StartStitch; exit; end;
  191,VK_HOME: begin StraightenTexture(true,false); exit; end;
  188: begin RotateTexture(st_left);  exit; end; {<}
  190: begin RotateTexture(st_right); exit; end; {>}
  VK_UP: RaiseObject(ro_up);
  VK_DOWN: RaiseObject(ro_down);
  Ord('C'): miCopyClick(nil);
  Ord('V'): miPasteClick(nil);
  Ord('P'): if map_mode=MM_SF then
           begin
            StartUndoRec('Flatten surface');
             if FlattenSurface(Level.Sectors[Cur_SC].Surfaces[Cur_SF]) then
             PanMessage(mt_info,'Surface planarized')
             else PanMessage(mt_info,'Surface wasn''t fully planarized');
             Invalidate;
            end;
  Ord('X'): if map_mode=MM_SF then
            begin
             surf:=Level.Sectors[cur_sc].Surfaces[cur_sf];
             if surf.adjoin<>nil then exit;
             s:=DoubleToStr(surf.ExtrudeSize);
             if not InputQuery('Extrude&&Expand','By:',s) then exit;
             if (Not ValDouble(s,dx)) or (dx=0) then
             begin ShowMessage('Invalid Value: '+s); exit; end;

             StartUndoRec('Extrude&&Expand surface');

             ExtrudeNExpandSurface(surf,dx);
             SetCurSF(Cur_SC,Cur_SF);
            end;
 end;
 exit;
 end;*/
                return;
            }

            switch (e.KeyCode)
            {
                case Keys.Escape:
                    CancelMouseMode();
                    break;

                case Keys.Back:
                    ClearMultiSelection();
                    break;

                case Keys.Add:
                    renderer.scale /= 1.5;
                    break;

                case Keys.Subtract:
                    renderer.scale *= 1.5;
                    break;
            }

            if (GlobalVars.MapRot == GlobalVars.MapRotation.Old)
            {
                bool updatecam = true;

                switch (e.KeyCode)
                {
                    case Keys.D1: camOrient = new Orient(0.0, 0.0, 0.0); break;
                    case Keys.D2: camOrient = new Orient(-90.0, 180.0, 0.0); break;
                    case Keys.D3: camOrient = new Orient(90.0, 0.0, 90.0); break;
                    case Keys.D4: camOrient = new Orient(0.0, 180.0, 0.0); break;
                    case Keys.D5: camOrient = new Orient(90.0, 0.0, 0.0); break;
                    case Keys.D6: camOrient = new Orient(0.0, -90.0, -90.0); break;
                    default: updatecam = false; break;
                }

                if (updatecam)
                    SetRendFromPYR();
            }
            else
            {
                switch (e.KeyCode)
                {
                    case Keys.D1: renderer.SetZ(Vec3.PosZ); renderer.SetX(Vec3.PosX); break;
                    case Keys.D2: renderer.SetZ(Vec3.PosY); renderer.SetX(Vec3.NegX); break;
                    case Keys.D3: renderer.SetZ(Vec3.PosX); renderer.SetX(Vec3.PosY); break;
                    case Keys.D4: renderer.SetZ(Vec3.NegZ); renderer.SetX(Vec3.NegX); break;
                    case Keys.D5: renderer.SetZ(Vec3.NegY); renderer.SetX(Vec3.PosX); break;
                    case Keys.D6: renderer.SetZ(Vec3.NegX); renderer.SetX(Vec3.NegY); break;
                }
            }

            if (e.KeyCode == Keys.Up || e.KeyCode == Keys.Down || e.KeyCode == Keys.Left || e.KeyCode == Keys.Right || e.KeyCode == Keys.Prior || e.KeyCode == Keys.Next)
            {
                Vec3 nx = renderer.XVec * renderer.scale;
                Vec3 ny = renderer.YVec * renderer.scale;
                Vec3 nz = renderer.ZVec * renderer.scale;

                switch (e.KeyCode)
                {
                    case Keys.Left: renderer.CamPos += nx; break;
                    case Keys.Right: renderer.CamPos -= nx; break;
                    case Keys.Down: renderer.CamPos += ny; break;
                    case Keys.Up: renderer.CamPos -= ny; break;
                    case Keys.Next: renderer.CamPos += nz; break;
                    case Keys.Prior: renderer.CamPos -= nz; break;
                }
            }


            switch (e.KeyCode)
            {
                case Keys.R:
                    {
                        Vec2? mpt = GetMousePos();
                        if (mpt.HasValue)
                            Do_StartRotateCam(mpt.Value);
                    }
                    break;

                case Keys.C:
                    SetMouseMode(MouseMode.Cleave);
                    break;

                case Keys.Y:
                    if (map_mode == MapMode.Surface)
                    {
                        JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];
                        string st = "ID: " + surf.uniqueID.ToString() + "\n";
                        foreach (JKSurface.GeoChange c in surf.geoChanges)
                            st += c.who + ": " + c.geo.ToString() + "\n";
                        MessageBox.Show(st);
                    }
                    break;

                case Keys.S:
                    SetMapMode(MapMode.Sector);
                    break;

                case Keys.F:
                    SetMapMode(MapMode.Surface);
                    break;

                case Keys.Space:
                    Do_MultiSelect();
                    break;

                case Keys.K:
                    SetMouseMode(MouseMode.CreateSector);
                    break;

                case Keys.T:
                    SetMapMode(MapMode.Thing);
                    break;

                case Keys.V:
                    SetMapMode(MapMode.Vertex);
                    break;

                case Keys.L:
                    SetMapMode(MapMode.Light);
                    break;

                case Keys.N:
                    NextObject();
                    break;

                case Keys.P:
                    PreviousObject();
                    break;

                case Keys.A:
                    switch (map_mode)
                    {
                        case MapMode.Sector:
                            {
                                int i = scsel.AddSC(Cur_SC);
                                if (scsel.Count == 2)
                                {
                                    Undo.StartUndoRec("Adjoin Sectors");
                                    if (Utils.AdjoinSectors(Level.sectors[scsel.GetSC(0)], Level.sectors[scsel.GetSC(1)]))
                                        Utils.PanMessage(Utils.MessageType.Info, "Sectors adjoined");
                                    else
                                        MessageBox.Show("You must have 2 sectors selected");
                                    scsel.DeleteN(i);
                                }
                            }
                            break;

                        case MapMode.Surface:
                            {
                                int x = sfsel.AddSF(Cur_SC, Cur_SF);
                                int nsel = 0;
                                Undo.StartUndoRec("Adjoin surfaces");
                                for (int i = 0; i < sfsel.Count; i++)
                                {
                                    int nsc, nsf;
                                    sfsel.GetSCSF(i, out nsc, out nsf);
                                    if (Utils.MakeAdjoin(Level.sectors[nsc].surfaces[nsf]))
                                        nsel++;
                                }
                                Utils.PanMessage(Utils.MessageType.Info, nsel.ToString() + " adjoins formed");
                                SetCurSF(Cur_SC, Cur_SF);
                                sfsel.DeleteN(x);
                            }
                            break;

                        case MapMode.Thing:
                            LayerThings();
                            break;
                    }
                    break;

                case Keys.Insert:
                    {
                        Vec2? xy = GetMousePos();
                        if(xy.HasValue)
                            switch (map_mode)
                            {
                                case MapMode.Sector:
                                    CopySectorsAt(xy.Value);
                                    break;

                                case MapMode.Light:
                                    AddLightsAt(xy.Value);
                                    break;

                                case MapMode.Thing:
                                    AddThingsAt(xy.Value);
                                    break;

                                case MapMode.Frame:
                                    AddFramesAt(xy.Value);
                                    break;
                            }
                    }
                    break;

                case Keys.Delete:
                    switch (map_mode)
                    {
                        case MapMode.Sector:
                            scsel.AddSC(Cur_SC);
                            scsel.Sort();
                            if (scsel.Count >= Level.sectors.Count)
                                return;

                            Undo.StartUndoRec("Delete Sector(s)");

                            for (int i = scsel.Count - 1; i >= 0; i--)
                                Utils.DeleteSector(Level, scsel.GetSC(i));
                            scsel.Clear();
                            break;

                        case MapMode.Thing:
                            thsel.AddTH(Cur_TH);
                            thsel.Sort();
                            if (thsel.Count >= Level.things.Count)
                                return;

                            Undo.StartUndoRec("Delete Thing(s)");

                            for (int i = thsel.Count - 1; i >= 0; i--)
                                Utils.DeleteThing(Level, thsel.GetTH(i));
                            thsel.Clear();
                            break;

                        case MapMode.Frame:
                            frsel.AddFR(Cur_TH, Cur_FR);
                            frsel.Sort();

                            Undo.StartUndoRec("Delete Frame(s)");

                            for (int i = frsel.Count - 1; i >= 0; i--)
                            {
                                int nth, nfr;
                                frsel.GetTHFR(i, out nth, out nfr);
                                DeleteFrame(nth, nfr);
                            }
                            frsel.Clear();
                            break;

                        case MapMode.Light:
                            ltsel.AddLT(Cur_LT);
                            ltsel.Sort();
                            if (ltsel.Count >= Level.lights.Count)
                                return;

                            Undo.StartUndoRec("Delete Light(s)");

                            for (int i = ltsel.Count - 1; i >= 0; i--)
                                Utils.DeleteLight(Level, ltsel.GetLT(i));
                            ltsel.Clear();
                            break;

                        case MapMode.Vertex:
                            {
                                vxsel.AddVX(Cur_SC, Cur_VX);
                                vxsel.Sort();
                                if (vxsel.Count == 1)
                                    Undo.StartUndoRec("Delete Vertex");
                                else
                                    Undo.StartUndoRec("Delete Vertices");

                                int a = 0;
                                for (int i = vxsel.Count - 1; i >= 0; i--)
                                {
                                    int x, y;
                                    vxsel.GetSCVX(i, out x, out y);
                                    if(Utils.DeleteVertex(Level.sectors[x], y, vxsel.Count!=1))
                                        a++;
                                }
                                Utils.PanMessage(Utils.MessageType.Info, a.ToString() + " vertices deleted");
                                vxsel.Clear();
                            }
                            break;
                    }
                    SetMapMode(map_mode);
                    break;

                case Keys.E:
                    SetMapMode(MapMode.Edge);
                    break;

                case Keys.X:
                    if (map_mode == MapMode.Surface)
                    {
                        JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];

                        Undo.StartUndoRec("Extrude surface");

                        Utils.ExtrudeSurface(surf, surf.ExtrudeSize);

                        SetCurSF(Cur_SC, Cur_SF);
                    }
                    break;

                case Keys.J:
                    switch (map_mode)
                    {
                        case MapMode.Sector:
                            ConnectSCs();
                            break;

                        case MapMode.Surface:
                            ConnectSFs();
                            break;
                    }
                    break;

                case Keys.M:
                    switch (map_mode)
                    {
                        case MapMode.Sector:
                            {
                                /*int i =*/ scsel.AddSC(Cur_SC);
                                scsel.Sort();
                                int nsel = 0;
                                Undo.StartUndoRec("Merge sectors");

                                while (scsel.Count > 1)
                                {
                                    JKSector sec = Level.sectors[scsel.GetSC(scsel.Count - 1)];
                                    for (int i = 0; i < scsel.Count - 1; i++)
                                        if (Utils.MergeSectors(Level.sectors[scsel.GetSC(i)], sec) != null)
                                        {
                                            nsel++;
                                            break;
                                        }

                                    scsel.DeleteN(scsel.Count - 1);
                                }

                                SetCurSC(scsel.GetSC(0));
                                scsel.Clear();
                                Utils.PanMessage(Utils.MessageType.Info, nsel.ToString() + " sectors merged");
                            }
                            break;

                        case MapMode.Surface:
                            {
                                int i = sfsel.AddSF(Cur_SC, Cur_SF);

                                if (sfsel.Count == 1)
                                {
                                    JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];
                                    if (surf.adjoin == null)
                                        return;

                                    Undo.StartUndoRec("Merge sectors");
                                    JKSector sec = Utils.MergeSectors(surf.sector, surf.adjoin.sector);
                                    if (sec == null)
                                        Utils.PanMessage(Utils.MessageType.Info, "Sectors weren't merged");
                                    else
                                    {
                                        SetCurSF(sec.num, 0);
                                        sfsel.Clear();
                                    }
                                }
                                else
                                {
                                    sfsel.Sort();
                                    int nsel = 0;

                                    Undo.StartUndoRec("Merge surfaces");

                                    while (sfsel.Count > 1)
                                    {
                                        int nsc, nsf;
                                        sfsel.GetSCSF(sfsel.Count - 1, out nsc, out nsf);
                                        JKSurface surf = Level.sectors[nsc].surfaces[nsf];
                                        for (int a = 0; a < sfsel.Count - 1; a++)
                                        {
                                            sfsel.GetSCSF(a, out nsc, out nsf);
                                            if (Utils.MergeSurfaces(Level.sectors[nsc].surfaces[nsf], surf) != null)
                                            {
                                                nsel++;
                                                break;
                                            }
                                        }
                                        sfsel.DeleteN(sfsel.Count - 1);
                                    }

                                    {
                                        int nsc, nsf;
                                        sfsel.GetSCSF(0, out nsc, out nsf);
                                        JKSurface surf = Level.sectors[nsc].surfaces[nsf];
                                        sfsel.Clear();
                                        SetCurSF(surf.sector.num, surf.num);
                                        Utils.PanMessage(Utils.MessageType.Info, nsel.ToString() + " surfaces merged");
                                    }
                                }

                                sfsel.DeleteN(i);
                            }
                            break;

                        case MapMode.Vertex:
                            Utils.PanMessage(Utils.MessageType.Info, Utils.MergeVertices(Level, vxsel).ToString() + " vertices merged");
                            vxsel.Clear();
                            break;

                        case MapMode.Edge:
                            {
                                JKSurface surf = Level.sectors[Cur_SC].surfaces[Cur_SF];

                                Undo.StartUndoRec("Merge surfaces");

                                if (Utils.MergeSurfacesAt(surf, Cur_ED))
                                    edsel.Clear();
                            }
                            break;
                    }
                    SetMapMode(map_mode);
                    break;
            }

            /*procedure TJedMain.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var  th:TJKThing;
     i,a,nth,nfr,x,y:integer;
     n,xn:TVector;
     dx,dy,dz:double;
     nx,ny,nz:TVector;
     surf,surf1:TJKSurface;
     sec,sec1:TJKSector;
     mx:TMat3x3;
     nsc,nsf:integer;
     nsel:integer;
     s:string;
begin

 case key of  
  VK_RETURN: begin if IsSelvalid then ItemEdit.Show; exit; end;
 
  Ord('B'): case map_mode of
             MM_TH: BringThingToSurf(cur_TH,cur_SC,Cur_SF);
             MM_LT: BringLightToSurf(cur_LT,cur_SC,Cur_SF);
            end;
  VK_MULTIPLY: SetViewToGrid;

  Ord('G'): if GetMousePos(X,Y) then begin SetMouseMode(MM_TranslateGrid); Do_StartTranslateGrid(X,Y); end;
  
  Ord('U'): UseInCog;
  Ord('Q'): LoadDLLPlugin('h:\rr\mv\mv.dll');
 
  {;}  186: begin StartStitch; exit; end;
  {'}  222: begin DoStitch;  exit; end;
  {/}  191: begin StraightenTexture(false,false);  exit; end;
  {<}  188: begin ShiftTexture(st_left);  exit; end;
  {>}  190: begin ShiftTexture(st_right);  exit; end;
  {[}  219: RaiseObject(ro_up);
  {]}  221: RaiseObject(ro_down);

  else exit;
 end;
 Key:=0;
 Invalidate;
end;         */
        }

        public void GotoSC(int sc)
        {
            SetCurSC(sc);
            SetMapMode(MapMode.Sector);
            CenterViewOnObject(false);
        }

        public void GotoSF(int sc, int sf)
        {
            SetCurSF(sc, sf);
            SetMapMode(MapMode.Surface);
            CenterViewOnObject(false);
        }

        public void GotoTH(int th)
        {
            SetCurTH(th);
            SetMapMode(MapMode.Thing);
            CenterViewOnObject(false);
        }

        public void Goto_Adjoin()
        {
            JKSurface sf = Level.sectors[Cur_SC].surfaces[Cur_SF];
            if (sf.adjoin == null)
                return;

            SetCurSF(sf.adjoin.sector.num, sf.adjoin.num);
        }

        public void GotoXYZ(Vec3 c, bool force)
        {
            if (!force)
            {
                renderer.BeginRectPick(new Vec2(0.0, 0.0), new Vec2((double)renderer.vpw, (double)renderer.vph));
                force = !renderer.IsVertexInRect(c);
            }

            if (!force)
                return;

            renderer.CamPos = -c;
        }

        public void CenterViewOnObject(bool force)
        {
            switch (map_mode)
            {
                case MapMode.Sector:
                    GotoXYZ(Level.sectors[Cur_SC].vertices[0].v, force);
                    return;

                case MapMode.Surface:
                    GotoXYZ(Level.sectors[Cur_SC].surfaces[Cur_SF].vertices[0].v, force);
                    break;

                case MapMode.Edge:
                    GotoXYZ(Level.sectors[Cur_SC].surfaces[Cur_SF].vertices[Cur_ED].v, force);
                    break;

                case MapMode.Vertex:
                    GotoXYZ(Level.sectors[Cur_SC].vertices[Cur_VX].v, force);
                    break;

                case MapMode.Thing:
                    GotoXYZ(Level.things[Cur_TH].pos, force);
                    break;

                case MapMode.Frame:
                    if (Cur_FR != -1)
                    {
                        Vec3 fp;
                        Orient fo;
                        Level.things[Cur_TH].vals[Cur_FR].GetFrame(out fp, out fo);
                        GotoXYZ(fp, force);
                    }
                    else
                        GotoXYZ(Level.things[Cur_TH].pos, force);
                    break;

                case MapMode.Light:
                    GotoXYZ(Level.lights[Cur_LT].Pos, force);
                    break;
            }
        }

        public object GetCurObjForCog(COGType ct)
        {
            switch (map_mode)
            {
                case MapMode.Sector:
                    if (ct == COGType.Sector)
                        return Level.sectors[Cur_SC];

                    break;

                case MapMode.Surface:
                    if (ct == COGType.Surface)
                        return Level.sectors[Cur_SC].surfaces[Cur_SF];
                    else
                        if (ct == COGType.Sector)
                            return Level.sectors[Cur_SC];

                    break;

                case MapMode.Thing:
                    if (ct == COGType.Thing)
                        return Level.things[Cur_TH];

                    break;
            }

            MessageBox.Show("Ned map is not in the appropriate mode");
            return null;
        }

        public void ConnectSFs()
        {
            if (map_mode != MapMode.Surface)
                return;

            int a = sfsel.AddSF(Cur_SC, Cur_SF);
            if (sfsel.Count == 2)
            {
                Undo.StartUndoRec("Connect surfaces");

                int sc1, sf1;
                sfsel.GetSCSF(0, out sc1, out sf1);

                int sc2, sf2;
                sfsel.GetSCSF(1, out sc2, out sf2);

                if (Utils.ConnectSurfaces(Level.sectors[sc1].surfaces[sf1], Level.sectors[sc2].surfaces[sf2]))
                    Utils.PanMessage(Utils.MessageType.Info, "The surfaces successfully connected");
                UpdateItemEditor();
            }
            sfsel.DeleteN(a);
        }

        public void ConnectSCs()
        {
            if (map_mode != MapMode.Sector)
                return;

            int a = scsel.AddSC(Cur_SC);
            if (scsel.Count == 2)
            {
                Undo.StartUndoRec("Connect sectors");
                if (Utils.ConnectSectors(Level.sectors[scsel.GetSC(0)], Level.sectors[scsel.GetSC(1)]))
                    Utils.PanMessage(Utils.MessageType.Info, "The sectors successfully connected");
                UpdateItemEditor();
            }
            scsel.DeleteN(a);
        }

        public bool GetXYZAt(Vec2 pt, out Vec3 d)
        {
            if (!renderer.GetGridAt(pt, out d))
                return false;

            if (SnapToGrid)
                renderer.GetNearestGrid(ref d);

            return true;
        }

        public void CopySectorsAt(Vec2 pt)
        {
            Vec3 d;
            if (!GetXYZAt(pt, out d))
                return;

            JKSector sec = Level.sectors[Cur_SC];

            d -= sec.vertices[0].v;

            scsel.AddSC(Cur_SC);

            int fnew = Level.sectors.Count;

            Undo.StartUndoRec("Copy sector(s)");

            for (int i = 0; i < scsel.Count; i++)
            {
                sec = Level.sectors[scsel.GetSC(i)];
                JKSector newsec = Level.NewSector();

                Utils.DuplicateSector(sec, newsec, sec.vertices[0].v, renderer.GridXVec, renderer.GridYVec, renderer.GridNormal, d);
                Level.sectors.Add(newsec);
                Undo.SaveSecUndo(newsec, Undo.Change.Added, Undo.SectorChange.Both);
            }

            Level.RenumSecs();
            scsel.Clear();

            for (int i = fnew; i < Level.sectors.Count; i++)
            {
                sec = Level.sectors[i];
                foreach (JKSurface surf in sec.surfaces)
                    if (surf.nadj == 1)
                        Utils.MakeAdjoinSCUP(surf, fnew);

                SectorAdded(sec);
                scsel.AddSC(i);
            }

            SetCurSC(fnew);
            if (scsel.Count == 1)
                scsel.Clear();
        }

        public void AddLightsAt(Vec2 pt)
        {
            Vec3 l;
            if (!GetXYZAt(pt, out l))
                return;

            Undo.StartUndoRec("Add light(s)");

            if (Level.lights.Count == 0)
            {
                JedLight lt = Level.NewLight();
                lt.Pos = l;
                Level.lights.Add(lt);
                Undo.SaveLightUndo(lt, Undo.Change.Added);
                LevelChanged(true);
                return;
            }

            l -= Level.lights[Cur_LT].Pos;

            ltsel.AddLT(Cur_LT);

            int nlts = Level.lights.Count;
            for (int i = 0; i < ltsel.Count; i++)
            {
                JedLight lt = Level.lights[ltsel.GetLT(i)];
                JedLight nlt = Level.NewLight();
                nlt.Assign(lt);
                nlt.Pos = lt.Pos + l;
                Level.lights.Add(nlt);
                Undo.SaveLightUndo(nlt, Undo.Change.Added);
            }

            ltsel.Clear();
            for (int i = nlts; i < Level.lights.Count; i++)
                ltsel.AddLT(i);

            SetCurLT(nlts);
            if (ltsel.Count == 1)
                ltsel.Clear();

            LevelChanged(true);
        }

        public void AddThingsAt(Vec2 pt)
        {
            Vec3 t;
            if (!GetXYZAt(pt, out t))
                return;

            Undo.StartUndoRec("Add thing(s)");

            if (Level.things.Count == 0)
            {
                JKThing th = Level.NewThing();
                Level.things.Add(th);
                Undo.SaveThingUndo(th, Undo.Change.Added);
                ThingAdded(th);
                LayerThing(0);
                Undo.SaveThingUndo(th, Undo.Change.Added);// is this really supposed to be here twice?
                return;
            }

            t -= Level.things[Cur_TH].pos;
            thsel.AddTH(Cur_TH);

            if (GlobalVars.NewOnFloor && thsel.Count == 1)
                t -= Level.things[Cur_TH].box.LRF;// is this right

            int nths = Level.things.Count;
            for (int i = 0; i < thsel.Count; i++)
            {
                JKThing oldth = Level.things[thsel.GetTH(i)];
                JKThing th = Level.NewThing();
                th.Assign(oldth);
                th.pos = t + oldth.pos;

                foreach (TemplateValue tv in th.vals)
                {
                    if (tv.adjType != AdjType.Frame)
                        continue;

                    Vec3 fp;
                    Orient fo;
                    tv.GetFrame(out fp, out fo);
                    tv.SetFrame(fp + t, fo);
                }

                th.num = Level.things.Count;
                Level.things.Add(th);
                LayerThing(th.num);
                ThingAdded(th);
                Undo.SaveThingUndo(th, Undo.Change.Added);
            }

            thsel.Clear();

            for (int i = nths; i < Level.things.Count; i++)
                thsel.AddTH(i);
            SetCurTH(nths);
            if (thsel.Count == 1)
                thsel.Clear();
        }

        public void AddFramesAt(Vec2 pt)
        {
            Vec3 f;
            if (!GetXYZAt(pt, out f))
                return;

            JKThing th = Level.things[Cur_TH];

            Undo.StartUndoRec("Add frame");
            Undo.SaveThingUndo(th, Undo.Change.Changed);

            Orient o = th.orient;
            string sval = "(" + f.x.ToString("#0.000000") + "/" + f.y.ToString("#0.000000") + "/" + f.z.ToString("#0.000000") + ":" + o.pitch.ToString("#0.000000") + "/" + o.yaw.ToString("#0.000000") + "/" + o.roll.ToString("#0.000000") + ")";

            TemplateValue tv;
            if (Cur_FR == -1)
                tv = th.AddValue("frame", sval);
            else
                tv = th.InsertValue(Cur_FR + 1, "frame", sval);

            int nv = th.nframes;
            int i = th.vals.IndexOfName("numframes");
            if (i == -1)
                th.InsertValue(0, "numframes", nv.ToString());
            else
                th.vals[i].Val(nv.ToString());

            Cur_FR = th.vals.IndexOf(tv);

            SetCurFR(Cur_TH, Cur_FR);
        }

        public void DeleteFrame(int th, int fr)
        {
            if (fr == -1)
                return;

            JKThing thing = Level.things[th];
            if (fr >= thing.vals.Count)
                return;

            TemplateValue vl = thing.vals[fr];
            if (!vl.name.Equals("frame", StringComparison.InvariantCultureIgnoreCase))
                return;

            Undo.SaveThingUndo(thing, Undo.Change.Changed);

            thing.vals.Remove(vl);

            TemplateValue av;
            if (fr >= thing.vals.Count)
                av = thing.vals[thing.vals.Count - 1];
            else
                av = thing.vals[fr];

            int nv = thing.nframes;
            int i = thing.vals.IndexOfName("numframes");

            if (nv == 0)
            {
                if (i != -1)
                    thing.vals.RemoveAt(i);
                return;
            }

            if (i == -1)
                thing.InsertValue(0, "numframes", nv.ToString());
            else
                thing.vals[i].Val(nv.ToString());

            nv = thing.vals.IndexOf(av);

            fr = -1;

            i = thing.NextFrame(-1);

            while (i != -1)
            {
                fr = i + 1;
                i = thing.NextFrame(i);
            }

            SetCurFR(th, fr);
        }

        public void LayerThings()
        {
            if (map_mode != MapMode.Thing)
                return;

            int n = 0;
            int a = thsel.AddTH(Cur_TH);
            for(int i=0; i<thsel.Count; i++)
                if(LayerThing(thsel.GetTH(i)))
                    n++;
            thsel.DeleteN(a);
            Utils.PanMessage(Utils.MessageType.Info, n.ToString() + " thing(s) layered");
        }

        private void WireframesClick(object sender, EventArgs e)
        {
            if (sender == Dots)
                SetThingView(ThingView.Dots);
            else if (sender == Boxes)
                SetThingView(ThingView.Boxes);
            else if (sender == Wireframes)
                SetThingView(ThingView.Wireframes);
        }

        public void SetCam(Vec3 p, Orient o)
        {
            renderer.CamPos = -p;

            camOrient = o;
            SetRendFromPYR();
        }

        public static Orient sysGetPYR(Vec3 x, Vec3 y, Vec3 z)
        {
            Orient o = Orient.Zero;

            double l = Math.Sqrt(x.x * x.x + x.z * x.z);
            if (MathLib.CompareScalar(l, 0.0) != 0)
                o.yaw = -MathLib.ToDeg(Math.Asin(x.z / l));

            if (MathLib.CompareScalar(x.x, 0.0) < 0)
                o.yaw = 180.0 - o.yaw;
            if (o.yaw < 0.0)
                o.yaw += 360.0;


            Vec3 nx = x;
            Utils.RotateVector(ref nx, 0.0, 0.0, -o.yaw/*wtf?*/);
            o.roll = MathLib.ToDeg(Math.Acos(nx.x));

            if (MathLib.CompareScalar(nx.y, 0.0) < 0)
                o.roll = 360.0 - o.roll;

            Vec3 nz = z;
            Utils.RotateVector(ref nz, 0.0, 0.0, -o.yaw/*wtf?*/);
            Utils.RotateVector(ref nz, 0.0, -o.roll/*wtf?*/, 0.0);

            o.pitch = MathLib.ToDeg(Math.Acos(nz.z));
            if (MathLib.CompareScalar(nz.y, 0.0) > 0)
                o.pitch = 360.0 - o.pitch;

            return o;
        }

        public Orient GetPYR()
        {
            return sysGetPYR(renderer.XVec, renderer.YVec, renderer.ZVec);
        }

        private void Options1_Click(object sender, EventArgs e)
        {
            if (new Options().ShowDialog(this) != System.Windows.Forms.DialogResult.OK)
                return;

            if (GlobalVars.MapRot == GlobalVars.MapRotation.Old)
                camOrient = GetPYR();
        }

        private void NedMain_FormClosed(object sender, FormClosedEventArgs e)
        {
            GlobalVars.MWMaxed = WindowState == FormWindowState.Maximized;
            if(WindowState == FormWindowState.Normal)
                GlobalVars.MWinPos = Bounds;

            //Toolbar.Close();
            //GlobalVars.ItemEdit.Close();
            //GlobalVars.Preview3D.Close();

            // clear out the skeletons
            GlobalVars.Cleanup();
        }

        private void NedMain_Load(object sender, EventArgs e)
        {
            NedRunning = true;

            if (!GlobalVars.MWinPos.IsEmpty)
                Bounds = GlobalVars.MWinPos;
            WindowState = GlobalVars.MWMaxed ? FormWindowState.Maximized : FormWindowState.Normal;
        }

        public static bool NedRunning = false;

        public static bool NedFocused
        {
            get
            {
                if (!NedRunning)
                    return false;

                Form active = Form.ActiveForm;
                if (active == null)
                    return false;

                if (object.ReferenceEquals(active, GlobalVars.NedMain))
                    return true;

                if (active is SubForm)
                    return true;

                return false;
            }
        }

        private void N3DPreview1_Click(object sender, EventArgs e)
        {
            GlobalVars.Preview3D.ShowPreview();
        }

        private void BNSC_Click(object sender, EventArgs e)
        {
            if (sender == BNSC)
                SetMapMode(MapMode.Sector);
            else if (sender == BNSF)
                SetMapMode(MapMode.Surface);
            else if (sender == BNED)
                SetMapMode(MapMode.Edge);
            else if (sender == BNVX)
                SetMapMode(MapMode.Vertex);
            else if (sender == BNTH)
                SetMapMode(MapMode.Thing);
            else if (sender == BNLT)
                SetMapMode(MapMode.Light);
            else if (sender == BNEX)
                SetMapMode(MapMode.Extra);
            else if (sender == BNFR)
                SetMapMode(MapMode.Frame);
        }

        private void TemplateCreator1_Click(object sender, EventArgs e)
        {
            LoadTemplates();
            TemplateCreator.CreateTemplate();
        }

        private void SaveAs1_Click(object sender, EventArgs e)
        {
            SysSaveProject(true);
        }

        private void Save1_Click(object sender, EventArgs e)
        {
            SaveProject();
        }

        private void Reverttosaved1_Click(object sender, EventArgs e)
        {
            if (AskSaveAs)
            {
                MessageBox.Show("The project wasn't saved!");
                return;
            }

            if (GlobalVars.ConfirmRevert && changed)
            {
                if (MessageBox.Show("Revert to Saved?", "Warning", MessageBoxButtons.YesNo) != System.Windows.Forms.DialogResult.Yes)
                    return;
            }

            OpenProject(LevelFile, OpenMethod.Revert);
        }

        private void MakeaBackupCopy1_Click(object sender, EventArgs e)
        {
            if (string.IsNullOrEmpty(GlobalVars.ProjectDir))
            {
                MessageBox.Show("Save the level first");
                return;
            }

            Directory.CreateDirectory(Path.Combine(GlobalVars.ProjectDir, "backup"));

            string outfile = string.Empty;
            for (int n = 0; n < 100; n++)
            {
                outfile = Path.Combine(Path.Combine(Path.GetDirectoryName(LevelFile), "backup"), Path.GetFileNameWithoutExtension(LevelFile) + "-" + n.ToString("00") + Path.GetExtension(LevelFile));
                if (!File.Exists(outfile))
                    break;
            }

            if (string.IsNullOrEmpty(outfile))
            {
                MessageBox.Show("You have 100 backup copies! Remove some");
                return;
            }

            File.Copy(LevelFile, outfile);
            Utils.PanMessage(Utils.MessageType.Info, "Made a backup copy " + outfile);
        }

        private int AddMat(JKModel model, string mat)
        {
            for (int i = 0; i < model.mats.Count; i++)
                if (model.mats[i].Equals(mat, StringComparison.InvariantCultureIgnoreCase))
                    return i;

            model.mats.Add(mat);
            return (model.mats.Count - 1);
        }

        private void AddSecToMesh(JKModel model, JKModel.Mesh mesh, JKSector sec, Vec3 center, bool skipSky)
        {
            // add vertices
            foreach (JKVertex vert in sec.vertices)
                vert.mark = mesh.AddVertex(vert.v - center);

            // add faces
            foreach (JKSurface surf in sec.surfaces)
            {
                if (surf.adjoin != null)
                    continue;

                if (skipSky && (surf.surfaceFlags & (Surface.Flag.Ceiling | Surface.Flag.Horizon)) != 0)
                    continue;

                JKModel.Face face = new JKModel.Face();
                face.imat = AddMat(model, surf.material);
                face.ftype = 0;
                if ((surf.faceFlags & Surface.FaceFlag.Transparent) != 0)
                    face.ftype = 2;
                face.geo = (int)surf.geo;
                face.light = (int)surf.light;
                face.tex = (int)surf.tex;
                face.extra_l = surf.extraLight;
                for (int j = surf.vertices.Count - 1; j >= 0; j--)
                {
                    Vertex jv = surf.vertices[j];
                    TexVertex jtv = surf.texvertices[j];

                    int n = face.AddVertex(mesh.vertices[jv.mark]);
                    face.texvertices[n].v = jtv.v;

                    face.vertNormals.AddRange(surf.vertNormals);
                }

                mesh.faces.Add(face);
            }
        }

        public enum ExportModelMode
        {
            JK3do,
            OBJ
        }

        public void ExportModel(string outpath, ExportModelMode mode)
        {
            JKSector sec = Level.sectors[Cur_SC];
            JKModel model = new JKModel();

            Vec3 center = Vec3.Zero;
            if (Level.hnodes.Count == 0)
                center = Utils.FindCenter(sec);

            SCMultiSel nsel = new SCMultiSel();
            for (int i = 0; i < scsel.Count; i++)
                nsel.AddSC(scsel.GetSC(i));

            nsel.AddSC(Cur_SC);

            while (nsel.Count > 0)
            {
                sec = Level.sectors[nsel.GetSC(0)];
                JKModel.Mesh mesh = new JKModel.Mesh();
                mesh.name = Level.layers[sec.layer];
                model.meshes.Add(mesh);

                for (int i = nsel.Count - 1; i >= 0; i--)
                {
                    JKSector csec = Level.sectors[nsel.GetSC(i)];
                    if (csec.layer != sec.layer)
                        continue;

                    csec.Renumber();
                    AddSecToMesh(model, mesh, csec, center, mode == ExportModelMode.OBJ/*skip sky if OBJ? should be option*/);
                    nsel.DeleteN(i);
                }
            }

            // build hierarchy
            if (Level.hnodes.Count == 0)
            {
                // no hierarchy defined; create default

                JKModel.HNode hnode = new JKModel.HNode();
                model.hnodes.Add(hnode);
                if (model.meshes.Count == 1)
                {
                    hnode.nodename = model.meshes[0].name;
                    hnode.nmesh = 0;
                }
                else
                {
                    hnode.nodename = "$$DUMMY";
                    for (int i = 0; i < model.meshes.Count; i++)
                    {
                        hnode = new JKModel.HNode();
                        model.hnodes.Add(hnode);
                        hnode.nodename = model.meshes[i].name;
                        hnode.nmesh = i;
                        hnode.parent = 0;
                    }
                }
            }
            else
            {
                for (int i = 0; i < Level.hnodes.Count; i++)
                {
                    JKModel.HNode hnode = new JKModel.HNode();
                    hnode.Assign(Level.hnodes[i]);
                    model.hnodes.Add(hnode);
                }

                for (int i = 0; i < model.meshes.Count; i++)
                {
                    int n = 0;
                    for (int j = 0; j < model.hnodes.Count; j++)
                        if (model.hnodes[j].nodename.Equals(model.meshes[i].name, StringComparison.InvariantCultureIgnoreCase))
                        {
                            n++;
                            model.hnodes[j].nmesh = i;
                        }

                    if (n == 0)
                    {
                        // no nodes for this mesh
                        JKModel.HNode hnode = new JKModel.HNode();
                        hnode.parent = 0;
                        hnode.nmesh = i;
                        hnode.nodename = model.meshes[i].name;
                        model.hnodes.Add(hnode);
                    }
                }

                // check that there are no nodes pointing to non-existing meshes
                for (int i = 0; i < model.hnodes.Count; i++)
                {
                    JKModel.HNode hnode = model.hnodes[i];
                    int n = 0;
                    for (int j = 0; j < model.meshes.Count; j++)
                        if (hnode.nodename.Equals(model.meshes[j].name, StringComparison.InvariantCultureIgnoreCase))
                            n++;

                    if (n == 0)
                        hnode.nmesh = -1;
                }
            }

            foreach (JKModel.Mesh mesh in model.meshes)
                foreach (JKModel.Face face in mesh.faces)
                    face.Recalc();

            if (mode == ExportModelMode.JK3do)
                model.SaveTo3DO(outpath);
            else if (mode == ExportModelMode.OBJ)
                model.SaveToOBJ(outpath, JKModel.OBJScale, (Level.sectors[0].flags & Sector.Flag.Preview3DO)!=0, Level.GetMasterCMP(), JKModel.OBJTextureOutput.Tga);
        }

        private void ExportSectoras3DO1_Click(object sender, EventArgs e)
        {
            if (map_mode != MapMode.Sector)
                return;

            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "3DO Files (*.3DO)|*.3DO";
            sfd.Title = "Save 3DO";
            if (sfd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            ExportModel(sfd.FileName, ExportModelMode.JK3do);
        }

        private void ExportSectorasShape1_Click(object sender, EventArgs e)
        {
            string s = string.Empty;
            if (Utils.InputBox("New Shape Name", "Name:", ref s) != System.Windows.Forms.DialogResult.OK)
                return;

            Prefabs.AddPrefab(s, Level.sectors[Cur_SC]);
            Prefabs.SavePrefabs(Path.Combine(GlobalVars.BaseDir, @"JedData\shapes.tpl"));
            Toolbar.UpdatePrefabs();
        }

        void Import3DO(string name)
        {
            GlobalVars.NedMain.HaltLightCalc();

            JKModel model = new JKModel(name, 0);
            Level.Clear();
            try
            {
                foreach (JKModel.Mesh mesh in model.meshes)
                {
                    JKSector sec = Level.NewSector();
                    sec.flags |= Sector.Flag.Preview3DO;
                    sec.layer = Level.AddLayer(mesh.name);
                    Level.sectors.Add(sec);
                    for (int j = 0; j < mesh.vertices.Count; j++)
                    {
                        Vertex v3 = mesh.vertices[j];
                        v3.num = j;
                        JKVertex v = sec.NewVertex();
                        v.v = v3.v;
                    }

                    for (int j = 0; j < mesh.faces.Count; j++)
                    {
                        JKModel.Face face = mesh.faces[j] as JKModel.Face/*hope so*/;
                        JKSurface surf = sec.NewSurface();
                        sec.surfaces.Add(surf);

                        surf.adjoin = null;
                        surf.material = model.GetMat(face.imat);
                        surf.SetGeo("Import3DO", (Surface.GeoModeType)face.geo);
                        surf.light = (Surface.LightModeType)face.light;
                        surf.tex = (Surface.TextureModeType)face.tex;
                        if ((face.ftype & 0x02) != 0)
                            surf.faceFlags |= Surface.FaceFlag.Transparent;
                        surf.extraLight = face.extra_l;

                        int n = face.vertices.Count;
                        for (int k = n - 1; k >= 0; k--)
                        {
                            surf.AddVertex(sec.vertices[face.vertices[k].num]);
                            surf.texvertices[n - k - 1].Assign(face.texvertices[k]);
                            if(face.vertNormals.Count > 0)
                                surf.vertNormals.Add(face.vertNormals[k]);
                        }

                        surf.Recalc();
                        surf.FindUVScales();
                    }
                }

                for (int i = 0; i < model.hnodes.Count; i++)
                {
                    JKModel.HNode hnode = Level.New3DONode();
                    hnode.Assign(model.hnodes[i]);
                    Level.hnodes.Add(hnode);
                    hnode.pivot = Vec3.Zero;
                }
            }
            finally
            {
                Undo.ClearUndoBuffer();
                Level.things.Add(Level.NewThing());
                Level.RenumSecs();
                foreach (JKSector sec in Level.sectors)
                    sec.Renumber();
            }

            GlobalVars.NedMain.ResumeLightCalc();
        }

        bool Import3DS(string name)
        {
#if false
            Import3DS importDialog = new Import3DS();
            if (importDialog.ShowDialog() != DialogResult.OK)
                return;

            GlobalVars.ProjectDir = importDialog.outputdir.Text;
#else
            FolderBrowserDialog fbd = new FolderBrowserDialog();
            fbd.Description = "Select output project folder";
            fbd.ShowNewFolderButton = true;
            if (fbd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return false;

            string scaletxt = "1.0";
            if (Utils.InputBox("Scale", "Input scale factor", ref scaletxt) != DialogResult.OK)
                return false;

            double scale;
            if (!double.TryParse(scaletxt, out scale) || scale <= 0.0)
                return false;

            GlobalVars.ProjectDir = fbd.SelectedPath;
#endif

            LevelFile = Path.Combine(GlobalVars.ProjectDir, Path.GetFileName(name));
            File.Copy(name, LevelFile, true);

            GlobalVars.NedMain.HaltLightCalc();

            Autodesk3DS max = new Autodesk3DS(name);
            JKModel model = max.GenerateJKModel(GlobalVars.ProjectDir, scale);
            Level.Clear();
            try
            {
                foreach (JKModel.Mesh mesh in model.meshes)
                {
                    JKSector sec = Level.NewSector();
                    sec.flags |= Sector.Flag.Preview3DO;
                    sec.layer = Level.AddLayer(mesh.name);
                    Level.sectors.Add(sec);
                    for (int j = 0; j < mesh.vertices.Count; j++)
                    {
                        Vertex v3 = mesh.vertices[j];
                        v3.num = j;
                        JKVertex v = sec.NewVertex();
                        v.v = v3.v;
                    }

                    for (int j = 0; j < mesh.faces.Count; j++)
                    {
                        JKModel.Face face = mesh.faces[j] as JKModel.Face/*hope so*/;
                        JKSurface surf = sec.NewSurface();
                        sec.surfaces.Add(surf);

                        surf.adjoin = null;
                        surf.material = model.GetMat(face.imat);
                        surf.SetGeo("Import3DS", (Surface.GeoModeType)face.geo);
                        surf.light = (Surface.LightModeType)face.light;
                        surf.tex = (Surface.TextureModeType)face.tex;
                        if ((face.ftype & 0x02) != 0)
                            surf.faceFlags |= Surface.FaceFlag.Transparent;
                        surf.extraLight = face.extra_l;

                        int n = face.vertices.Count;
                        for (int k = n - 1; k >= 0; k--)
                        {
                            surf.AddVertex(sec.vertices[face.vertices[k].num]);
                            surf.texvertices[n - k - 1].Assign(face.texvertices[k]);
                        }

                        surf.Recalc();
                        surf.FindUVScales();
                    }
                }

                for (int i = 0; i < model.hnodes.Count; i++)
                {
                    JKModel.HNode hnode = Level.New3DONode();
                    hnode.Assign(model.hnodes[i]);
                    Level.hnodes.Add(hnode);
                    hnode.pivot = Vec3.Zero;
                }
            }
            finally
            {
                Undo.ClearUndoBuffer();
                Level.things.Add(Level.NewThing());
                Level.RenumSecs();
                foreach (JKSector sec in Level.sectors)
                    sec.Renumber();
            }

            GlobalVars.NedMain.ResumeLightCalc();

            return true;
        }

        private void SaveJKL1_Click(object sender, EventArgs e)
        {
            DoSaveJKL();
        }

        string GenerateJKLFileName()
        {
            return FileOperations.ChangeFileExt(Path.GetFileName(LevelFile), ".jkl");
        }

        public bool DoSaveJKL()
        {
            if (FileOperations.IsInContainer(LevelFile))
            {
                Utils.PanMessage(Utils.MessageType.Error, "File is in container - try \"SaveJKL to\"");
                return false;
            }

            if (NewLevel)
            {
                Utils.PanMessage(Utils.MessageType.Error, "Project was never saved");
                return false;
            }

            string s = GetTargetJKLName();
            string dirroot = Path.GetDirectoryName(s);
            Directory.CreateDirectory(dirroot);

            string dstfile = Path.Combine(GlobalVars.ProjectDir, GenerateJKLFileName());
            if (File.Exists(dstfile))
                File.Delete(dstfile);

            Level.SaveToJKL(s, false);
            GlobalVars.EpisodeEdit.Save();

            return true;
        }

        string GetTargetJKLName()
        {
            if (FileOperations.ExtractExt(LevelFile).Equals(".jkl", StringComparison.InvariantCultureIgnoreCase))
                return LevelFile;

            return Path.Combine(Path.Combine(GlobalVars.ProjectDir, "jkl"), FileOperations.ChangeFileExt(Path.GetFileName(LevelFile), ".jkl"));
        }

        private void Export1_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "JKL|*.JKL";
            sfd.Title = "Save file as";
            sfd.FileName = FileOperations.ChangeFileExt(FileOperations.ExtractName(LevelFile), ".jkl");
            if (sfd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            Level.SaveToJKL(sfd.FileName, false);
        }

        private string lastSaveGob = string.Empty;

        private class GobProjHelper
        {
            string name;
            GOB2Creator gob;
            List<string> flist = new List<string>();
            List<string> slist = new List<string>();

            int njkls = 0;

            public GobProjHelper(string _name)
            {
                name = _name;
                gob = new GOB2Creator(name);
            }

            void AddFiles(string subdir, string mask)
            {
                foreach (string fn in FileOperations.ListDirMask(Path.Combine(GlobalVars.ProjectDir, subdir), mask))
                {
                    string sfn = Path.Combine(subdir, fn);

                    int match = -1;
                    for(int n=0; n<flist.Count; n++)
                        if (flist[n].Equals(sfn, StringComparison.InvariantCultureIgnoreCase))
                        {
                            match = n;
                            break;
                        }

                    if (match != -1)
                    {
                        Utils.PanMessage(Utils.MessageType.Warning, "Duplicate files in project directory: " + slist[match] + " and " + sfn);
                        slist[match] = sfn;
                    }
                    else
                    {
                        flist.Add(sfn);
                        slist.Add(sfn);
                    }

                    if (mask == "*.jkl")
                        njkls++;
                }
            }

            List<string> extlist;
            List<string> checklist;

            void InitSmartGob()
            {
                extlist = new List<string>(new string[] {
                    ".wav",
                    ".mat",
                    ".ai",
                    ".3do",
                    ".spr",
                    ".key",
                    ".pup",
                    ".snd",
                    ".cog"
                });
                extlist.Sort();

                checklist = new List<string>();

                foreach (string fn in slist)
                    if (Path.GetExtension(fn).Equals(".jkl", StringComparison.InvariantCultureIgnoreCase))
                        ListRes.LoadJKLLists(Path.Combine(GlobalVars.ProjectDir, fn), ref checklist);

                ItemsDatFile id = new ItemsDatFile();
                id.Load("items.dat");
                checklist.AddRange(id.COGs);

                ModelsDatFile md = new ModelsDatFile();
                md.Load("models.dat");
                checklist.AddRange(md.snds);
                checklist.AddRange(md.a3dos);
            }

            void CheckFList()
            {
                for (int i = slist.Count - 1; i >= 0; i--)
                {
                    string s = Path.GetFileName(slist[i]);
                    if (!extlist.Contains(Path.GetExtension(s), StringComparer.InvariantCultureIgnoreCase))
                        continue;

                    if (!checklist.Contains(s, StringComparer.InvariantCultureIgnoreCase))
                        continue;

                    slist.RemoveAt(i);
                    flist.RemoveAt(i);
                }
            }

            public bool Do()
            {
                foreach (string s in FileOperations.ListDirMask(GlobalVars.ProjectDir, "*.*"))
                {
                    switch (Path.GetExtension(s).ToUpperInvariant())
                    {
                        case ".JK": flist.Add(s); break;
                        case ".JKL": flist.Add(@"jkl\" + s); njkls++; break;
                        case ".COG": flist.Add(@"cog\" + s); break;
                        case ".3DO": flist.Add(@"3do\" + s); break;
                        case ".MAT": flist.Add(@"mat\" + s); break;
                        case ".DAT": flist.Add(@"misc\" + s); break;
                        case ".UNI":
                            string tmp = Path.GetFileName(s);
                            if (tmp.Equals("cogstrings.uni", StringComparison.InvariantCultureIgnoreCase) || tmp.Equals("sithstrings.uni", StringComparison.InvariantCultureIgnoreCase))
                                flist.Add(@"misc\" + s);
                            else
                                flist.Add(@"ui\" + s);
                            break;
                        case ".KEY": flist.Add(@"3do\key\" + s); break;
                        case ".AI0": case ".AI2": case ".AI": flist.Add(@"misc\ai\" + s); break;
                        case ".CMP": flist.Add(@"misc\cmp\" + s); break;
                        case ".PAR": flist.Add(@"misc\par\" + s); break;
                        case ".PER": flist.Add(@"misc\per\" + s); break;
                        case ".PUP": flist.Add(@"misc\pup\" + s); break;
                        case ".SND": flist.Add(@"misc\snd\" + s); break;
                        case ".SPR": flist.Add(@"misc\spr\" + s); break;
                        case ".WAV": flist.Add(@"sound\" + s); break;
                        case ".BM": flist.Add(@"ui\bm\" + s); break;
                        case ".SFT": flist.Add(@"ui\sft\" + s); break;
                        case ".SMK": flist.Add(@"video\" + s); break;
                        case ".SAN": flist.Add(@"video\" + s); break;
                    }
                }

                // LEFT OFF HERE
                foreach (string fn in flist)
                    slist.Add(Path.GetFileName(fn));

                AddFiles(@"cog\", "*.cog");
                AddFiles(@"jkl\", "*.jkl");
                AddFiles(@"3do\", "*.3do");
                AddFiles(@"3do\mat\", "*.mat");
                AddFiles(@"mat\", "*.mat");
                AddFiles(@"misc\", "*.uni");
                AddFiles(@"ui\", "*.uni");
                AddFiles(@"misc\", "*.dat");
                AddFiles(@"3do\key\", "*.key");
                AddFiles(@"misc\ai\", "*.ai;*.ai2;*.ai0");
                AddFiles(@"misc\cmp\", "*.cmp");
                AddFiles(@"misc\par\", "*.par");
                AddFiles(@"misc\per\", "*.per");
                AddFiles(@"misc\pup\", "*.pup");
                AddFiles(@"misc\snd\", "*.snd");
                AddFiles(@"misc\spr\", "*.spr");
                AddFiles(@"sound\", "*.wav");
                AddFiles(@"voice\", "*.wav");
                AddFiles(@"voiceuu\", "*.wav");
                AddFiles(@"ui\bm\", "*.bm");
                AddFiles(@"ui\sft\", "*.sft");

                if (njkls == 0)
                    Utils.PanMessage(Utils.MessageType.Warning, "No JKL files found in project directory");

                if (GlobalVars.GOBSmart)
                {
                    InitSmartGob();
                    CheckFList();
                }

                ProgressDialog.Reset(flist.Count, 0);

                gob.PrepareHeader(flist);
                foreach (string fn in slist)
                {
                    ProgressDialog.Step();
                    using (BaseFile bf = FileOperations.OpenFileRead(Path.Combine(GlobalVars.ProjectDir, fn)))
                        gob.AddFile(bf);
                }

                gob.Dispose();

                ProgressDialog.Cleanup();

                return true;
            }
        }

        public bool GobProj(string name)
        {
            if (NewLevel)
            {
                Utils.PanMessage(Utils.MessageType.Error, "Project was never saved");
                return false;
            }

            if (FileOperations.IsInContainer(LevelFile))
            {
                Utils.PanMessage(Utils.MessageType.Error, "Level is in container - can't gob");
                return false;
            }

            GobProjHelper helper = new GobProjHelper(name);
            helper.Do();

            return true;
        }

        private void SaveJKLGob1_Click(object sender, EventArgs e)
        {
            if (!DoSaveJKL())
                return;

            string ext = Path.GetExtension(LevelFile).ToUpper();

            SaveFileDialog savegob = new SaveFileDialog();
            savegob.Filter = "GOB Files|*.gob|GOO Files|*.goo";
            savegob.Title = "Select a GOB to create";
            string gobdir = !string.IsNullOrEmpty(lastSaveGob) ? Path.GetDirectoryName(lastSaveGob) : Path.Combine(GlobalVars.GameDir, @"EPISODE");
            savegob.InitialDirectory = gobdir;

            if (isMots)
            {
                savegob.FileName = /*Path.Combine(gobdir, */Path.GetFileNameWithoutExtension(LevelFile) + ".goo";//);
                savegob.FilterIndex = 2;
                savegob.DefaultExt = "goo";
            }
            else
            {
                savegob.FileName = /*Path.Combine(gobdir, */Path.GetFileNameWithoutExtension(LevelFile) + ".gob";//);
                savegob.FilterIndex = 1;
                savegob.DefaultExt = "gob";
            }

            if (savegob.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            lastSaveGob = savegob.FileName;
            GobProj(savegob.FileName);
        }

        public void LoadGOBResources()
        {
            Resource.Cleanup(true, true, true, "NED LoadGOBResources");

            if (Program.DemoMode)
            {
                Resource.AddDirectory(Program.ExecutableDirectory);
                Resource.AddGOB("res1demo.gob", false, "NED LoadGOBResources");
                return;
            }

            Resource.AddDirectory(GlobalVars.JKDir);
            Resource.AddDirectory(Path.Combine(GlobalVars.JKDir, "Resource"));

            Resource.AddDirectory(GlobalVars.DWDir);//need a better arch for all this

            if(JK.GameType == GameType.DW)
            {
                Resource.AddGOB("dwHD.gob", false, "NED LoadGOBResources");
                Resource.AddGOB("dwMin.gob", false, "NED LoadGOBResources");
                Resource.AddGOB("dwStream.gob", false, "NED LoadGOBResources");
                Resource.AddGOB("dwCD.gob", false, "NED LoadGOBResources");
            }
            else
            if (isMots)
            {
                Resource.AddGOB("jkmres.goo", false, "NED LoadGOBResources");
                Resource.AddGOB("jkmsndlo.goo", false, "NED LoadGOBResources");
            }
            else
            {
                Resource.AddGOB("res2.gob", false, "NED LoadGOBResources");
                Resource.AddGOB("res1hi.gob", false, "NED LoadGOBResources");
            }
        }

        public string LevelFileJKL
        {
            get
            {
                return LevelFile.Substring(LevelFile.LastIndexOfAny(new char[] { '>', '\\', '/' }) + 1);
            }
        }

        public string[] GenerateSmithCmdLine(bool neo, bool mini, bool sync, bool noprecache)
        {
            List<string> args = new List<string>(new string[] {
                "DUMMY",    // normally would be the application path
                "-ignorejed",
                "-clearbvhcache"
            });

            if (Level.mots)
                args.Add("-mots");

            if (!string.IsNullOrEmpty(GlobalVars.ProjectGob))
            {
                // just run JKL from gob
                args.AddRange(new string[] {
                    "-path",
                    "\"" + GlobalVars.ProjectGob + "\"",
                    "-jkl",
                    "\"" + LevelFileJKL + "\""
                });
            }
            else
            {
                // only add for active projects;  if loading JKL from gob who cares  do normal launch
                if (noprecache)
                    args.Add("-noprecache");

                args.AddRange(new string[] {
                    "-path",
                    "\"" + GlobalVars.ProjectDir + "\"",
                    "-jkl",
                    "\"" + GenerateJKLFileName() + "\""
                });
            }

            if (mini)
                args.Add("-mini");

            // reuse some of jed shit to generate frame string for teleporting the player
            if (GlobalVars.Preview3D.Visible && GlobalVars.Preview3D.renderer != null && sync)
            {
                Value frame = new Value();
                frame.name = "frame";
                frame.SetFrame(GlobalVars.Preview3D.renderer.CamPos, GlobalVars.Preview3D.renderer.CamOrient);
                args.AddRange(new string[] {
                    "-teleport",
                    "\"" + frame.AsString + "\""
                });
            }

            if (neo)
                args.Add("-neo");

            return args.ToArray();
        }

        private void SaveJKLGOBandTestSmith1_Click(object sender, EventArgs e)
        {
            if (FileOperations.IsInContainer(LevelFile))
            {
                //Utils.PanMessage(Utils.MessageType.Error, "File is in container - try \"SaveJKL to\"")
                if (MessageBox.Show("This level is still inside a GOB and not yet saved.\n\nDo you want to just Smith on the original gob/jkl?", "Run Original", MessageBoxButtons.OKCancel) != System.Windows.Forms.DialogResult.OK)
                    return;
            }
            else
            {
                if (!DoSaveJKL())
                    return;
            }

            // clear whatever we have loaded for ned
            Resource.Cleanup(true, true, true, "NED SaveJKLGOBandTestSmith");

            Program.RunSmith(this, GenerateSmithCmdLine(GlobalVars.SmithNeo, GlobalVars.SmithMini, GlobalVars.SmithSyncCameras, GlobalVars.SmithNoPrecache));

            // reload our stuff
            LoadGOBResources();
        }

        [DllImport("user32.dll")]
        static extern uint PostMessage(IntPtr hWnd, uint Msg, uint wParam, uint lParam);

        private bool handleFirstClickOnActivated = false;

        protected override void OnActivated(EventArgs e)
        {
            base.OnActivated(e);

            if (handleFirstClickOnActivated)
            {
                Point cursorPosition = Cursor.Position;
                Point clientPoint = PointToClient(cursorPosition);
                Control child = GetChildAtPoint(clientPoint);
                while (handleFirstClickOnActivated && child != null)
                {
                    ToolStrip toolstrip = child as ToolStrip;
                    if (toolstrip != null)
                    {
                        handleFirstClickOnActivated = false;
                        clientPoint = toolstrip.PointToClient(cursorPosition);
                        foreach(ToolStripItem ts in toolstrip.Items)
                            if (ts != null && ts.Bounds.Contains(clientPoint))
                            {
                                ToolStripMenuItem mi = ts as ToolStripMenuItem;
                                if (mi != null)
                                {
                                    mi.ShowDropDown();
                                }
                                else
                                {
                                    ts.PerformClick();
                                }
                                break;
                            }
                    }
                    else
                    {
                        clientPoint = child.PointToClient(cursorPosition);
                        child = child.GetChildAtPoint(clientPoint);
                    }
                }
                handleFirstClickOnActivated = false;
            }

            // trigger a real mouse event upon first activation if we are inside the form (eg. we clicked on it rather than alt-tabbed or selected from taskbar)
            /*if (DesktopBounds.Contains(Cursor.Position))
            {
                int x = Cursor.Position.X - DesktopBounds.Left;
                int y = Cursor.Position.Y - DesktopBounds.Top;
                uint xy = (uint)(x | (y << 16));

                Application.DoEvents();// this is required.. i dunno why.. it works i dont ask questions

                // trigger click
                PostMessage(Handle, 0x201, 1, xy);
                PostMessage(Handle, 0x202, 1, xy);
            }*/
        }

        private void PlcedCogs1_Click(object sender, EventArgs e)
        {
            GlobalVars.CogForm.Show();
        }

        private void Messages1_Click(object sender, EventArgs e)
        {
            GlobalVars.MsgForm.Show();
        }

        private void miUndo_Click(object sender, EventArgs e)
        {
            Undo.ApplyUndo();
            VerifyMultiSelection();
            UpdateItemEditor();
        }

        private void N3DOHierarchy1_Click(object sender, EventArgs e)
        {
            GlobalVars.UrqForm.Reload();
            GlobalVars.UrqForm.Show();
        }

        private void ConsistencyCheck1_Click(object sender, EventArgs e)
        {
            GlobalVars.ConsistencyChecker.Check();
        }

        private void CheckResources1_Click(object sender, EventArgs e)
        {
            GlobalVars.ConsistencyChecker.CheckResources();
        }

        private void HideThings_Click(object sender, EventArgs e)
        {
            SetHideThings(!ThingsHidden);
        }

        private void HideLights_Click(object sender, EventArgs e)
        {
            SetHideLights(!LightsHidden);
        }

        private void CalculateLights1_Click(object sender, EventArgs e)
        {
            if (GlobalVars.CalculateLightsAuto)
            {
                // just do the dirty flag
                lightsDirty = true;

                //ensure, since we're on autocalculte, that the thread is resumed
                ResumeLightCalc();

                return;
            }

            SCMultiSel scs = new SCMultiSel();

            if(object.ReferenceEquals(sender, CalcLightOnLayers))
            {
                for(int i=0; i<Level.sectors.Count; i++)
                    if(Toolbar.IsLayerVisible(Level.sectors[i].layer))
                        scs.AddSC(i);
            } else if(object.ReferenceEquals(sender, CalcLightInSel))
            {
                switch(map_mode)
                {
                    case MapMode.Sector:
                        for (int i = 0; i < scsel.Count; i++)
                            scs.AddSC(scsel.GetSC(i));
                        break;

                    case MapMode.Surface:
                        for (int i = 0; i < sfsel.Count; i++)
                        {
                            int sc, sf;
                            sfsel.GetSCSF(i, out sc, out sf);
                            scs.AddSC(sc);
                        }
                        break;

                    case MapMode.Edge:
                        for (int i = 0; i < edsel.Count; i++)
                        {
                            int sc, sf, a;
                            edsel.GetSCSFED(i, out sc, out sf, out a);
                            scs.AddSC(sc);
                        }
                        break;

                    case MapMode.Vertex:
                        for (int i = 0; i < vxsel.Count; i++)
                        {
                            int sc, a;
                            vxsel.GetSCVX(i, out sc, out a);
                            scs.AddSC(sc);
                        }
                        break;

                    default:
                        MessageBox.Show("You must be in sector,surface,edge or vertex mode to use this option");
                        return;
                }
            } else {
                for(int i=0; i<Level.sectors.Count; i++)
                    scs.AddSC(i);
            }

            Undo.ClearUndoBuffer();
            DateTime start = DateTime.UtcNow;

            if(GlobalVars.NewLightCalc)
                Utils.CalcLightingNew(Level, scs, false);
            else
                Utils.CalcLighting(Level, scs, false);

            int msecs = (int)DateTime.UtcNow.Subtract(start).TotalMilliseconds;
            Utils.PanMessage(Utils.MessageType.Info, Level.lights.Count.ToString() + " lights processed in " + msecs.ToString());

            if(Level.sectors.Count == scs.Count)
                GlobalVars.Preview3D.ReloadLevel();
            else
                for(int i=0; i<scs.Count; i++)
                    SectorChanged(Level.sectors[scs.GetSC(i)], false/*dont fire another one off*/);

            LevelChanged(false/*dont fire another one off*/);
        }

        private void Exit1_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void Levelheadereditor1_Click(object sender, EventArgs e)
        {
            LevelHeaderEditor le = new LevelHeaderEditor();

            if (le.ShowDialog(this) != DialogResult.OK)
                return;

            if (le.NeedPreviewReload)
                GlobalVars.Preview3D.ReloadLevel();

            LevelChanged(false);
        }

        private void splitContainer1_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            DoPreviewKeyDown(e);
        }

        private void splitContainer1_KeyDown(object sender, KeyEventArgs e)
        {
            DoKeyDown(e);
        }

        private void splitContainer1_KeyUp(object sender, KeyEventArgs e)
        {
            DoKeyUp(e);
        }

        private void Tools1_DropDownOpening(object sender, EventArgs e)
        {
            if (GlobalVars.CalculateLightsAuto)
            {
                CalcLightInSel.Visible = false;
                CalcLightOnLayers.Visible = false;

                CalculateLights1.Enabled = !lightsDirty && !calcingLights;
                calcLightsSep.Visible = true;
            }
            else
            {
                CalculateLights1.Enabled = true;
                CalculateLights1.Visible = true;
                CalcLightInSel.Visible = true;
                CalcLightOnLayers.Visible = true;
                calcLightsSep.Visible = true;
            }
        }

        private void matAttributeEdit_Click(object sender, EventArgs e)
        {
            MaterialAttributeEditor mae = new MaterialAttributeEditor();
            mae.ShowDialog(this);
        }

        private void GenerateModGob_Click(object sender, EventArgs e)
        {
            JKLevel.ClassSaveToJKL.UsedListMatrix matrix = new JKLevel.ClassSaveToJKL.UsedListMatrix(true);

            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "Mod Gob (*.gob)|*.gob";
            if (sfd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            matrix.staticJKLPath = Path.GetTempFileName();

            using (StreamWriter sw = File.CreateText(matrix.staticJKLPath))
            {
                JKLevel.ClassSaveToJKL c = new JKLevel.ClassSaveToJKL(Level, sw);
                c.Save(false, true, matrix);
            }

            matrix.Finish(true);

            matrix.SaveGOB(sfd.FileName);
        }

        private void extractProjectFromGOB_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "All acceptable files|*.gob;*.goo";
            ofd.Title = "Select GOB";
            if (ofd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            FolderBrowserDialog fbd = new FolderBrowserDialog();
            fbd.Description = "Select output project folder";
            fbd.ShowNewFolderButton = true;
            if (fbd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            ExtractProject(ofd.FileName, fbd.SelectedPath, true);
        }

        private void viewToGridToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetViewToGrid();
        }

        private void gridToViewToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetGridToView();
        }

        public void SetGridToView()
        {
            Vec3 x = Vec3.PosX;
            Vec3 z = Vec3.NegY;

            x = renderer.XVec;
            z = renderer.ZVec;

            renderer.SetGridNormal(z);
            renderer.SetGridXNormal(x);
        }

        public void SetViewToGrid()
        {
            renderer.SetZ(renderer.GridNormal);
            renderer.SetX(renderer.GridXVec);

            renderer.CamPos = -renderer.GridPos;

            if (GlobalVars.MapRot == GlobalVars.MapRotation.Old)
            {
                GetPYR();
                SetRendFromPYR();
            }
        }

        private void ExportSectorAsOBJ_Click(object sender, EventArgs e)
        {
            if (map_mode != MapMode.Sector)
                return;

            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "OBJ Files (*.OBJ)|*.OBJ";
            sfd.Title = "Save OBJ";
            if (sfd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            ExportModel(sfd.FileName, ExportModelMode.OBJ);
        }

        private void saveJKLToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DoSaveJKL();
        }

        private void showWeirdAdjoinsIn3DPreviewToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetShowWeirdAdjoins(!ShowWeirdAdjoins);
        }
    }
}
